/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <math.h>
#include <gtk/gtk.h>

#include "cpk_bibutton.h"

#define DEFAULT_SPACING   7
#define DEFAULT_LEFT_POS  0
#define DEFAULT_TOP_POS   0
#define MIN_BIBUTTON_WIDTH                 2
#define ARROW_SIZE                         8
#define CPK_BIBUTTON_INITIAL_TIMER_DELAY   200
#define CPK_BIBUTTON_TIMER_DELAY           20
#define MAX_TIMER_CALLS                    5
#define MAX_TEXT_LENGTH                    256
#define EPSILON                            1e-5

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/


/* Local data */

static GtkContainerClass *parent_class = NULL;
static guint              bibutton_signals[BIBUTTON_LAST_SIGNAL] = { 0 };

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static void cpk_bibutton_class_init    (CpkBibuttonClass  *klass) ;
static void cpk_bibutton_init          (CpkBibutton       *button) ;
static void cpk_bibutton_size_request  (GtkWidget         *widget, 
                                        GtkRequisition    *requisition) ;
static void cpk_bibutton_size_allocate (GtkWidget         *widget,
                                        GtkAllocation     *allocation) ;
static void cpk_bibutton_map           (GtkWidget         *widget) ;
static void cpk_bibutton_unmap         (GtkWidget         *widget) ;
static void cpk_bibutton_realize       (GtkWidget         *widget) ;
static void cpk_bibutton_unrealize     (GtkWidget         *widget) ;
static gint cpk_bibutton_expose        (GtkWidget         *widget,
					GdkEventExpose    *event) ;
static void cpk_bibutton_paint         (GtkWidget         *widget, 
                                        GdkRectangle      *area) ;
static void cpk_bibutton_draw          (GtkWidget         *widget,
                                        GdkRectangle      *area) ;
static gint cpk_bibutton_button_press  (GtkWidget         *widget, 
                                        GdkEventButton    *event) ;
static gint cpk_bibutton_button_release(GtkWidget         *widget, 
                                        GdkEventButton    *event) ;
static gint cpk_bibutton_enter_notify  (GtkWidget         *widget, 
                                        GdkEventCrossing  *event) ;
static gint cpk_bibutton_leave_notify  (GtkWidget         *widget, 
                                        GdkEventCrossing  *event) ;
static gint cpk_bibutton_motion_notify (GtkWidget         *widget,  
                                        GdkEventMotion    *event) ;

static void cpk_bibutton_draw_arrow    (CpkBibutton       *bibutton, 
                                        guint              arrow) ;
static gint cpk_bibutton_timer         (CpkBibutton       *bibutton) ;
static void cpk_bibutton_set_adj       (CpkBibutton       *button, 
                                        gfloat             increment) ;
static void adjustment_changed_cb      (GtkAdjustment     *adjustment, 
                                        gpointer           data) ;
static GtkWidget *cpk_bibutton_icon    (GtkWidget         *parent, 
                                        gchar            **xpm_data) ;
static void cpk_bibutton_popup         (CpkBibutton       *button) ;
static gint cpk_bibutton_popup_destroy (CpkBibutton       *button) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

GtkType
cpk_bibutton_get_type (void)
{
  static GtkType bibutton_type = 0;

  if (!bibutton_type)
    {
      static const GtkTypeInfo bibutton_info =
      {
        "CpkBibutton",
        sizeof (CpkBibutton),
        sizeof (CpkBibuttonClass),
        (GtkClassInitFunc) cpk_bibutton_class_init,
        (GtkObjectInitFunc) cpk_bibutton_init,
        /* reserved_1 */ NULL,
        /* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      bibutton_type = gtk_type_unique (GTK_TYPE_BUTTON, &bibutton_info);
    }
  return bibutton_type;
}

static void
cpk_bibutton_class_init (CpkBibuttonClass *class)
{
  GtkObjectClass    *object_class;
  GtkWidgetClass    *widget_class;
  GtkContainerClass *container_class;
  GtkButtonClass    *button_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;
  container_class = (GtkContainerClass*) class;
  button_class = (GtkButtonClass*) class;

  parent_class = gtk_type_class (GTK_TYPE_BUTTON);

   bibutton_signals[LEFT_PRESSED] =
   gtk_signal_new ("left-pressed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (CpkBibuttonClass, lpressed),
                    gtk_marshal_NONE__NONE,
		    GTK_TYPE_NONE, 0) ;
   bibutton_signals[RIGHT_PRESSED] =
   gtk_signal_new ("right-pressed",
                    GTK_RUN_FIRST,
                    object_class->type,
                    GTK_SIGNAL_OFFSET (CpkBibuttonClass, rpressed),
                    gtk_marshal_NONE__NONE,
		    GTK_TYPE_NONE, 0) ;
   gtk_object_class_add_signals (object_class, bibutton_signals, 
         BIBUTTON_LAST_SIGNAL);

   widget_class->size_request = cpk_bibutton_size_request ;
   widget_class->size_allocate = cpk_bibutton_size_allocate ;
   widget_class->map = cpk_bibutton_map ;
   widget_class->unmap = cpk_bibutton_unmap ;
   widget_class->realize = cpk_bibutton_realize ;
   widget_class->unrealize = cpk_bibutton_unrealize ;
   widget_class->draw = cpk_bibutton_draw ;
   widget_class->expose_event = cpk_bibutton_expose ;
   widget_class->button_press_event = cpk_bibutton_button_press ;
   widget_class->button_release_event = cpk_bibutton_button_release ;
   widget_class->enter_notify_event = cpk_bibutton_enter_notify ;
   widget_class->leave_notify_event = cpk_bibutton_leave_notify ;
   widget_class->motion_notify_event = cpk_bibutton_motion_notify ;

   class->lpressed = NULL ;
   class->rpressed = NULL ;

}

static void
cpk_bibutton_init (CpkBibutton *bibutton)
{
   bibutton->adjustment = NULL ;
   bibutton->pixmap = NULL ;
   bibutton->popup = NULL ;
   bibutton->event_window = NULL ;
   bibutton->timer = 0 ;
   bibutton->pop_timer = 0 ;
   bibutton->climb_rate = 0.0 ;
   bibutton->timer_step = 0.0 ;
   bibutton->in_child = 0 ;
   bibutton->click_child = 0 ;
   bibutton->mbutton = 0 ;
   bibutton->need_timer = FALSE ;
   bibutton->timer_calls = 0 ;
   bibutton->wrap = FALSE ;
   bibutton->lactive = FALSE ;
   bibutton->ractive = FALSE ;
   GTK_WIDGET_UNSET_FLAGS (bibutton, GTK_NO_WINDOW) ;

   cpk_bibutton_set_adjustment (bibutton, 
         (GtkAdjustment*) gtk_adjustment_new (0, 0, 0, 0, 0, 0)) ;
}

static void
cpk_bibutton_map (GtkWidget *widget)
{
   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (widget)) ;

   /* if (GTK_WIDGET_NO_WINDOW (widget))
      gdk_window_show (CPK_BIBUTTON (widget)->event_window) ; */

   GTK_WIDGET_CLASS (parent_class)->map (widget) ;
}

static void
cpk_bibutton_unmap (GtkWidget *widget)
{
   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (widget)) ;

   /* if (GTK_WIDGET_NO_WINDOW (widget))
      gdk_window_hide (CPK_BIBUTTON (widget)->event_window) ; */
   GTK_WIDGET_CLASS (parent_class)->unmap (widget) ;
}

static void
cpk_bibutton_realize (GtkWidget *widget)
{
   CpkBibutton  *button ;
   GdkWindowAttr attributes ;
   gint          attributes_mask ;
   gint          border_width ;

   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (widget)) ;

   button = CPK_BIBUTTON (widget) ;
   /* GTK_WIDGET_CLASS (parent_class)->realize (widget) ; */
   GTK_WIDGET_SET_FLAGS (widget, GTK_REALIZED) ;

   border_width = GTK_CONTAINER (widget)->border_width ;

   attributes.window_type = GDK_WINDOW_CHILD ;
   attributes.x = widget->allocation.x + border_width ;
   attributes.y = widget->allocation.y + border_width ;
   attributes.width = widget->allocation.width - border_width * 2 ;
   attributes.height = widget->allocation.height - border_width * 2 ;
   attributes.wclass = GDK_INPUT_OUTPUT ;
   attributes.visual = gtk_widget_get_visual (widget) ;
   attributes.colormap = gtk_widget_get_colormap (widget) ;
   attributes.event_mask = gtk_widget_get_events (widget) ;
   attributes.event_mask |= (GDK_EXPOSURE_MASK | 
                             GDK_BUTTON_PRESS_MASK | 
                             GDK_BUTTON_RELEASE_MASK | 
                             GDK_LEAVE_NOTIFY_MASK | 
                             GDK_ENTER_NOTIFY_MASK | 
                             GDK_POINTER_MOTION_MASK | 
                             GDK_POINTER_MOTION_HINT_MASK) ;

   attributes_mask = GDK_WA_X | GDK_WA_Y | GDK_WA_VISUAL | GDK_WA_COLORMAP ;

   widget->window = gdk_window_new (gtk_widget_get_parent_window (widget),
				    &attributes, attributes_mask) ;
   gdk_window_set_user_data (widget->window, button) ;
   button->event_window = widget->window ;

   widget->style = gtk_style_attach (widget->style, widget->window) ;
   gtk_style_set_background (widget->style, widget->window, GTK_STATE_NORMAL) ;
}

static void
cpk_bibutton_unrealize (GtkWidget *widget)
{
  CpkBibutton *button;

  g_return_if_fail (widget != NULL);
  g_return_if_fail (CPK_IS_BIBUTTON (widget));

  button = CPK_BIBUTTON (widget);

  GTK_WIDGET_CLASS (parent_class)->unrealize (widget);

  return ;

  if (button->event_window)
    {
      gdk_window_set_user_data (button->event_window, NULL);
      gdk_window_destroy (button->event_window);
      button->event_window = NULL;
    }
}

static void
cpk_bibutton_size_request (GtkWidget *widget, GtkRequisition *requisition)
{
   GtkWidget *child ;
   gint       x, y, w, h, child_extra ;
   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (requisition != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (widget)) ;

   GTK_WIDGET_CLASS (parent_class)->size_request (widget, requisition) ;

   child = GTK_BIN (widget)->child ;
   x = MIN_BIBUTTON_WIDTH + 2 * ARROW_SIZE ;
   h =  ARROW_SIZE ;
   if (child && GTK_WIDGET_VISIBLE (child)) {
      x = child->requisition.width + child->requisition.height + 4 ;
      h = child->requisition.height / 1.3 ;
   }
   w = MAX (ARROW_SIZE, h / 2) ;
   x += 2 * (w - h/2) ;

   CPK_BIBUTTON (widget)->w = w ;
   CPK_BIBUTTON (widget)->h = h ;

   requisition->width = x + ARROW_SIZE ;
}

static void
cpk_bibutton_size_allocate (GtkWidget *widget, GtkAllocation *allocation)
{
   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (widget)) ;
   g_return_if_fail (allocation != NULL) ;

   GTK_WIDGET_CLASS (parent_class)->size_allocate (widget, allocation) ;
}

static gint
cpk_bibutton_expose (GtkWidget *widget, GdkEventExpose *event)
{
   GtkBin *bin ;
   GdkEventExpose child_event ;
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_BIBUTTON (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   if (GTK_WIDGET_DRAWABLE (widget)) {
      bin = GTK_BIN (widget) ;
      cpk_bibutton_paint (widget, &event->area) ;
      child_event = *event ;
      if (bin->child && GTK_WIDGET_NO_WINDOW (bin->child) &&
          gtk_widget_intersect (bin->child, &event->area, &child_event.area))
         gtk_widget_event (bin->child, (GdkEvent*) &child_event) ;
   }

  return TRUE ;
}

static gint
cpk_bibutton_enter_notify (GtkWidget *widget, GdkEventCrossing *event)
{
   CpkBibutton *button ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_BIBUTTON (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   button = CPK_BIBUTTON (widget) ;

   if (event->window == widget->window) {
      gint x ;
      gint y ;
      gdk_window_get_pointer (widget->window, &x, &y, NULL) ;

      if (x <= widget->requisition.width / 2) {
         button->in_child = GTK_ARROW_LEFT ;
         if (button->click_child == 0) 
            cpk_bibutton_draw_arrow (button, GTK_ARROW_LEFT) ;
      }
      else {
         button->in_child = GTK_ARROW_RIGHT ;
         if (button->click_child == 0) 
            cpk_bibutton_draw_arrow (button, GTK_ARROW_RIGHT) ;
      }
   }
   return FALSE ;
}

static gint
cpk_bibutton_leave_notify (GtkWidget *widget, GdkEventCrossing *event)
{
   CpkBibutton *button ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_BIBUTTON (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   button = CPK_BIBUTTON (widget) ;

   if (event->window == widget->window && button->click_child == 0) {
      if (button->in_child == GTK_ARROW_RIGHT) {
         button->in_child = 0 ;
         cpk_bibutton_draw_arrow (button, GTK_ARROW_RIGHT) ;
      }
      else {
         button->in_child = 0 ;
         cpk_bibutton_draw_arrow (button, GTK_ARROW_LEFT) ;
      }
   }
   return FALSE ;
}

static gint
cpk_bibutton_motion_notify (GtkWidget *widget, GdkEventMotion *event)
{
   CpkBibutton *button ;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_BIBUTTON (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   button = CPK_BIBUTTON (widget) ;

   if (button->mbutton)
      return FALSE ;

   if (event->window == widget->window) {
      gint x ;

      x = event->x ;
      if (event->is_hint)
         gdk_window_get_pointer (widget->window, &x, NULL, NULL) ;

      if (x <= widget->requisition.width / 2 && 
	   button->in_child == GTK_ARROW_RIGHT) {
         button->in_child = GTK_ARROW_LEFT ;
         cpk_bibutton_draw_arrow (button, GTK_ARROW_LEFT) ;
         cpk_bibutton_draw_arrow (button, GTK_ARROW_RIGHT) ;
      }
      else if (x > widget->requisition.width / 2 && 
	   button->in_child == GTK_ARROW_LEFT) {
         button->in_child = GTK_ARROW_RIGHT ;
         cpk_bibutton_draw_arrow (button, GTK_ARROW_LEFT) ;
         cpk_bibutton_draw_arrow (button, GTK_ARROW_RIGHT) ;
      }
      return FALSE ;
   }

   return GTK_WIDGET_CLASS (parent_class)->motion_notify_event (widget, event) ;
}

static gint
cpk_bibutton_button_press (GtkWidget *widget, GdkEventButton *event)
{
   CpkBibutton *button;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_BIBUTTON (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   button = CPK_BIBUTTON (widget) ;

   if (button->mbutton)
      return FALSE ;
   if (event->type == GDK_BUTTON_PRESS) {

      if (!GTK_WIDGET_HAS_FOCUS (widget))
         gtk_widget_grab_focus (widget) ;
      gtk_grab_add (widget) ;
      button->mbutton = event->button ;

      if (event->x < widget->allocation.width/2) {
         button->click_child = GTK_ARROW_LEFT ;
         if (event->button == 1) {
            cpk_bibutton_set_adj (button, -button->adjustment->step_increment) ;
            if (!button->timer) {
               button->timer_step = button->adjustment->step_increment ;
               button->need_timer = TRUE ;
               button->timer = gtk_timeout_add (CPK_BIBUTTON_INITIAL_TIMER_DELAY,
                     (GtkFunction) cpk_bibutton_timer, (gpointer) button) ;
            }
         }
         cpk_bibutton_draw_arrow (button, GTK_ARROW_LEFT) ;
         
      }
      else {
         button->click_child = GTK_ARROW_RIGHT ;
         if (event->button == 1) {
            cpk_bibutton_set_adj (button, button->adjustment->step_increment) ;
            if (!button->timer) {
               button->timer_step = button->adjustment->step_increment ;
               button->need_timer = TRUE ;
               button->timer = gtk_timeout_add (CPK_BIBUTTON_INITIAL_TIMER_DELAY,
                     (GtkFunction) cpk_bibutton_timer, (gpointer) button) ;
            }
         }
         cpk_bibutton_draw_arrow (button, GTK_ARROW_RIGHT) ;
      }
   }
   return FALSE ;
}

static gint
cpk_bibutton_button_release (GtkWidget *widget, GdkEventButton *event)
{
   CpkBibutton *button;
   guint        click_child;

   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (CPK_IS_BIBUTTON (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   button = CPK_BIBUTTON (widget) ;

   if (event->button == button->mbutton) {

      if (button->timer) {
         gtk_timeout_remove (button->timer) ;
         button->timer = 0 ;
         button->timer_calls = 0 ;
         button->need_timer = FALSE ;
      }

      gtk_grab_remove (widget) ;
      click_child = button->click_child ;
      button->click_child = 0 ;
      button->mbutton = 0 ;
      cpk_bibutton_draw_arrow (button, click_child) ;
   }
   else
      GTK_WIDGET_CLASS (parent_class)->button_release_event (widget, event) ;

   if (button->popup)
      button->pop_timer = gtk_timeout_add (600, 
            (GtkFunction) cpk_bibutton_popup_destroy, (gpointer) button) ;

   return FALSE ;
}

static gint
cpk_bibutton_timer (CpkBibutton *bibutton)
{
   gboolean retval = FALSE ;
  
   GDK_THREADS_ENTER () ;

   if (bibutton->timer) {
      if (bibutton->click_child == GTK_ARROW_RIGHT)
         cpk_bibutton_set_adj (bibutton, bibutton->timer_step) ;
      else
         cpk_bibutton_set_adj (bibutton, -bibutton->timer_step) ;

      if (bibutton->need_timer) {
         bibutton->need_timer = FALSE ;
         bibutton->timer = gtk_timeout_add (CPK_BIBUTTON_TIMER_DELAY, 
               (GtkFunction) cpk_bibutton_timer, (gpointer) bibutton) ;
      }
      else {
         if (bibutton->climb_rate > 0.0 && bibutton->timer_step 
	      < bibutton->adjustment->page_increment) {
            if (bibutton->timer_calls < MAX_TIMER_CALLS)
               bibutton->timer_calls++ ;
            else {
               bibutton->timer_calls = 0 ;
               bibutton->timer_step += bibutton->climb_rate ;
            }
         }
         retval = TRUE ;
      }
   }

   GDK_THREADS_LEAVE () ;

   return retval ;
}

static void
cpk_bibutton_set_adj (CpkBibutton *button, gfloat increment)
{
   GtkAdjustment *adj ;
   gfloat new_value = 0.0 ;

   g_return_if_fail (button != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (button)) ;

   adj = button->adjustment ;

   new_value = adj->value + increment ;

   if (increment > 0) {
      if (button->wrap) {
         if (fabs (adj->value - adj->upper) < EPSILON)
            new_value = adj->lower ;
         else if (new_value > adj->upper)
            new_value = adj->upper ;
      }
      else
         new_value = MIN (new_value, adj->upper) ;
   }
   else if (increment < 0) {
      if (button->wrap) {
         if (fabs (adj->value - adj->lower) < EPSILON)
            new_value = adj->upper ;
         else if (new_value < adj->lower)
            new_value = adj->lower ;
      }
      else
         new_value = MAX (new_value, adj->lower) ;
   }

   if (fabs (new_value - adj->value) > EPSILON) {
      if (button->popup) {
         GdkRectangle area ;
         area.x = area.y = 0 ;
         area.width = button->popup->allocation.width ;
         area.height = button->popup->allocation.height ;
         gtk_widget_draw (button->popup, &area) ;
      }
      else
         cpk_bibutton_popup (button) ;
      gtk_adjustment_set_value (adj, new_value) ;
   }
 }

static void
cpk_bibutton_paint (GtkWidget *widget, GdkRectangle *area)
{
   GtkButton     *button ;
   CpkBibutton   *bibutton ;
   GtkShadowType  shadow_type ;
   gint           width, height,x, y ;

   button = GTK_BUTTON (widget);
   bibutton = CPK_BIBUTTON (widget);
   shadow_type = GTK_SHADOW_NONE ;
  
   if (GTK_WIDGET_DRAWABLE (widget)) {
      x = 0;
      y = 0;
      width = widget->allocation.width - GTK_CONTAINER (widget)->border_width * 2;
      height = widget->allocation.height - GTK_CONTAINER (widget)->border_width * 2;

      gdk_window_set_back_pixmap (widget->window, NULL, TRUE) ;
      gdk_window_clear_area (widget->window, area->x, area->y, area->width, area->height);

      if (GTK_WIDGET_DRAWABLE (widget)) {
         if (shadow_type != GTK_SHADOW_NONE)
            gtk_paint_box (widget->style, widget->window,
               GTK_STATE_NORMAL, shadow_type, area, widget, "bibutton",0, 0, 
               widget->requisition.width, widget->requisition.height) ;
         else {
            gdk_window_set_back_pixmap (widget->window, NULL, TRUE) ;
            gdk_window_clear_area (widget->window, area->x, area->y, area->width, area->height);
         }
         /* Draw first back button, then arrows */
         GTK_WIDGET_CLASS (parent_class)->draw (widget, area) ;
         cpk_bibutton_draw_arrow (bibutton, GTK_ARROW_LEFT) ;
         cpk_bibutton_draw_arrow (bibutton, GTK_ARROW_RIGHT) ;
      }
   }
}

static void
cpk_bibutton_draw (GtkWidget *widget, GdkRectangle *area)
{
   CpkBibutton *button ;
   GdkRectangle child_area ;
   GdkRectangle tmp_area ;
   GtkBin *bin ;

   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (widget)) ;
   g_return_if_fail (area != NULL) ;

   button = CPK_BIBUTTON (widget) ;
   bin = GTK_BIN (widget) ;

   if (GTK_WIDGET_DRAWABLE (widget) /* && !GTK_WIDGET_NO_WINDOW (widget) */) {
      tmp_area = *area ;
      tmp_area.x -= GTK_CONTAINER (widget)->border_width ;
      tmp_area.y -= GTK_CONTAINER (widget)->border_width ;

      cpk_bibutton_paint (widget, &tmp_area) ;

      if (bin->child && 
          gtk_widget_intersect (bin->child, &tmp_area, &child_area))
         gtk_widget_draw (bin->child, &child_area) ;
   }
}

static void
cpk_bibutton_draw_arrow (CpkBibutton *bibutton, guint arrow)
{
   GtkStateType  state_type ;
   GtkShadowType shadow_type ;
   GtkShadowType button_shadow_type ;
   GtkWidget    *widget ;
   gint          x, y, x0 ;

   g_return_if_fail (bibutton != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (bibutton)) ;

   widget = GTK_WIDGET (bibutton) ;

   /* button_shadow_type = gtk_button_get_shadow_type (GTK_BUTTON (bibutton)) ; */
   button_shadow_type = GTK_SHADOW_NONE ;

   if (GTK_WIDGET_DRAWABLE (bibutton)) {
      if (((arrow == GTK_ARROW_RIGHT &&
            (bibutton->adjustment->upper - bibutton->adjustment->value
	    <= bibutton->adjustment->step_increment/2))) || 
          ((arrow == GTK_ARROW_LEFT &&
	    (bibutton->adjustment->value - bibutton->adjustment->lower
	    <= bibutton->adjustment->step_increment/2)))) {
         shadow_type = GTK_SHADOW_ETCHED_IN ;
         state_type = GTK_STATE_NORMAL ;
      }
      else {
         if (bibutton->in_child == arrow) {
            if (bibutton->click_child == arrow)
               state_type = GTK_STATE_ACTIVE;
            else
               state_type = GTK_STATE_PRELIGHT;
         }
         else
            state_type = GTK_STATE_NORMAL;

         if (bibutton->click_child == arrow)
            shadow_type = GTK_SHADOW_IN;
         else
            shadow_type = GTK_SHADOW_OUT;
      }

      x0 = (widget->requisition.width - ARROW_SIZE) / 2 ;
      x = (widget->allocation.width) / 2 ;
      y = (widget->allocation.height - bibutton->h) / 2 + 1 ;

      if (arrow == GTK_ARROW_RIGHT) {
         x += (x0 - bibutton->w) ;
         if (button_shadow_type == GTK_SHADOW_NONE)
            x -= 1 ; 
      }
      else {
         x -= x0 ;
         if (button_shadow_type != GTK_SHADOW_NONE)
            x += 1 ;
      }
      gtk_paint_arrow (widget->style, widget->window, 
            state_type, shadow_type, NULL, widget, "bibutton", 
            arrow, TRUE, x, y, bibutton->w, bibutton->h) ;
   }
}

static void
adjustment_changed_cb (GtkAdjustment *adjustment, gpointer data)
{
  CpkBibutton *bibutton ;

  bibutton = CPK_BIBUTTON (data) ;

  cpk_bibutton_draw_arrow (bibutton, GTK_ARROW_LEFT) ;
  cpk_bibutton_draw_arrow (bibutton, GTK_ARROW_RIGHT) ;
}

static void
cpk_bibutton_value_changed (GtkAdjustment *adjustment, CpkBibutton *bibutton)
{
   char buf[MAX_TEXT_LENGTH];

   g_return_if_fail (adjustment != NULL);
   g_return_if_fail (GTK_IS_ADJUSTMENT (adjustment)) ;

   sprintf (buf, "%0.*f", 2, adjustment->value) ;
  /* Scrivere sul tooltip */
   

   if (bibutton->click_child == GTK_ARROW_RIGHT)
      cpk_bibutton_right_pressed (bibutton) ;
   else
      cpk_bibutton_left_pressed (bibutton) ;

   cpk_bibutton_draw_arrow (bibutton, GTK_ARROW_LEFT) ;
   cpk_bibutton_draw_arrow (bibutton, GTK_ARROW_RIGHT) ;
}

static GtkWidget *
cpk_bibutton_icon (GtkWidget *parent, gchar **xpm_data)
{
   GtkWidget *pixmap ;
   GdkPixmap *bpixmap ;
   GdkBitmap *mask ;
   GtkStyle  *style ;

   style = gtk_widget_get_style (parent) ;
   bpixmap = gdk_pixmap_colormap_create_from_xpm_d (parent->window,
         gtk_widget_get_colormap (parent), 
         &mask, 
         &style->bg[GTK_STATE_NORMAL],
         (gchar **)xpm_data) ;
   pixmap = gtk_pixmap_new (bpixmap, mask) ;

   return pixmap ;
}

static void
cpk_bibutton_popup (CpkBibutton *button)
{
   GtkWidget *pbar ;
   gint       x, y, scr_w, w;

   GDK_THREADS_ENTER () ;

   if (button->pop_timer) {
      gtk_timeout_remove (button->popup) ;
      button->pop_timer = 0 ;
      gtk_widget_destroy (button->popup) ;
   }

   button->popup = gtk_window_new (GTK_WINDOW_POPUP) ;
   /* set_font ("-*-times-medium-r-normal--10-140-*-*-*-*-*-*") ; */
   pbar = gtk_progress_bar_new_with_adjustment (button->adjustment) ;
   gtk_widget_set_usize (button->popup, 120, 22) ;
   gtk_container_add (GTK_CONTAINER (button->popup), pbar) ;
   gtk_progress_set_format_string (GTK_PROGRESS (pbar), "%p%%") ;
   gtk_progress_set_show_text (GTK_PROGRESS (pbar), TRUE) ;
   gtk_widget_show_all (pbar) ;

   gdk_window_get_pointer (NULL, &x, &y, NULL) ;
   x -= 60 ;  
   y -= GTK_WIDGET (button)->allocation.height + 15 ;
   gtk_widget_popup (button->popup, x, y) ;
   /* set_font ("-*-times-medium-r-normal--17-140-*-*-*-*-*-*") ; */

   GDK_THREADS_LEAVE () ;
}

static gint
cpk_bibutton_popup_destroy (CpkBibutton *button)
{
   GDK_THREADS_ENTER () ;

   if (button->pop_timer) {
      gtk_widget_destroy (button->popup) ;
      button->popup = NULL ;
      button->pop_timer = 0 ;
   }

   GDK_THREADS_LEAVE () ;
   return FALSE ;
}

/*----------------------------------------------------------------------------
 *							Public interface
 *--------------------------------------------------------------------------*/

GtkWidget *
cpk_bibutton_new (GtkWidget *parent, gchar **xpm_data, const gchar *tip, 
      const gchar *lab2, GtkAdjustment *adjustment)
{
   GtkTooltips *ttip ;
   GtkWidget   *button ;

   button =  GTK_WIDGET (gtk_type_new (TYPE_CPK_BIBUTTON)) ;

   CPK_BIBUTTON (button)->pixmap = cpk_bibutton_icon (parent, xpm_data) ;
   gtk_container_add (GTK_CONTAINER (button), CPK_BIBUTTON (button)->pixmap) ;

   if (tip) {
      ttip = gtk_tooltips_new() ;
      gtk_tooltips_set_tip(GTK_TOOLTIPS (ttip), button, tip, NULL) ;
   }
   if (adjustment)
      cpk_bibutton_set_adjustment (CPK_BIBUTTON (button), adjustment) ;
   else
      adjustment = CPK_BIBUTTON (button)->adjustment ;
   gtk_adjustment_value_changed (adjustment) ;

   return button ;
}

void
cpk_bibutton_set_adjustment (CpkBibutton *bibutton, GtkAdjustment *adjustment)
{
   g_return_if_fail (bibutton != NULL) ;
   g_return_if_fail (CPK_IS_BIBUTTON (bibutton)) ;

   if (bibutton->adjustment != adjustment) {
      if (bibutton->adjustment) {
         gtk_signal_disconnect_by_data (GTK_OBJECT (bibutton->adjustment), 
               (gpointer) bibutton) ;
         gtk_object_unref (GTK_OBJECT (bibutton->adjustment)) ;
      }
      bibutton->adjustment = adjustment ;
      if (adjustment) {
         gtk_object_ref (GTK_OBJECT (adjustment)) ;
         gtk_object_sink (GTK_OBJECT (adjustment)) ;
         gtk_signal_connect (GTK_OBJECT (adjustment), "value_changed", 
               (GtkSignalFunc) cpk_bibutton_value_changed, 
               (gpointer) bibutton) ;
         gtk_signal_connect (GTK_OBJECT (adjustment), "changed",
               (GtkSignalFunc) adjustment_changed_cb, (gpointer) bibutton) ;
      }
   }
}

GtkAdjustment *
cpk_bibutton_get_adjustment (CpkBibutton *bibutton)
{
  g_return_val_if_fail (bibutton != NULL, NULL) ;
  g_return_val_if_fail (CPK_IS_BIBUTTON (bibutton), NULL) ;

  return bibutton->adjustment;
}

void
cpk_bibutton_left_pressed (CpkBibutton *bibutton)
{
  g_return_if_fail (bibutton != NULL);
  g_return_if_fail (CPK_IS_BIBUTTON (bibutton));

  gtk_signal_emit (GTK_OBJECT (bibutton), bibutton_signals[LEFT_PRESSED]);
}

void
cpk_bibutton_right_pressed (CpkBibutton *bibutton)
{
  g_return_if_fail (bibutton != NULL);
  g_return_if_fail (CPK_IS_BIBUTTON (bibutton));

  gtk_signal_emit (GTK_OBJECT (bibutton), bibutton_signals[RIGHT_PRESSED]);
}
