/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include <math.h>

#include <gtk/gtk.h>

#include "catalib.h"
#include "cpk_gtk_util.h"
#include "cpk_plot.h"
#include "cpk_map.h"

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef M_PI
#define M_PI    (3.14159265358979323846)
#endif   /* M_PI */
#define TWOPI (2.0*M_PI)
#ifndef RADTODEG
#define RADTODEG(a) ((a)*360.0/TWOPI)
#endif
#ifndef DEGTORAD
#define DEGTORAD(a) ((a)*TWOPI/360.0)
#endif

#define HYPOT(a,b) (sqrt ((a)*(a)+(b)*(b)))

#define MAP_NBRIGHT  (250)
#define MAP_NPLOT    (2000)

enum {
   MAP_NORMAL,
   MAP_FLAGGED,
   MAP_HIGHLIGHT,
   MAP_GRAY,
   MAP_SHADE,
   MAP_DISPL
} ;

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

/* Local data */

static CpkDrawClass  *parent_class = NULL ;
gdouble sph_scale = 3600.0/0.25 ;
gdouble lin_scale = 1.0 ;

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/
                                               
static void        cpk_map_class_init       (CpkMapClass      *klass) ;

static void        cpk_map_init             (CpkMap           *map) ;
 
static gdouble     cpk_map_x             (CpkDraw   *draw, 
                                          gint       i) ;
static gdouble     cpk_map_y             (CpkDraw   *draw, 
                                          gint       i) ;
static void        cpk_map_button_press  (CpkDraw   *draw, 
                                          GtkWidget *darea,
                                          gdouble    x,
                                          gdouble    y,
                                          GdkModifierType state) ;
static void        cpk_map_geometry      (CpkDraw   *draw, 
                                          GtkWidget *darea, 
                                          gdouble    x, 
                                          gdouble    y) ;
static void        cpk_map_draw          (CpkDraw   *draw, 
                                          GdkPixmap *pixmap, 
                                          gint       mode) ;

static void        world_to_draw         (CpkDraw       *draw,
                                          gdouble        xw,
                                          gdouble        yw,
                                          gdouble       *xd,
                                          gdouble       *yd) ;

static void        draw_to_world         (CpkDraw       *draw,
                                          gdouble        xd,
                                          gdouble        yd,
                                          gdouble       *xw,
                                          gdouble       *yw) ;

static void        cpk_map_check_fnorm   (CpkDraw       *draw) ;
static GdkGC     **cpk_map_set_stack_gc  (CpkDraw       *draw) ;
static void        cpk_map_free_stack_gc (CpkMap        *map) ;
static void        cpk_map_set_const     (CpkDraw       *draw) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

GtkType
cpk_map_get_type (void)
{
  static GtkType map_type = 0;

  if (!map_type)
    {
      static const GtkTypeInfo map_info =
      {
	"CpkMap",
	sizeof (CpkMap),
	sizeof (CpkMapClass),
	(GtkClassInitFunc) cpk_map_class_init,
	(GtkObjectInitFunc) cpk_map_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      map_type = gtk_type_unique (TYPE_CPK_DRAW, &map_info);
    }

  return map_type ;
}

static void
cpk_map_class_init (CpkMapClass *class)
{
   GtkObjectClass    *object_class ;
   GtkWidgetClass    *widget_class ;
   CpkDrawClass      *draw_class ;

   object_class = (GtkObjectClass*) class ;
   widget_class = (GtkWidgetClass*) class ;
   draw_class = (CpkDrawClass*) class ;

   parent_class = gtk_type_class (TYPE_CPK_MAP) ;

   draw_class->x = cpk_map_x ;
   draw_class->y = cpk_map_y ;
   draw_class->draw_geometry = cpk_map_geometry ;
   draw_class->draw_button_press = cpk_map_button_press ;
   draw_class->draw_paint = cpk_map_draw ;
   draw_class->draw_pan = NULL ;
   draw_class->draw_pan_border = NULL ;
   draw_class->draw_mgf = NULL ;
   draw_class->world2draw = world_to_draw ;
   draw_class->draw2world = draw_to_world ;
   draw_class->world2pan = NULL ;
   draw_class->pan2world = NULL ;
   draw_class->world2mgf = NULL ;
}

static void
cpk_map_init (CpkMap *map)
{
   map->dragx = 0.0 ;
   map->dragy = 0.0 ;
}

/****** non serve : si puo' delegare la realizzazione alle sottoclassi 
                    cpk_map, cpk_graph, cpk_histo, cpk_map... */
GtkWidget *
cpk_map_new (GtkWidget *widget)
{
   CpkMap *map ;

   g_return_val_if_fail (widget != NULL, NULL) ;
   g_return_val_if_fail (CPK_IS_PLOT (widget), NULL) ;
   
   map = gtk_type_new (cpk_map_get_type ()) ;

   CPK_DRAW (map)->nx = widget->allocation.width ;
   CPK_DRAW (map)->ny = widget->allocation.height ;

   gtk_container_add (GTK_CONTAINER (widget), GTK_WIDGET (map)) ;

   return GTK_WIDGET (map) ;
}

static gdouble 
cpk_map_x (CpkDraw *draw, gint i)
{
   CpkMap *map = CPK_MAP (draw) ;
   return map->list->data[i]->xrot ;
}

static gdouble 
cpk_map_y (CpkDraw *draw, gint i)
{
   CpkMap *map = CPK_MAP (draw) ;
   return map->list->data[i]->yrot ;
}


static void
cpk_map_button_press  (CpkDraw   *draw, 
                       GtkWidget *darea, 
                       gdouble    x, 
                       gdouble    y, 
                       GdkModifierType state)
{
   CpkMap *map = CPK_MAP (draw) ;

   if (map->list == NULL)
      return ;
   if (state & (GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) 
      return ;

   CPKDEBUG5 ("cpk_map_button_press", 5) ;

   draw_to_world (draw, x, y, &map->dragx, &map->dragy) ;

   map->dragz = map->zoom ;
   map->dragd = HYPOT (x - darea->allocation.width / 2.0,
                       darea->allocation.height / 2.0 - y) ;
   map->draga = map->rot ;
   map->dragr = atan2 (darea->allocation.height / 2.0 - y, 
                       x - darea->allocation.width / 2.0) ;
}

static void
cpk_map_geometry (CpkDraw *draw, GtkWidget *darea, gdouble x, gdouble y)
{
   CpkMap *map = CPK_MAP (draw) ;
   gdouble dx, dy, dr, ix ;
   guint p_action = CPK_PLOT (GTK_WIDGET (draw)->parent)->p_action ;

   switch (p_action) {
      case PAN :
         map->panx -= (draw->xcur-map->dragx) ;
         map->pany -= (draw->ycur-map->dragy) ;
         break ;
      case ZOOM :
         if (ABS (map->dragd) < 1.0e-5)
            return ;
         dx = x - darea->allocation.width / 2.0 ;
         dy = darea->allocation.height / 2.0 - y ;
         dr = HYPOT (dx, dy) ;
         map->zoom = map->dragz * MAX ( ABS (dr / map->dragd), 1.0e-5) ;
         map->izoom = -1 ;
         break ;
      case ROTATE :
         dx = x - darea->allocation.width / 2.0 ;
         dy = darea->allocation.height / 2.0 - y ;
         if (ABS (dx) < 1.0e-15 && ABS (dy) < 1.0e-15) 
            return ;
         dr =  atan2 (dy, dx) ;
         ix = ((map->list->out_sph) ? 1.0 : -1.0) * map->swx * map->swy ;
         map->rot = ang_res_sym (map->draga - ix * (dr - map->dragr)) ;
         break ;
   }
}

static void
cpk_map_draw (CpkDraw *draw, GdkPixmap *pixmap, gint mode)
{
   CpkMap    *map = CPK_MAP (draw) ;
   CpkPlot   *plot = CPK_PLOT (GTK_WIDGET (draw)->parent) ;
   gint       i, nplot ;
   guint      j ;
   gdouble    x, y, r, rmx, mag, x0, y0 ;
   gdouble    dx1, dx2, dy1, dy2 ;
   Stack     *tab ;
   gboolean   displacement = (!CPK_PLOT_BLINKING (plot) && 
                              map->model->status > HIGH_ALIGN) ;

   if (!map->list)
      return ;
   CPKDEBUG ("cpk_map_draw");

   cpk_map_set_const (draw) ;

   dx1 = dy1 = 1.0e10 ;
   dx2 = dy2 = -1.0e10 ;

   /* Set GCs for STACK MAP only */
   if (mode != 2)
      cpk_map_check_fnorm (draw) ;

   rmx = map->radius * map->zoom ;
   nplot = CPK_PLOT_MOVING (plot) ? 
           MIN (MAP_NPLOT, map->list->nstar) : map->list->nstar ; 

   for (i = nplot-1; i >= 0; i--) {
      world_to_draw (draw, cpk_map_x (draw, i), cpk_map_y (draw, i), &x, &y) ;
      if (CPK_PLOT_SCALE (plot)) {
         if (x < dx1)
            dx1 = x ;
         if (x > dx2)
            dx2 = x ;
         if (y < dy1)
            dy1 = y ;
         if (y > dy2)
            dy2 = y ;
      }
      if (x > -rmx && x < draw->nx + rmx && y > -rmx && y < draw->ny + rmx) {
         mag = map->list->data[i]->mcal ;
         mag = (mag > map->mcut2) ? map->mcut2 : 
               (mag < map->mcut1) ? map->mcut1 : mag ;
         r = rmx * (map->mcut2 - mag) / map->mrange + 0.5 ;
         switch (mode) {
         case 0 :
            if (displacement) {
               world_to_draw (draw, 
                              (map->list->data[i]->xprj * map->model->pc[0][0] + 
                               map->list->data[i]->yprj * map->model->pc[0][1]) * 
                               map->model->cdelt[0],
                              (map->list->data[i]->xprj * map->model->pc[1][0] + 
                               map->list->data[i]->yprj * map->model->pc[1][1]) *
                               map->model->cdelt[1], 
                              &x0, &y0) ;
               cpk_plot_seg (pixmap, map->gc[MAP_DISPL], x0, y0, x, y) ;
            }
            if (map->list->data[i]->flag == 0) 
               cpk_plot_symb (pixmap, map->gc[MAP_NORMAL], CPK_CIRCLE, x, y, 0., r, 25) ;
            else
               cpk_plot_symb (pixmap, map->gc[MAP_FLAGGED], CPK_CIRCLE, x, y, 0., r, 25) ;
            break ;
         case 1 :
            /* cpk_plot_symb (pixmap, map->gc[MAP_SHADE], CPK_CROSS, x, y, 0., r, 9) ; */
            cpk_plot_symb (pixmap, map->gc[MAP_SHADE], CPK_CIRCLE, x, y, 45., r, 4) ;
            break ;
         case 2 :
            tab = (Stack*)map->list->data[i]->p ;
            j = bin_get_count (tab->ns) + bin_get_count (tab->ns2) - 1 ;            
            cpk_plot_symb (pixmap, map->stack_gc[j], CPK_CIRCLE, x, y, 0., r, 25) ;
            break ;
         }
      }
   }
   if (CPK_PLOT_SCALE (plot)) {
      draw->dx = dx2 - dx1 ; 
      draw->dy = dy2 - dy1 ;
      CPK_PLOT_UNSET_FLAGS (plot, CPK_SCALE) ;
   }

   /* map->origin[0] = map->model->crpix[0] ;
   map->origin[1] = map->model->crpix[1] ; */
   map->origin[0] = map->list->origin[0] ;
   map->origin[1] = map->list->origin[1] ;
}

static void
world_to_draw (CpkDraw *draw, gdouble x, gdouble y, gdouble *xd, gdouble *yd)
{
   CpkMap *map = CPK_MAP (draw) ;
   gdouble dx, dy ;
   dx = x - map->panx ;
   dy = y - map->pany ;
   *xd = (map->cos_rot*dx - map->sin_rot*dy) * map->map_scale * map->swx *
         map->swxx + draw->nx / 2.0 ;
   *yd = (-map->sin_rot*dx - map->cos_rot*dy) * map->map_scale * map->swy +
          draw->ny / 2.0 ;
}

static void
draw_to_world (CpkDraw *draw, gdouble xd, gdouble yd, gdouble *x, gdouble *y)
{
   CpkMap *map = CPK_MAP (draw) ;
   gdouble dx, dy ;
   dx = (draw->nx / 2.0 - xd) * map->swxx * map->swx ;
   dy = (draw->ny / 2.0 - yd) * map->swy ;
   *x = - ( map->cos_rot*dx - map->sin_rot*dy) / map->map_scale + map->panx ;
   *y = - (-map->sin_rot*dx - map->cos_rot*dy) / map->map_scale + map->pany ;
}

static void
cpk_map_check_fnorm (CpkDraw *draw)
{
   CpkMap  *map = CPK_MAP (draw) ;
   if (map->stack_gc && map->ngc != map->fnorm)
      cpk_map_free_stack_gc (map) ;
   if (!map->stack_gc)
      map->stack_gc = cpk_map_set_stack_gc (draw) ;
}

static GdkGC **
cpk_map_set_stack_gc (CpkDraw *draw)
{
   CpkMap  *map = CPK_MAP (draw) ;
   GdkGC  **gc = NULL ;
   guint    i, nf ;
   gfloat   f ;

   nf = map->fnorm ;
   gc = (GdkGC **) calloc ((size_t)nf, sizeof(GdkGC *)) ;

   for (i = 0; i < nf; i++) {
      gc[i] = gdk_gc_new (CPK_PLOT (GTK_WIDGET (draw)->parent)->parent->window) ;
      if (nf < 2)
         f = 1.0 ;
      else
         f = i / (gfloat) (map->fnorm-1) ;
      cpk_set_ns_color (CPK_PLOT (GTK_WIDGET (draw)->parent)->parent, gc[i], f) ;
   }
   map->ngc = nf ;

   return gc ;
}

static void
cpk_map_free_stack_gc (CpkMap *map)
{
   guint  i = 0 ;
   for (i = 0; i < map->ngc; i++) 
      gdk_gc_unref (map->stack_gc[i]) ;
   free (map->stack_gc) ;
   map->stack_gc = NULL ;
}

static void
cpk_map_set_const (CpkDraw *draw)
{
   CpkMap *map = CPK_MAP (draw) ;
   gdouble scale ;

   map->swxx = (map->list->out_sph) ? -1.0 : 1.0 ;

   scale = (map->list->out_sph) ? sph_scale : lin_scale ;
   scale *= map->zoom ;

   if (map->map_scale != scale || 
       map->cos_rot != cos (map->rot) ||
       map->sin_rot != sin (map->rot)) {

      CPK_PLOT_SET_FLAGS (CPK_PLOT (GTK_WIDGET (draw)->parent), CPK_SCALE) ;

      map->map_scale = scale ;
      map->cos_rot = cos (map->rot) ;
      map->sin_rot = sin (map->rot) ;
   }
}

