/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include <math.h>
#include "common.h"

#include <gtk/gtk.h>
#include "cpk_bibutton.h"
#include "cataxcorr_p.h"
#include "starlist.h"

#include "cxc_gtk_main.h"
#include "cxc_data.h"
#include "cxc_map.h"
#include "cxc_buttons.h"
#include "cxc_magnifier.h"
#include "cxc_frames.h"


/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#define IS_PAIR ( display->mode == FRAME_PAIR || display->nframes == 2 )

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static GtkWidget  *frame_get                      (GtkWidget        *map,
                                                   cxc_display      *display) ;
static void        reset_frame_mode               (cxc_display      *display,
                                                   enum mode_tag     mode) ;
static void        cxc_rframe_show_current        (cxc_display      *display) ;

static void        set_paner                      (cxc_display      *display) ;

static void        frame_reparent_box             (GtkWidget        *box, 
                                                   cxc_display      *display) ;
static void        frame_reparent_table           (cxc_display      *display) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

void
cxc_frames_init (cxc_display *display)
{
   guint     i, nfr ;

   display->nframes = (guint) CXC_DATA (display->data)->n ;
   display->user_reference = display->nframes ;
   nfr = display->nframes + 1 ;
   if (display->nframes < 4) 
      display->tile_mode = 1 ;

   display->paner  = (GtkWidget **) malloc (nfr * sizeof(GtkWidget *)) ;
   display->frames = (GtkWidget **) malloc (nfr * sizeof(GtkWidget *)) ;

   display->show =  (guint*) malloc (display->nframes * sizeof(guint)) ;
   display->blink = (guint*) malloc (display->nframes * sizeof(guint)) ;

   for (i = 0; i < display->nframes; i++) {
      display->frames[i] = frame_get (display->map[i], display) ;
      display->show[i] = 1 ;
   }

   display->rframe = frame_get (display->map[display->nframes], display) ;
}

GtkWidget *
frames_realize (GtkWidget *window, cxc_display *display)
{
   guint      ncol, nrow, nf ;
   GtkWidget *box, *hbox, *vbox, *tbox ;

   nf = display->nframes ;
   ncol = ceil (sqrt ((gdouble)nf)) ;
   nrow = ceil (nf / (gdouble)ncol) ;

   box = gtk_hbox_new (FALSE, 0) ;

   display->frame_hbox = hbox = gtk_hbox_new (FALSE, 2) ;
   display->frame_vbox = vbox = gtk_vbox_new (FALSE, 2) ;
   display->frame_tbox = tbox = gtk_table_new (nrow, ncol, FALSE) ;
   display->tbox_nrow = nrow ;
   display->tbox_ncol = ncol ;
   gtk_table_set_row_spacings (GTK_TABLE (tbox), 2);
   gtk_table_set_row_spacings (GTK_TABLE (tbox), 2);

   gtk_container_add (GTK_CONTAINER (box), hbox) ;
   gtk_container_add (GTK_CONTAINER (box), vbox) ;
   gtk_container_add (GTK_CONTAINER (box), tbox) ;

   cxc_frame_reparent (display) ;

   gtk_container_add (GTK_CONTAINER (box), display->rframe) ;

   gtk_widget_show_all (box) ;
   return box ;
}

void
cxc_frame_reparent (cxc_display *display)
{
   if (!display->frame_hbox)
      return ;
   gtk_widget_hide (display->frame_hbox) ;
   gtk_widget_hide (display->frame_vbox) ;
   gtk_widget_hide (display->frame_tbox) ;

   switch  (display->mode) {
      case FRAME_TILE :
         switch (display->tile_mode) {
            case 0 : 
               frame_reparent_table (display) ;
               break ;
            case 1 : 
               frame_reparent_box (display->frame_hbox, display) ;
               break ;
            case 2 :
               frame_reparent_box (display->frame_vbox, display) ;
               break ;
            default :
               ;
         }
         break ;
      case FRAME_BLINK : case FRAME_COINC :
         break ;
      default :
         if (display->last_tile_mode == 1)
            frame_reparent_box (display->frame_hbox, display) ;
         else
            frame_reparent_box (display->frame_vbox, display) ;
      break ;
   }
   cxc_set_cursor (NULL, display) ;
}

static void
frame_reparent_box (GtkWidget *new_parent, cxc_display *display)
{
   GtkWidget *old_parent = NULL ;
   guint i ;
   if (!display->frames[0])
      return ;
   if (!new_parent)
      return ;
   for (i = 0; i < display->nframes; i++) {
      old_parent = GTK_WIDGET (display->frames[i])->parent ;
      if (old_parent) {
         gtk_widget_ref (GTK_WIDGET (display->frames[i])) ;
         gtk_container_remove (GTK_CONTAINER (old_parent), 
               GTK_WIDGET (display->frames[i])) ;
      }
      gtk_container_add (GTK_CONTAINER (new_parent), 
            GTK_WIDGET (display->frames[i])) ;
      if (old_parent) 
         gtk_widget_unref (GTK_WIDGET (display->frames[i])) ;
   }
   gtk_widget_show (new_parent) ;
}

static void
frame_reparent_table (cxc_display *display)
{
   GtkWidget *old_parent ;
   guint i, nc, nr, nf, ncol ;

   nf = 0 ;
   for (i = 0; i < display->nframes; i++)
      if (display->show[i])
         nf++ ;
   ncol = ceil (sqrt ((gdouble)nf)) ;

   if (!display->frames[0])
      return ;

   nc = nr = 0 ;
   for (i = 0; i < display->nframes; i++) {
      old_parent = GTK_WIDGET (display->frames[i])->parent ;
      if (old_parent) {
         gtk_widget_ref (GTK_WIDGET (display->frames[i])) ;
         gtk_container_remove (GTK_CONTAINER (old_parent), 
               GTK_WIDGET (display->frames[i])) ;
      }
      gtk_table_attach (GTK_TABLE (display->frame_tbox), 
            GTK_WIDGET (display->frames[i]), nc, nc+1, nr, nr+1,
            GTK_SHRINK | GTK_EXPAND | GTK_FILL, 
            GTK_SHRINK | GTK_EXPAND | GTK_FILL, 0, 0) ;
      if (display->show[i]) {
         if (nc < ncol-1)
            nc++ ;
         else {
            nc = 0 ;
            nr++ ;
         }
      }
      if (old_parent) 
         gtk_widget_unref (GTK_WIDGET (display->frames[i])) ;
   }
   gtk_widget_show (display->frame_tbox) ;
}

static GtkWidget *
frame_get (GtkWidget *map, cxc_display *display)
{
   GtkWidget *frame ;

   frame = gtk_frame_new (NULL) ;
   gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 0) ;
   gtk_container_add (GTK_CONTAINER (frame), CXC_MAP (map)->draw_area) ;

   return frame ;
}

void 
frame_reset (GtkWidget *toggle, cxc_display *display)
{
   CxcMap  *map = CXC_MAP (display->map[display->current]) ;
   CxcData *data = CXC_DATA (display->data) ;
   gint     n = display->current ;

   if (cpk_confirm (display->window, "CataXcorr: Reset Frame", 
         " Reset all settings for this frame ? "))
      return ;

   cxc_data_stack_remove (data, n) ;

   data->model[n] = model_init (data->list[data->nref], data->list[n], 
         data->model[n]) ;
   data->list[n]->out_sph = data->list[n]->spheric ;
   data->list[n]->proj = FALSE ;
   star_list_pix2prj (data->list[n], data->model[n]->crpix) ;
   cxc_map_reset (map) ;
   cxc_buttons_swap (display) ;
   cxc_buttons_wcs_status (display) ;
}

gint
cxc_frame_select (cxc_display *display)
{
   guint  i ;
   if (IS_PAIR) {
      for (i = 0; i < display->nframes; i++)
         if (i != CXC_DATA (display->data)->nref)
            if (GTK_IS_WIDGET (display->frames[i]) && 
                GTK_WIDGET_VISIBLE (display->frames[i]))
               return cxc_data_file_select (CXC_DATA (display->data), i) ;
   }
   else 
      if (display->current != CXC_DATA (display->data)->nref &&
          display->current != display->nframes) 
         return cxc_data_file_select (CXC_DATA (display->data), display->current) ;
   return 1 ;
}

static void
set_paner (cxc_display *display)
{
   guint i ;
   /* Map current paner frame */
   for (i = 0; i <= display->nframes; i++)
      if (GTK_IS_WIDGET (display->paner[i]))
         gtk_widget_hide (display->paner[i]) ;
   if (display->paner[display->current])
      gtk_widget_show (display->paner[display->current]) ;
}

guint
cxc_second_frame (cxc_display *display)
{
   guint i ;
   if (IS_PAIR) {
      for (i = 0; i < display->nframes; i++)
         if (i != CXC_DATA (display->data)->nref)
            if (GTK_IS_WIDGET (display->frames[i]) && 
                GTK_WIDGET_VISIBLE (display->frames[i]))
               return i ;
   }
   else 
      return display->current ;
   return display->nframes ;
}

static void 
reset_frame_mode (cxc_display *display, enum mode_tag mode)
{
   gint  i ;
   CPKDEBUG5 ("reset_frame_mode",5) ;

   /* remove the timer on blinking */
   if (display->mode == FRAME_BLINK)
      blink_timeout_remove (display) ;

   /* This causes both frames to be reconfigured */
   if (display->mode == FRAME_MERGE || mode == FRAME_MERGE) {
      CXC_MAP_REMAP_LATER (display->map[CXC_DATA (display->data)->nref]) ;
      CXC_MAP_REMAP_LATER (display->map[display->current]) ;
   }

   /* Hide all frames */
   for (i = 0; i < display->nframes; i++)
      if (GTK_IS_WIDGET (display->frames[i]) && 
          GTK_WIDGET_VISIBLE (display->frames[i]))
         gtk_widget_hide (display->frames[i]) ;

   if (GTK_IS_WIDGET (display->rframe) && GTK_WIDGET_VISIBLE (display->rframe))
      gtk_widget_hide (display->rframe) ;

   cxc_buttons_frame_mode (display, mode) ;

   /* Last, set current mode */ 
   display->mode = mode ;
   cxc_frame_reparent (display) ;
}

void 
set_frame_single (cxc_display *display, guint action, GtkWidget *item)
{
   if (display->mode == FRAME_SINGLE)
      return ;
   CPKDEBUG ("set_frame_single") ;

   reset_frame_mode (display, FRAME_SINGLE) ;

   if (display->current == display->nframes)
      frame_select (display, display->old_current) ;
   else
      frame_select (display, display->current) ;

   if (!display->show[display->current])
      frame_next (NULL, display) ;

   if (display->frames[display->current])
      gtk_widget_show_all (display->frames[display->current]) ;

   cxc_set_cursor (NULL, display) ;
   cxc_buttons_mode_status (display, FRAME_SINGLE) ;
}

void 
set_frame_pair (cxc_display *display, guint action, GtkWidget *item)
{
   guint n ;

   if (display->mode == FRAME_PAIR)
      return ;
   CPKDEBUG ("set_frame_pair") ;

   reset_frame_mode (display, FRAME_PAIR) ;
   n = CXC_DATA (display->data)->nref ;

   if (display->current == display->nframes)
      frame_select (display, display->old_current) ;
   else
      frame_select (display, display->current) ;

   if (!display->show[display->current] || display->current == n)
      frame_next (NULL, display) ;

   if (display->frames[n])
      gtk_widget_show_all (display->frames[n]) ;
   if (display->frames[display->current])
      gtk_widget_show_all (display->frames[display->current]) ;
   cxc_set_cursor (NULL, display) ;
   cxc_buttons_mode_status (display, FRAME_PAIR) ;
}

void 
set_frame_tile (cxc_display *display, guint action, GtkWidget *item)
{
   guint i ;

   if (display->mode == FRAME_TILE)
      return ;
   CPKDEBUG ("set_frame_tile") ;

   reset_frame_mode (display, FRAME_TILE) ;

   if (display->current == display->nframes)
      frame_select (display, display->old_current) ;
   else
      frame_select (display, display->current) ;

   if (!display->show[display->current])
      frame_next (NULL, display) ;

   for (i = 0; i < display->nframes; i++)
      if (display->show[i] && GTK_IS_WIDGET (display->frames[i]))
         gtk_widget_show_all (display->frames[i]) ;

   cxc_set_cursor (NULL, display) ;
   cxc_buttons_mode_status (display, FRAME_TILE) ;
}

void 
set_frame_blink (cxc_display *display, guint action, GtkWidget *item)
{
   guint32 rate[] = { 100, 200, 500, 1000, 2000, 4000, 8000 } ;
   guint   i ;

   if (display->mode == FRAME_BLINK)
      return ;
   CPKDEBUG ("set_frame_blink") ;

   /* Check for active frames */
   if (IS_PAIR) {
      for (i = 0; i < display->nframes; i++)
         if (GTK_IS_WIDGET (display->frames[i]) && 
             GTK_WIDGET_VISIBLE (display->frames[i]))
            display->blink[i] = 1 ;
         else
            display->blink[i] = 0 ;
   }
   else
      for (i = 0; i < display->nframes; i++)
         display->blink[i] = display->show[i] ;

   reset_frame_mode (display, FRAME_BLINK) ;
   frame_select (display, display->nframes) ;

   gtk_widget_hide (display->frame_hbox) ;
   gtk_widget_hide (display->frame_vbox) ;
   gtk_widget_hide (display->frame_tbox) ;
   gtk_widget_show_all (display->rframe) ;

   /* Add the timer on blinking */
   display->timer = gtk_timeout_add (rate[display->blink_rate], 
         blink_timer, display) ;
   cxc_buttons_mode_status (display, FRAME_BLINK) ;
}

void 
set_frame_merge (cxc_display *display, guint action, GtkWidget *item)
{
   guint n ;

   if (display->mode == FRAME_MERGE)
      return ;
   CPKDEBUG ("set_frame_merge") ;

   n = CXC_DATA (display->data)->nref ;
   reset_frame_mode (display, FRAME_MERGE) ;

   if (display->current == display->nframes)
      frame_select (display, display->old_current) ;
   else
      frame_select (display, display->current) ;

   if (!display->show[display->current] || display->current == n)
      frame_next (NULL, display) ;

   gtk_widget_show_all (display->frames[n]) ;
   gtk_widget_hide (display->frames[n]) ;

   if (display->frames[display->current])
      gtk_widget_show_all (display->frames[display->current]) ;

   cxc_set_cursor (NULL, display) ;
   cxc_buttons_mode_status (display, FRAME_MERGE) ;
}

void 
set_frame_coinc (cxc_display *display, guint action, GtkWidget *item)
{
   if (display->mode == FRAME_COINC)
      return ;
   CPKDEBUG ("set_frame_coinc") ;

   reset_frame_mode (display, FRAME_COINC) ;
   frame_select (display, display->nframes) ;

   gtk_widget_hide (display->frame_hbox) ;
   gtk_widget_hide (display->frame_vbox) ;
   gtk_widget_hide (display->frame_tbox) ;
   gtk_widget_show_all (display->rframe) ;

   cxc_set_cursor (NULL, display) ;
   cxc_buttons_mode_status (display, FRAME_COINC) ;
}

void
frame_next (GtkWidget *button, cxc_display *display)
{
   guint n = display->current ;
   guint i = display->current ;
   guint r = CXC_DATA (display->data)->nref ;
   CPKDEBUG ("frame_next") ;

   if (display->mode == FRAME_PAIR) {
      for (n = 0; n < display->nframes; n++)
         if (GTK_WIDGET_VISIBLE (display->frames[n]) && n != r)
            break ;
      if (n == display->nframes)
         n = display->current ;
   }
   do {
      i++ ;
      if (i >= display->nframes)
         i = 0 ;
   }
   while ((!display->show[i] && i != n) || 
          (i == r && display->mode == FRAME_PAIR)) ;

   if (display->mode == FRAME_SINGLE || display->mode == FRAME_PAIR) {
      if (display->frames[n])
         gtk_widget_hide (display->frames[n]) ;
      gtk_widget_show (display->frames[i]) ;
   }

   frame_select (display, i) ;
}

void
frame_prev (GtkWidget *button, cxc_display *display)
{
   guint n = display->current ;
   guint i = display->current ;
   guint r = CXC_DATA (display->data)->nref ;
   CPKDEBUG ("frame_prev") ;

   if (display->mode == FRAME_PAIR) {
      for (n = 0; n < display->nframes; n++)
         if (GTK_WIDGET_VISIBLE (display->frames[n])&&n!=r)
            break ;
      if (n == display->nframes)
         n = display->current ;
   }
   do {
      i-- ;
      if ((gint)i < 0)
         i = display->nframes-1 ;
   }
   while ((!display->show[i] && i != n) || 
          (i == r && display->mode == FRAME_PAIR)) ;

   if (display->mode == FRAME_SINGLE || display->mode == FRAME_PAIR) {
      if (display->frames[n])
         gtk_widget_hide (display->frames[n]) ;
      gtk_widget_show (display->frames[i]) ;
   }

   frame_select (display, i) ;
}

gint
blink_timer (gpointer data)
{
   cxc_display *display = (cxc_display *)data ;
   guint        count = 0 ;

   do {
      count++ ;
      display->current++ ;
      if (display->current >= display->nframes)
         display->current = 0 ;
   }
   while (!display->blink[display->current] && count <= display->nframes) ;

   cxc_rframe_show_current (display) ;

   /* As this is a timeout function, return TRUE so that it 
    * continues to get called */
   return (TRUE);
}

static void
cxc_rframe_show_current (cxc_display *display)
{
   GtkWidget   *draw = CXC_MAP (display->map[display->nframes])->draw_area ;

   cxc_map_configure_pixmap (draw, CXC_MAP (display->map[display->current])) ;
   gdk_draw_pixmap (draw->window,
         draw->style->fg_gc[GTK_WIDGET_STATE (draw)],
         CXC_MAP (display->map[display->current])->pixmap,
         0, 0, 0, 0, draw->allocation.width, draw->allocation.height) ;

   if (!CXC_MAP (display->map[display->current])->pixmap_mgf)
      cxc_mgf_remap (NULL, display) ;
   else
      gdk_draw_pixmap (display->magnifier->window,
         display->magnifier->style->fg_gc[GTK_WIDGET_STATE (display->magnifier)],
         CXC_MAP (display->map[display->current])->pixmap_mgf, 
         0, 0, 0, 0, -1,-1) ;
}

void 
blink_timeout_remove (gpointer data)
{
   cxc_display *display = (cxc_display *)data ;

   gtk_timeout_remove (display->timer) ;
   display->timer = 0 ; 
   display->current = display->nframes ;
   frame_select (display, display->old_current) ;
} 

void
frame_select (cxc_display *display, guint new_selection)
{
   gint  izoom ;
   guint old_selection = display->current ;

   CPKDEBUG ("frame_select") ;

   if (old_selection != display->nframes && 
       old_selection != new_selection)
      display->old_current = old_selection ;

   /* Draw focus border */
   if (old_selection != display->nframes) {
      if (CXC_MAP (display->map[old_selection])->selected) {
         CXC_MAP (display->map[old_selection])->selected = FALSE ;
         gtk_widget_draw (GTK_WIDGET (
               CXC_MAP (display->map[old_selection])->draw_area), NULL) ;
      }
   }
   if (display->mode == FRAME_TILE || display->mode == FRAME_PAIR) {
      CXC_MAP (display->map[new_selection])->selected = TRUE ;
      gtk_widget_draw (GTK_WIDGET (
            CXC_MAP (display->map[new_selection])->draw_area), NULL) ;
   }
   if (display->current == new_selection)
      return ;

   display->current = new_selection ;

   /* Set current frame zoom */
   izoom = CXC_MAP (display->map[new_selection])->izoom ;
   if (izoom >= 0 && izoom < 9)
      if (display->zoom_radio1[izoom])
         gtk_check_menu_item_set_active (
             GTK_CHECK_MENU_ITEM (display->zoom_radio1[izoom]), TRUE) ;

   cxc_buttons_swap (display) ;
   cxc_buttons_wcs_status (display) ;

   set_paner (display) ;

   /* Set current values for Contrast and Luminosity control widgets */
   /* gtk_signal_disconnect_by_func (
         GTK_OBJECT (CPK_BIBUTTON (display->contrast)->adjustment), 
         GTK_SIGNAL_FUNC (cxc_toolbar_set_contrast),
          (gpointer)display) ;
   gtk_signal_disconnect_by_func (
         GTK_OBJECT (CPK_BIBUTTON (display->luminosity)->adjustment), 
         GTK_SIGNAL_FUNC (cxc_toolbar_set_luminosity),
          (gpointer)display) ;
 */
   /* gtk_signal_emit_stop_by_name (
         GTK_OBJECT (CPK_BIBUTTON (display->contrast)->adjustment),
         "value-changed") ;
   gtk_signal_emit_stop_by_name (
         GTK_OBJECT (CPK_BIBUTTON (display->luminosity)->adjustment),
         "value-changed") ; */

   gtk_adjustment_set_value (CPK_BIBUTTON (display->contrast)->adjustment, 
         1 / CXC_MAP (display->map[new_selection])->mrange) ;
   gtk_adjustment_set_value (CPK_BIBUTTON (display->luminosity)->adjustment, 
         CXC_MAP (display->map[new_selection])->radius) ;

   /* gtk_signal_connect_object (
         GTK_OBJECT (CPK_BIBUTTON (display->contrast)->adjustment), 
         "value-changed", GTK_SIGNAL_FUNC (cxc_toolbar_set_contrast),
          (gpointer)display) ;
   gtk_signal_connect_object (
         GTK_OBJECT (CPK_BIBUTTON (display->luminosity)->adjustment), 
         "value-changed", GTK_SIGNAL_FUNC (cxc_toolbar_set_luminosity),
          (gpointer)display) ; */
}

