/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include <math.h>
#include "common.h"

#include <gtk/gtk.h>

#include "catalib.h"
#include "cxc_data.h"
#include "cxc_map.h"
#include "cxc_magnifier.h"

#include "cxc_gtk_main.h"

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef MAX
#define MAX(a,b) (((a) >= (b)) ? (a) : (b))
#endif
#define MGF_DELAY    (150)

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static gint        cxc_mgf_configure        (GtkWidget         *widget, 
                                             GdkEventConfigure *event, 
                                             cxc_display       *display) ;
static gint        cxc_mgf_expose           (GtkWidget         *widget, 
                                             GdkEventExpose    *event, 
                                             cxc_display       *display) ;
static void        cxc_mgf_draw             (CxcMap            *map,
                                             cxc_display       *display) ;
static gint        cxc_mgf_remap_all        (gpointer           data) ;

static void        mgf_world_to_draw        (CxcMap            *map,
                                             gdouble            x, 
                                             gdouble            y, 
                                             gdouble           *xd, 
                                             gdouble           *yd) ;

/*----------------------------------------------------------------------------
 * 							Magnifier
 *--------------------------------------------------------------------------*/

GtkWidget *
cxc_magnifier_new (GtkWidget *window, cxc_display *display)
{
   GtkWidget *box, *frame ;

   box = gtk_hbox_new (FALSE, 0) ;
   frame = gtk_frame_new (NULL) ;
   gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN) ;
   gtk_widget_set_usize (frame, 120, 120) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 0) ;
   gtk_container_add (GTK_CONTAINER (box), frame) ;

   display->magnifier = gtk_drawing_area_new () ;
   gtk_signal_connect (GTK_OBJECT (display->magnifier),"configure_event",
         (GtkSignalFunc)cxc_mgf_configure, display) ;
   gtk_signal_connect (GTK_OBJECT (display->magnifier), "expose_event",
         (GtkSignalFunc)cxc_mgf_expose, display) ;

   gtk_container_add (GTK_CONTAINER (frame), display->magnifier) ;   

   gtk_widget_show_all (box) ;
   
   return box ;
}

static gint
cxc_mgf_configure (GtkWidget *widget, GdkEventConfigure *event, cxc_display *display)
{
   CxcMap *map = CXC_MAP (display->map[display->focused]) ;
   /* Allocate pixmap */
   if (map->pixmap_mgf) 
      gdk_pixmap_unref (map->pixmap_mgf) ;

   map->pixmap_mgf = gdk_pixmap_new (widget->window,
         widget->allocation.width, widget->allocation.height, -1) ;
   /* Set pixmap backgound to white */
   gdk_draw_rectangle (map->pixmap_mgf,
         display->magnifier->style->white_gc, TRUE, 0, 0, 
         display->magnifier->allocation.width, 
         display->magnifier->allocation.height) ;
   return TRUE;
}  

static gint
cxc_mgf_expose (GtkWidget *widget, GdkEventExpose *event, cxc_display *display)
{
   CxcMap *map = CXC_MAP (display->map[display->focused]) ;
   gdk_draw_pixmap (widget->window,
         widget->style->fg_gc[GTK_WIDGET_STATE (widget)], map->pixmap_mgf,
         event->area.x, event->area.y, event->area.x, event->area.y, 
         event->area.width, event->area.height) ;
  return FALSE;
}

gint
cxc_mgf_remap (GtkWidget *widget, cxc_display *display)
{
   CxcMap *map = CXC_MAP (display->map[display->focused]) ;

   if (display->mode == FRAME_BLINK) {
      gint  x, y ;
      map = CXC_MAP (display->map[display->current]) ;
      gdk_window_get_pointer (CXC_MAP (display->map[display->nframes])->draw_area->window, &x, &y, NULL) ;
      draw_to_world (map, (gdouble)x, (gdouble)y, &map->xcur, &map->ycur) ;
   }
   else
      if (!GTK_WIDGET_MAPPED (map->draw_area) ||
          !GTK_WIDGET_VISIBLE (map->draw_area->parent))
         return TRUE ;

   if (!map->pixmap_mgf)
      map->pixmap_mgf = gdk_pixmap_new (display->magnifier->window,
         display->magnifier->allocation.width, 
         display->magnifier->allocation.height, -1) ;

   /* Set pixmap backgound to white */
   gdk_draw_rectangle (map->pixmap_mgf, 
         display->magnifier->style->white_gc, TRUE, 0, 0,
         display->magnifier->allocation.width, 
         display->magnifier->allocation.height) ;

   /* draw map */
   cxc_mgf_draw (map, display) ;
   gdk_draw_pixmap (display->magnifier->window,
         display->magnifier->style->fg_gc[GTK_WIDGET_STATE (display->magnifier)],
         map->pixmap_mgf, 0, 0, 0, 0, -1,-1) ;

   if (display->mode != FRAME_BLINK) 
      cxc_mgf_id_draw (map, display) ;

   return TRUE ;
}

static gint
cxc_mgf_remap_all (gpointer data)
{
   cxc_display *display = (cxc_display *)data ;
   gint i ;

   GDK_THREADS_ENTER () ;
   if (display->mgf_timer)
      display->mgf_timer = 0 ;
   for (i = 0; i < display->nframes; i++)
      if (display->blink[i] && CXC_MAP (display->map[i])->pixmap_mgf) {
         gdk_pixmap_unref (CXC_MAP (display->map[i])->pixmap_mgf) ;
         CXC_MAP (display->map[i])->pixmap_mgf = NULL ;
      }
   
   GDK_THREADS_LEAVE () ;

   return FALSE ;
}

gint
cxc_mgf_enter (GtkWidget *widget, GdkEventCrossing *event, cxc_display *display)
{
   CPKDEBUG5 ("cxc_mgf_enter", 7);
   return TRUE ;
   cxc_mgf_remap (NULL, display) ;
   return TRUE ;
}


gint
cxc_mgf_update (GtkWidget *widget, GdkEventMotion *event, cxc_display *display)
{
   GdkModifierType state;

   if (display->mode == FRAME_BLINK) {
      if (display->mgf_timer)
         gtk_timeout_remove (display->mgf_timer) ;
      display->mgf_timer = gtk_timeout_add (MGF_DELAY, cxc_mgf_remap_all, display) ;
      return FALSE ;
   }

   if (event->is_hint) 
      gdk_window_get_pointer (event->window, NULL, NULL, &state);
   else 
      state = event->state;

   if (state & GDK_BUTTON1_MASK) 
      return TRUE ;

   cxc_mgf_remap (NULL, display) ;
   return TRUE ;
}

static void
cxc_mgf_draw (CxcMap *map, cxc_display *display)
{
   GdkPoint   points[10] ;
   GdkGC    **gc = NULL ;
   guint      i, j ;
   gint       lx, ly ;
   gdouble    x, y, r, mag, rmx ;
   gfloat     f ;
   Stack     *tab ;
   gdouble    scale ;

   /* if (!GTK_WIDGET_MAPPED (map->draw_area))
      return ; */

   if (map->color && SHOW_COLOR) 
      if (!map->stack_gc || map->ngc != map->fnorm)
         return ;

   scale  = (map->list->out_sph) ? 3600.0/0.25 : 1.0 ;
   scale *= map->zoom ;

   lx = display->magnifier->allocation.width / 2 ;
   ly = display->magnifier->allocation.height / 2 ;
   cpk_poly_plus (points, lx, ly, 0, 10) ;
   gdk_draw_polygon (map->pixmap_mgf, map->gc[MAP_GRAY], FALSE, points, 9) ;

   gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 2, GDK_LINE_SOLID,
         GDK_CAP_PROJECTING, GDK_JOIN_ROUND) ;

   rmx = map->radius * map->zoom * display->mgf_factor ;
   for (i = 0; i < (guint)map->list->nstar; i++) {

      if (ABS (map->list->data[i]->xrot-map->xcur) < lx/scale && 
          ABS (map->list->data[i]->yrot-map->ycur) < ly/scale ) {

         mgf_world_to_draw (map, map->list->data[i]->xrot, map->list->data[i]->yrot, 
               &x, &y) ;
         x = x * display->mgf_factor + lx ;
         y = y * display->mgf_factor + ly ;

         if (x > -rmx && x < display->magnifier->allocation.width + rmx &&
             y > -rmx && y < display->magnifier->allocation.height + rmx) {
            mag = map->list->data[i]->mcal ;
            mag = (mag > map->mcut2) ? map->mcut2 : 
                  (mag < map->mcut1) ? map->mcut1 : mag ;
            r = rmx * (map->mcut2 - mag) / map->mrange + 0.5 ;
            if (r >= 0) {
               if (map->color && SHOW_COLOR) {
                  tab = (Stack*)map->list->data[i]->p ;
                  j = bin_get_count (tab->ns) + bin_get_count (tab->ns2) - 1 ;
                  draw_circle (map->pixmap_mgf, map->stack_gc[j], x, y, r, 25) ;
               }
               else {
                  if (map->list->data[i]->flag == 0) 
                     draw_circle (map->pixmap_mgf, map->gc[MAP_NORMAL], 
                           x, y, r, 25) ;
                  else
                     draw_circle (map->pixmap_mgf, map->gc[MAP_FLAGGED], 
                           x, y, r, 25) ;
               }
            }
         }
      }
   }
}

void
cxc_mgf_id_draw (CxcMap *map, cxc_display *display)
{
   gint       lx, ly ;
   gdouble    x, y, r, mag, rmx ;

   CPKDEBUG5 ("cxc_mgf_id_draw", 8);

   if (!map->s)
      return ;
   if (map->color && SHOW_COLOR) 
      if (!map->stack_gc || map->ngc != map->fnorm)
         return ;

   lx = display->magnifier->allocation.width / 2 ;
   ly = display->magnifier->allocation.height / 2 ;

   gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 2, GDK_LINE_SOLID,
         GDK_CAP_PROJECTING, GDK_JOIN_ROUND) ;

   mgf_world_to_draw (map, map->s->xrot, map->s->yrot, &x, &y) ;
   x = x * display->mgf_factor + lx ;
   y = y * display->mgf_factor + ly ;

   rmx = map->radius * map->zoom * display->mgf_factor ;
   mag = map->s->mcal ;
   mag = (mag > map->mcut2) ? map->mcut2 : 
         (mag < map->mcut1) ? map->mcut1 : mag ;
   r = rmx * (map->mcut2 - mag) / map->mrange + 0.5 ;

   if (r < 0)
      return ;

   if (map->color && SHOW_COLOR) {
      Stack  *tab = (Stack*) map->s->p ;
      guint   j = bin_get_count (tab->ns) + bin_get_count (tab->ns2) - 1 ;
      gdk_gc_set_line_attributes (map->stack_gc[j], 2, GDK_LINE_SOLID,
            GDK_CAP_PROJECTING, GDK_JOIN_ROUND) ;
      draw_circle (display->magnifier->window, map->stack_gc[j], x, y, r, 25) ;
      gdk_gc_set_line_attributes (map->stack_gc[j], 1, GDK_LINE_SOLID,
            GDK_CAP_PROJECTING, GDK_JOIN_ROUND) ;
   }
   else {
      if (map->s->flag == 0) 
         cpk_set_fg_color (map->draw_area, map->gc[MAP_HIGHLIGHT], "black") ;
      else
         cpk_set_fg_color (map->draw_area, map->gc[MAP_HIGHLIGHT], "gray75") ;
      draw_circle (display->magnifier->window, map->gc[MAP_HIGHLIGHT], x, y, r, 25) ;
      cpk_set_fg_color (map->draw_area, map->gc[MAP_HIGHLIGHT], "blue") ;
   }
}

/* Transform world coordinates to drawing ones on magnifier area */
static void
mgf_world_to_draw (CxcMap *map, gdouble x, gdouble y, gdouble *xd, gdouble *yd)
{
   gdouble dx, dy ;

   dx = x - map->xcur ;
   dy = y - map->ycur ;

   *xd = ( map->cos_rot*dx - map->sin_rot*dy) * map->map_scale * map->swx *
           map->swxx ;
   *yd = (-map->sin_rot*dx - map->cos_rot*dy) * map->map_scale * map->swy ;
}

