/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include <math.h>
#include "common.h"

#include <gtk/gtk.h>

#include "cpk_gtk_util.h"
#include "cataxcorr.h"
#include "catalib.h"
#include "platemodel.h"
#include "cataxcorr_p.h"
#include "coordinates.h"
#include "starlist.h"
#include "model.h"


#include "cxc_data.h"
#include "cxc_map.h"

#include "cxc_gtk_main.h"


/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef MIN
#define MIN(a,b) (((a) <= (b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b) (((a) >= (b)) ? (a) : (b))
#endif
#ifndef M_PI
#define M_PI    (3.14159265358979323846)
#endif   /* M_PI */
#define TWOPI (2.0*M_PI)
#ifndef RADTODEG
#define RADTODEG(a) ((a)*360.0/TWOPI)
#endif
#ifndef DEGTORAD
#define DEGTORAD(a) ((a)*TWOPI/360.0)
#endif

#define HYPOT(a,b) (sqrt ((a)*(a)+(b)*(b)))

#define MAP_DELAY    (250)
#define MAP_DELAY_ID (50)
#define MAP_NBRIGHT  (250)
#define MAP_NPLOT    (2000)

#define REMAP              (map->nx == 0)
#define RESCALING          (map->dx < 1.0e-15)

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern double     IntNull ;

extern boolean    suspend ;
extern GtkWidget *statusbar ;

/* Local data */

static GtkVBoxClass *parent_class = NULL;
static guint         map_signals[MAP_LAST_SIGNAL] = { 0 };

static gdouble srad_deg, srad_pix ;

gdouble sph_scale = 3600.0/0.25 ;
gdouble lin_scale = 1.0 ;

/* static const gchar  *value_index_key = "gtk-value-index"; */

/*----------------------------------------------------------------------------
 * 							Private functions
 *--------------------------------------------------------------------------*/

static int sort_star_m (const Star **d1, const Star **d2)
{ 
   if ( (*d1)->mcal > (*d2)->mcal ) 
      return 1 ; 
   else 
      return -1 ; 
}

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/
                                               
static void        cxc_map_class_init           (CxcMapClass       *klass) ;
 
static void        cxc_map_init                 (CxcMap            *map) ;
 
static void        cxc_map_get_range            (CxcMap            *map) ;
 
static void        cxc_map_get_mag_range        (CxcMap            *map) ;
 
static void        cxc_map_set_cuts             (CxcMap            *map) ;

static GdkGC     **cxc_map_set_stack_gc         (CxcMap            *map) ;

static void        cxc_map_free_stack_gc        (CxcMap            *map) ;

static void        cxc_map_check_fnorm          (CxcMap            *map) ;

static Star       *cxc_map_find_nearest         (CxcMap            *map,
                                                 gdouble            x,
                                                 gdouble            y) ;
static gint        cxc_map_map_event            (GtkWidget         *widget,
                                                 GdkEvent          *event,
                                                 CxcMap            *map) ;
static gint        cxc_map_expose_event         (GtkWidget         *widget,
                                                 GdkEventExpose    *event,
                                                 CxcMap            *map) ;
static gint        cxc_map_configure_event      (GtkWidget         *widget,
                                                 GdkEventConfigure *event,
                                                 CxcMap            *map) ;
static void        cxc_map_realize              (GtkWidget         *widget,
                                                 CxcMap            *map) ;
 
static void        cxc_map_draw_focus           (GtkWidget         *widget) ;
static gint        cxc_map_focus_in             (GtkWidget         *widget,
		                                 GdkEventFocus     *event) ;
static gint        cxc_map_focus_out            (GtkWidget         *widget,
		                                 GdkEventFocus     *event) ;
static gint        cxc_map_button_press_event   (GtkWidget         *widget,
                                                 GdkEventButton    *event,
                                                 CxcMap            *map) ;
static gint        cxc_map_button_release       (GtkWidget         *widget,
                                                 GdkEventButton    *event,
                                                 CxcMap            *map) ;
static gint        cxc_map_enter_notify_event   (GtkWidget         *widget,
                                                 GdkEventCrossing  *event,
                                                 CxcMap            *map) ;
static gint        cxc_map_motion_notify_event  (GtkWidget         *widget,
                                                 GdkEventMotion    *event,
                                                 CxcMap            *map) ;
static gint        cxc_map_leave_notify_event   (GtkWidget         *widget,
                                                 GdkEventCrossing  *event,
                                                 CxcMap            *map) ;
static gint        cxc_map_timer                (gpointer           data) ;
 
static gint        cxc_map_timer_id             (gpointer           data) ;
 
static gint        cxc_map_geometry             (GtkWidget         *widget,
                                                 CxcMap            *map,
                                                 gdouble            x,
                                                 gdouble            y) ;
static void        cxc_map_draw_center          (GtkWidget         *widget,
                                                 CxcMap            *map) ;
static void        cxc_map_draw                 (GdkPixmap         *pixmap,
                                                 CxcMap            *map,
                                                 gint               mode) ;
static gint        cxc_pan_configure_event      (GtkWidget         *widget,
                                                 GdkEventConfigure *event,
                                                 CxcMap            *map) ;
extern gint        cxc_pan_remap                (GtkWidget         *widget,
                                                 CxcMap            *map) ;
static gint        cxc_pan_map_event            (GtkWidget         *widget,
                                                 GdkEvent          *event,
                                                 CxcMap            *map) ;
static gint        cxc_pan_expose_event         (GtkWidget         *widget,
                                                 GdkEventExpose    *event,
                                                 CxcMap            *map) ;
static void        cxc_pan_draw                 (CxcMap            *map) ;
 
static void        cxc_pan_draw_border          (CxcMap            *map) ;
 
static gint        cxc_pan_button_press_event   (GtkWidget         *widget,
                                                 GdkEventButton    *event,
                                                 CxcMap            *map) ;
static gint        cxc_pan_button_release       (GtkWidget         *widget,
                                                 GdkEventButton    *event,
                                                 CxcMap            *map) ;
static gint        cxc_pan_motion_notify_event  (GtkWidget         *widget,
                                                 GdkEventMotion    *event,
                                                 CxcMap            *map) ;
static void        pan_world_to_draw            (CxcMap            *map,
                                                 gdouble            x,
                                                 gdouble            y,
                                                 gdouble           *xd,
                                                 gdouble           *yd) ;
static void        pan_draw_to_world            (CxcMap            *map,
                                                 gdouble            xd,
                                                 gdouble            yd,
                                                 gdouble           *x,
                                                 gdouble           *y) ;
static void        cxc_map_set_const            (CxcMap            *map) ;
 
static void        cxc_map_set_pan_scale        (CxcMap            *map) ;

/*----------------------------------------------------------------------------
 * 							Defines
 *--------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

GtkType
cxc_map_get_type (void)
{
  static guint map_type = 0;

  if (!map_type)
    {
      static const GtkTypeInfo map_info =
      {
	"CxcMap",
	sizeof (CxcMap),
	sizeof (CxcMapClass),
	(GtkClassInitFunc) cxc_map_class_init,
	(GtkObjectInitFunc) cxc_map_init,
	/* reserved_1 */ NULL,
	/* reserved_2 */ NULL,
        (GtkClassInitFunc) NULL,
      };

      map_type = gtk_type_unique (gtk_vbox_get_type (), &map_info);
    }

  return map_type;
}

static void
cxc_map_class_init (CxcMapClass *class)
{
  GtkObjectClass    *object_class;
  GtkWidgetClass    *widget_class;
  GtkContainerClass *container_class;

  object_class = (GtkObjectClass*) class;
  widget_class = (GtkWidgetClass*) class;
  container_class = (GtkContainerClass*) class;

  parent_class = gtk_type_class (gtk_widget_get_type ()) ;

  map_signals[MAP_RANGE] =
        gtk_signal_new ("map-range",
        GTK_RUN_FIRST,
        object_class->type,
        GTK_SIGNAL_OFFSET (CxcMapClass, map_range),
        gtk_marshal_NONE__NONE,
	GTK_TYPE_NONE, 0) ;
  map_signals[MAP_MAGNITUDE] =
        gtk_signal_new ("map-magnitude",
        GTK_RUN_FIRST,
        object_class->type,
        GTK_SIGNAL_OFFSET (CxcMapClass, map_magnitude),
        gtk_marshal_NONE__NONE,
	GTK_TYPE_NONE, 0) ;
  map_signals[MAP_CHANGED] =
        gtk_signal_new ("map-changed",
        GTK_RUN_FIRST,
        object_class->type,
        GTK_SIGNAL_OFFSET (CxcMapClass, map_changed),
        gtk_marshal_NONE__NONE,
	GTK_TYPE_NONE, 0) ;
  map_signals[MAP_REMAPPED] =
        gtk_signal_new ("map-remapped",
        GTK_RUN_FIRST,
        object_class->type,
        GTK_SIGNAL_OFFSET (CxcMapClass, map_remapped),
        gtk_marshal_NONE__NONE,
	GTK_TYPE_NONE, 0) ;
  map_signals[MAP_UNLOCK] =
        gtk_signal_new ("map-unlock",
        GTK_RUN_FIRST,
        object_class->type,
        GTK_SIGNAL_OFFSET (CxcMapClass, map_unlock),
        gtk_marshal_NONE__NONE,
	GTK_TYPE_NONE, 0) ;
  map_signals[STAR_CHANGED] =
        gtk_signal_new ("star-changed",
        GTK_RUN_FIRST,
        object_class->type,
        GTK_SIGNAL_OFFSET (CxcMapClass, star_changed),
        gtk_marshal_NONE__NONE,
	GTK_TYPE_NONE, 0) ;

  gtk_object_class_add_signals (object_class, map_signals, MAP_LAST_SIGNAL);

}

static void
cxc_map_init (CxcMap *map)
{
   map->zoom = 1.0 ;
   map->izoom = 4 ;
   map->rot  = 0.0 ;
   map->swx = map->swy = 1 ;
   map->panx = map->pany = 0.0 ;
   map->dx = map->dy = 0.0 ;
   map->radius = 18.0 ;
   map->gc[0] = NULL ;
   map->stack_gc = NULL ;
   map->ngc = 0 ;
   map->list = NULL ;
   map->model = NULL ;
   map->map_id = 0 ;
   map->color = FALSE ;
   map->fnorm = 1 ;
   map->selected = FALSE ;
   map->pointer_action = NULL ;
   map->reference = NULL ;
   map->s = NULL ;
   map->grabbed = map->paner_drag = map->paner_active = FALSE ;
   map->timer = 0 ;
   map->timer_id = 0 ;

   /* gtk_widget_set_visual (GTK_WIDGET (map), gdk_rgb_get_visual ()) ;
   gtk_widget_set_colormap (GTK_WIDGET (map), gdk_rgb_get_cmap ()) ; */

   map->draw_area = gtk_drawing_area_new () ;

   /* gtk_widget_set_visual (GTK_WIDGET (map->draw_area), gdk_rgb_get_visual ()) ;
   gtk_widget_set_colormap (GTK_WIDGET (map->draw_area), gdk_rgb_get_cmap ()) ; */

   GTK_WIDGET_SET_FLAGS (map->draw_area, GTK_CAN_FOCUS) ;

   gtk_signal_connect (GTK_OBJECT (map), "map-range",
                      (GtkSignalFunc)cxc_map_get_range, map) ;
   gtk_signal_connect (GTK_OBJECT (map), "map-magnitude",
                       (GtkSignalFunc)cxc_map_get_mag_range, map) ;
   gtk_signal_connect (GTK_OBJECT (map), "map-changed",
                       (GtkSignalFunc)cxc_map_remap, map) ;

   gtk_signal_connect (GTK_OBJECT (map->draw_area), "map_event",
                       (GtkSignalFunc)cxc_map_map_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area), "expose_event",
                       (GtkSignalFunc)cxc_map_expose_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area),"configure_event",
                       (GtkSignalFunc)cxc_map_configure_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area),"focus_in_event",
                       (GtkSignalFunc)cxc_map_focus_in, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area),"focus_out_event",
                       (GtkSignalFunc)cxc_map_focus_out, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area), "button_press_event",
                       (GtkSignalFunc)cxc_map_button_press_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area), "button_release_event",
                       (GtkSignalFunc)cxc_map_button_release, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area), "enter_notify_event",
                       (GtkSignalFunc)cxc_map_enter_notify_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area), "motion_notify_event",
                       (GtkSignalFunc)cxc_map_motion_notify_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->draw_area), "leave_notify_event",
                       (GtkSignalFunc)cxc_map_leave_notify_event, map) ;
   gtk_widget_set_events (map->draw_area, GDK_EXPOSURE_MASK
                                        | GDK_ENTER_NOTIFY_MASK
                                        | GDK_LEAVE_NOTIFY_MASK
                                        | GDK_BUTTON_PRESS_MASK
                                        | GDK_BUTTON_RELEASE_MASK
                                        | GDK_POINTER_MOTION_MASK
                                        | GDK_POINTER_MOTION_HINT_MASK) ;


   gtk_widget_show_all (map->draw_area) ;

   map->paner = gtk_drawing_area_new () ;
   GTK_WIDGET_SET_FLAGS (map->paner, GTK_CAN_FOCUS) ;
   gtk_signal_connect (GTK_OBJECT (map->paner),"configure_event",
                       (GtkSignalFunc)cxc_pan_configure_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->paner), "map_event",
                       (GtkSignalFunc)cxc_pan_map_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->paner), "expose_event",
                       (GtkSignalFunc)cxc_pan_expose_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->paner), "button_press_event",
                       (GtkSignalFunc)cxc_pan_button_press_event, map) ;
   gtk_signal_connect (GTK_OBJECT (map->paner), "button_release_event",
                       (GtkSignalFunc)cxc_pan_button_release, map) ;
   gtk_signal_connect (GTK_OBJECT (map->paner), "motion_notify_event",
                       (GtkSignalFunc)cxc_pan_motion_notify_event, map) ;
   gtk_widget_set_events (map->paner, GDK_EXPOSURE_MASK
                                    | GDK_BUTTON_PRESS_MASK
                                    | GDK_BUTTON_RELEASE_MASK
                                    | GDK_POINTER_MOTION_MASK
                                    | GDK_POINTER_MOTION_HINT_MASK) ;
   gtk_widget_show_all (map->paner) ;

   return ;
}

GtkWidget *
cxc_map_new (CxcData *data, guint n)
{
   CxcMap  *map ;

   map = gtk_type_new (cxc_map_get_type ());

   srad_deg = data->srad_deg ;
   srad_pix = data->srad_pix ;
   if (n >= 0 && n < data->n) {
      gtk_signal_connect_object (GTK_OBJECT (data), "data-changed",
                         (GtkSignalFunc)cxc_map_get_range, GTK_OBJECT (map)) ;
      map->list = data->list[n] ;
      map->model = data->model[n] ;
   }
   if (n == data->n) {
      gtk_signal_connect_object (GTK_OBJECT (data), "stack-added",
                         (GtkSignalFunc)cxc_map_get_mag_range, GTK_OBJECT (map)) ;
      gtk_signal_connect_object (GTK_OBJECT (data), "stack-added",
                         (GtkSignalFunc)cxc_map_get_range, GTK_OBJECT (map)) ;
      gtk_signal_connect (GTK_OBJECT (data), "stack-added",
                         (GtkSignalFunc)cxc_map_remap, GTK_OBJECT (map)) ;

      map->list = data->stack->ref ;
      map->model = data->model[data->nref] ;
      map->color = TRUE ;
   }
   if (n == 0)
      map->color = TRUE ;
   map->map_id = n ;
   cxc_map_get_mag_range (map) ;
   cxc_map_get_range (map) ;
   map->panx = ( map->fx1 + map->fx2 ) / 2.0 ;
   map->pany = ( map->fy1 + map->fy2 ) / 2.0 ;

   qsort ((void*)map->list->data, map->list->nstar, sizeof(Star*),
          (int (*)(const void *, const void *))sort_star_m) ;

   return GTK_WIDGET (map) ;
}

void
cxc_map_destroy (CxcMap *map)
{
   if (map->pixmap_pan) 
      gdk_pixmap_unref (map->pixmap_pan) ;
   if (map->pixmap_mgf) 
      gdk_pixmap_unref (map->pixmap_mgf) ;
   if (!map->gc[0]) {
      guint i ;
      for (i = 0; i < 6; i++)
         gdk_gc_unref (map->gc[i]) ;
   }
   if (!map->stack_gc) {
      guint i ;
      for (i = 0; i < map->fnorm; i++)
         gdk_gc_unref (map->stack_gc[i]) ;
      free (map->stack_gc) ;
   }

   gtk_widget_destroy (map->paner) ;
   gtk_widget_destroy (map->draw_area) ;

   /* gtk_widget_destroy (GTK_WIDGET (map)) ; */

}

static void
cxc_map_get_range (CxcMap *map)
{
   size_t  ns = map->list->nstar, i ;
   double  fx1, fx2, fy1, fy2 ;

   CPKDEBUG5 ("cxc_map_get_range",4) ;

   if (ns == 0)
      return ;

   fx1 = fx2 = map->list->data[0]->xrot ;
   fy1 = fy2 = map->list->data[0]->yrot ;
   for (i = 0; i < ns; i++) {
      fx1 = MIN (fx1, map->list->data[i]->xrot) ;
      fx2 = MAX (fx2, map->list->data[i]->xrot) ;
      fy1 = MIN (fy1, map->list->data[i]->yrot) ;
      fy2 = MAX (fy2, map->list->data[i]->yrot) ;
   }

   cxc_map_set_const (map) ;

   map->radius = (map->list->out_sph ? srad_deg : srad_pix ) * 
         map->map_scale / map->zoom ;

   map->radius = MIN (map->radius, 30) ;
   map->radius = MAX (map->radius, 5) ;

   cxc_map_set_cuts (map) ;

   if (fx1 != map->fx1 || fx2 != map->fx2 || 
       fy1 != map->fy1 || fy2 != map->fy2) {
      map->fx1 = fx1 ;
      map->fx2 = fx2 ;
      map->fy1 = fy1 ;
      map->fy2 = fy2 ;
      /* gtk_signal_emit_by_name (GTK_OBJECT (map), "map-changed") ; */
   }
}

static void
cxc_map_set_cuts (CxcMap *map)
{
   CPKDEBUG5 ("cxc_map_set_cuts",5) ;

   map->mcut1 += map->model->dmag-map->old_mzero ;
   map->mcut2 += map->model->dmag-map->old_mzero ;
   map->old_mzero = map->model->dmag ;
}

void
cxc_map_range_set (CxcMap *map, gfloat val)
{
   gdouble mag ;
   CPKDEBUG5 ("cxc_map_range_set",6) ;
   if (map->mrange == (gdouble)val)
      return ;
   map->mrange =  (gdouble)val ;
   mag = (map->mcut1 + map->mcut2) / 2.0 ;
   map->mcut1 = mag - map->mrange / 2.0 ;
   map->mcut2 = mag + map->mrange / 2.0 ;
   if (map->timer)
      gtk_timeout_remove (map->timer) ;
   map->timer = 1 ;
   cxc_map_remap (NULL, map) ;
   map->timer = gtk_timeout_add (MAP_DELAY, cxc_map_timer, map) ;
}

void
cxc_map_radius_set (CxcMap *map, gfloat val)
{
   gdouble mag ;
   CPKDEBUG5 ("cxc_map_radius_set",6) ;
   if (map->radius == (gdouble)val)
      return ;
   map->radius = (gdouble)val ;
   if (map->timer)
      gtk_timeout_remove (map->timer) ;
   map->timer = 1 ;
   cxc_map_remap (NULL, map) ;
   map->timer = gtk_timeout_add (MAP_DELAY, cxc_map_timer, map) ;
}

static void
cxc_map_get_mag_range (CxcMap *map)
{
   size_t  ns = map->list->nstar ;
   guint   i ;

   CPKDEBUG5 ("cxc_map_get_mag_range",5) ;

   if (ns == 0)
      return ;
   
   qsort ((void*)map->list->data, map->list->nstar, sizeof(Star*),
          (int (*)(const void *, const void *))sort_star_m) ;

   map->fm1 = map->list->data[0]->mcal ; /*** FIXME: USELESS ***/
   map->nbright = MIN ((guint)ns-1, MAP_NBRIGHT) ;

   i = ns-1 ;
   while (i > 0 && map->list->data[i]->mcal == IntNull)
      i-- ;
   map->fm2 = map->list->data[i]->mcal ; /*** FIXME: USELESS ***/
   map->mrange = 5.0 ;
   map->mcut2 = map->list->data[(int)(0.5*(i))]->mcal ;
   map->mcut1 = map->mcut2 - map->mrange ;
   map->mcut1 = MAX (map->mcut1, map->fm1) ;
   map->mcut2 = map->mcut1 + map->mrange ;
   map->old_mzero = map->model->dmag ;
}

void
cxc_map_reset (CxcMap *map)
{
   cxc_map_get_range (map) ;
   cxc_map_get_mag_range (map) ;
   map->panx = ( map->fx1 + map->fx2 ) / 2.0 ;
   map->pany = ( map->fy1 + map->fy2 ) / 2.0 ;
   map->swx = map->swy = 1 ;
   map->zoom = 1.0 ;
   map->izoom = 4 ;
   map->rot  = 0.0 ;
   cxc_map_remap (NULL, map) ;
}

/* Redraw the screen from the backing pixmap */
static gint
cxc_map_map_event (GtkWidget *widget, GdkEvent *event, CxcMap *map)
{
   CPKDEBUG5 ("cxc_map_map_event",5) ;
   return FALSE;
}
static void
cxc_map_realize (GtkWidget *widget, CxcMap *map)
{
   CPKDEBUG5 ("cxc_map_realize",5) ;
   return ;
}

static gint
cxc_map_expose_event (GtkWidget *widget, GdkEventExpose *event, CxcMap *map)
{
   CPKDEBUG5 ("cxc_map_expose_event",7);
   gdk_draw_pixmap (widget->window,
                    widget->style->fg_gc[GTK_WIDGET_STATE (widget)],
                    map->pixmap,
                    event->area.x, event->area.y,
                    event->area.x, event->area.y,
                    event->area.width, event->area.height) ;
   if (map->selected && *map->mode != FRAME_SINGLE) {
      gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 3, 
            GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER) ;
      gdk_draw_rectangle (widget->window,
            map->gc[MAP_HIGHLIGHT], FALSE, 0, 0,
            widget->allocation.width-1, widget->allocation.height-1) ;
   }
   CPKDEBUG5 ("cxc_map_expose_event_end",7);
   return FALSE;
}

static gint
cxc_map_configure_event (GtkWidget *widget, GdkEventConfigure *event, CxcMap *map)
{
   CPKDEBUG5 ("cxc_map_configure_event",5);

   return cxc_map_configure_pixmap (widget, map) ;
}  

  /* Create a new backing pixmap of the appropriate size */
gint
cxc_map_configure_pixmap (GtkWidget *widget, CxcMap *map)
{
   if (map->pixmap) 
      if (map->nx == widget->allocation.width &&
          map->ny == widget->allocation.height)
         return TRUE ;

   CPKDEBUG5 ("configure_pixmap", 4) ;

   if (!map->gc[0]) {
       guint i ;
       /* CPKDEBUG ("cxc_map_gcinit"); */
       for (i = 0; i < 6; i++)
          map->gc[i] = gdk_gc_new (map->draw_area->window) ;
       cpk_set_fg_color (map->draw_area, map->gc[MAP_NORMAL], "black") ;
       cpk_set_fg_color (map->draw_area, map->gc[MAP_FLAGGED], "gray75") ;
       cpk_set_fg_color (map->draw_area, map->gc[MAP_HIGHLIGHT], "blue") ;
       cpk_set_fg_color (map->draw_area, map->gc[MAP_GRAY], "grey70") ;
       cpk_set_fg_color (map->draw_area, map->gc[MAP_SHADE], "brown") ;
       cpk_set_fg_color (map->draw_area, map->gc[MAP_DISPL], "SeaGreen3") ;
   }

   /* Allocate pixmap */
   if (map->pixmap) 
      gdk_pixmap_unref (map->pixmap) ;
   map->pixmap = gdk_pixmap_new (widget->window,
                                 widget->allocation.width,
                                 widget->allocation.height, -1) ;
   map->nx = widget->allocation.width ;
   map->ny = widget->allocation.height ;

   if (*map->mode == FRAME_MERGE) {
      map->reference->nx = map->nx ;
      map->reference->ny = map->ny ;
   }
   cxc_map_remap (widget, map) ;

   return TRUE;
}  

gint
cxc_map_remap (GtkWidget *widget, CxcMap *map)
{
   GtkWidget *draw = widget ;

   if (!map->pixmap)
      return TRUE ;

   if (!GTK_IS_DRAWING_AREA (draw)) 
      draw = map->draw_area ;

   if (!GTK_WIDGET_VISIBLE (draw->parent)) {
      CXC_MAP_REMAP_LATER (map) ;
      return FALSE ;
   }

   CPKDEBUG ("cxc_map_remap");

   /* Set pixmap backgound to white */
   gdk_draw_rectangle (map->pixmap,
         draw->style->white_gc, TRUE, 0, 0,
         draw->allocation.width, draw->allocation.height) ;

   /* draw map and coord grid */
   if (map->color && SHOW_COLOR)
      cxc_map_draw (map->pixmap, map, 2) ;
   else {
      if (*map->mode == FRAME_MERGE)
         cxc_map_draw (map->pixmap, map->reference, 1) ;
      cxc_map_draw (map->pixmap, map, 0) ;
   }

   /* Settare solo l'area esposta */
   gdk_draw_pixmap (map->draw_area->window,
         map->draw_area->style->fg_gc[GTK_WIDGET_STATE (map->draw_area)],
         map->pixmap, 0, 0, 0, 0, -1,-1) ;

   if (map->selected  && *map->mode != FRAME_SINGLE) {
      gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 3, 
            GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER) ;
      gdk_draw_rectangle (map->draw_area->window,
            map->gc[MAP_HIGHLIGHT], FALSE, 0, 0, 
            map->draw_area->allocation.width-1, 
            map->draw_area->allocation.height-1) ;
   }

   if (map->grabbed) 
      cxc_map_draw_center (draw, map) ;

   /* Remap associated paner */
   cxc_pan_remap (NULL, map) ;

   gtk_signal_emit_by_name (GTK_OBJECT (map), "map-remapped") ;
   return TRUE ;
}

static void
cxc_map_draw_focus (GtkWidget *widget)
{
   CPKDEBUG5 ("cxc_map_draw_focus",6);
   g_return_if_fail (widget != NULL);
   g_return_if_fail (GTK_IS_DRAWING_AREA (widget));

   gtk_widget_draw (GTK_WIDGET (widget), NULL);
}

static gint
cxc_map_focus_in (GtkWidget     *widget, GdkEventFocus *event)
{
   CPKDEBUG5 ("cxc_map_focus_in",6) ; 
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (GTK_IS_DRAWING_AREA (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   GTK_WIDGET_SET_FLAGS (widget, GTK_HAS_FOCUS) ;
   cxc_map_draw_focus (widget) ;

   return TRUE ;
}

static gint
cxc_map_focus_out (GtkWidget     *widget,
		   GdkEventFocus *event)
{
   CPKDEBUG5 ("cxc_map_focus_out",6) ;
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (GTK_IS_DRAWING_AREA (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   GTK_WIDGET_UNSET_FLAGS (widget, GTK_HAS_FOCUS);
   cxc_map_draw_focus (widget);

   return TRUE;
}

static gint
cxc_map_button_press_event (GtkWidget *widget, GdkEventButton *event, CxcMap *map)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (GTK_IS_DRAWING_AREA (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   if (map->list == NULL)
      return FALSE ;
   if (event->state & (GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) 
      return FALSE ;

   CPKDEBUG5 ("cxc_map_button_press_event",5);

   gtk_widget_grab_focus (widget) ;
   map->grabbed = TRUE ;

   draw_to_world (map, event->x, event->y, &map->xcur, &map->ycur) ;

   map->dragx = map->xcur ;
   map->dragy = map->ycur ;
   map->dragz = map->zoom ;
   map->dragd = HYPOT (event->x - widget->allocation.width / 2.0,
                        widget->allocation.height / 2.0 - event->y) ;
   map->draga = map->rot ;
   map->dragr = atan2 (widget->allocation.height / 2.0 - event->y, 
                       event->x - widget->allocation.width / 2.0) ;

   if (map->timer)
      gtk_timeout_remove (map->timer) ;
   map->timer = gtk_timeout_add (MAP_DELAY, cxc_map_timer, map) ;
   /* if (map->timer)
      gtk_timeout_remove (map->timer) ;
   map->timer = 0 ; */

   return TRUE ;
}
      
static gint
cxc_map_button_release (GtkWidget *widget, GdkEventButton *event, CxcMap *map)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (GTK_IS_DRAWING_AREA (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   if (map->list == NULL)
      return FALSE ;
   if (!map->grabbed)
      return FALSE ;
   if (event->state & (GDK_BUTTON2_MASK | GDK_BUTTON3_MASK)) 
      return FALSE ;

   if (*map->pointer_action != POINTER)
      gtk_widget_draw (map->draw_area, NULL) ; /*** FIXME: SERVE ?????? ***/

   draw_to_world (map, event->x, event->y, &map->xcur, &map->ycur) ;
   cxc_map_geometry (widget, map, event->x, event->y) ;
   map->grabbed = FALSE ;

   if (map->timer) {
      gtk_timeout_remove (map->timer) ;
      map->timer = 0 ;
      cxc_map_remap (NULL, map) ;
   }

   return TRUE ;
}

static gint
cxc_map_enter_notify_event (GtkWidget *widget, GdkEventCrossing *event, CxcMap *map)
{
   gdouble         xpos, ypos ;

   if (!map->list)
      return TRUE ;
   if (!GTK_WIDGET_MAPPED (map->draw_area))
      return FALSE ;
   CPKDEBUG5 ("cxc_map_enter_notify_event", 7) ;

   if (*map->mode == FRAME_MERGE) {
      draw_to_world (map->reference, event->x, event->y, &xpos, &ypos) ;
      map->reference->xcur = xpos ;
      map->reference->ycur = ypos ;
   }
   if (!map->grabbed)
      return FALSE ;
   if (map->timer) {
      gtk_timeout_remove (map->timer) ;
      map->timer = 0 ;
      /* cxc_map_remap (NULL, map) ; */
   }
   return TRUE ;
}

static gint
cxc_map_leave_notify_event (GtkWidget *widget, GdkEventCrossing *event, CxcMap *map)
{
   CPKDEBUG5 ("cxc_map_leave_notify_event", 4);
   if (!map->list)
      return TRUE ;
   if (map->timer) {
      gtk_timeout_remove (map->timer) ;
      map->timer = 0 ;
      cxc_map_remap (NULL, map) ;
   }
   if (map->timer_id)
      gtk_timeout_remove (map->timer_id) ;
   map->timer_id = 0 ;
   return TRUE ;
}

static gint
cxc_map_timer (gpointer data)
{
   CxcMap *map = CXC_MAP (data) ;
   CPKDEBUG5 ("cxc_map_timer", 4) ;

   GDK_THREADS_ENTER ();
   if (map->timer) {
      map->timer = 0 ;
      CXC_MAP_SET_RESCALING (map) ;
      cxc_map_remap (NULL, map) ;
   }
   GDK_THREADS_LEAVE ();

   return FALSE ;
}

static gint
cxc_map_motion_notify_event (GtkWidget *widget, GdkEventMotion *event, CxcMap *map)
{
   GdkModifierType state;
   gdouble         x, y ;
   gint            ix, iy ;
   gdouble         xpos, ypos ;

   CPKDEBUG5 ("cxc_map_motion_notify", 8) ;

   if (!map->list)
      return TRUE ;

   if (*map->mode == FRAME_BLINK) 
      return TRUE ;

   /* Get world cursor coordinates */
   if (event->is_hint) {
      gdk_window_get_pointer (event->window, &ix, &iy, &state);
      x = (gdouble)ix ;
      y = (gdouble)iy ;
   }
   else {
      x = event->x ;
      y = event->y ;
      state = event->state ;
   }
   if (*map->mode == FRAME_MERGE) {
      draw_to_world (map->reference, x, y, &xpos, &ypos) ;
      map->reference->xcur = xpos ;
      map->reference->ycur = ypos ;
   }
   draw_to_world (map, x, y, &xpos, &ypos) ;
   map->xcur = xpos ;
   map->ycur = ypos ;

   if (x < 0 || x > widget->allocation.width || 
       y < 0 || y > widget->allocation.height)
      return FALSE ;

   if (state & GDK_BUTTON1_MASK && map->pixmap != NULL) {
      if (!map->grabbed)
         return FALSE ;
      cxc_map_geometry (widget, map, x, y) ;

      if (map->timer)
         gtk_timeout_remove (map->timer) ;
      map->timer = gtk_timeout_add (MAP_DELAY, cxc_map_timer, map) ;
      cxc_map_remap (NULL, map) ;

      return TRUE ;
   }

   /* Look for nearest star */
   if (map->timer_id)
      gtk_timeout_remove (map->timer_id) ;
   map->timer_id = gtk_timeout_add (MAP_DELAY_ID, cxc_map_timer_id, map) ;

   CPKDEBUG5 ("cxc_map_motion_notify_end", 8);
   return TRUE ;
}

static void
cxc_map_draw_center (GtkWidget *widget, CxcMap *map)
{
   GdkPoint points[10] ;
   cpk_poly_plus (points, widget->allocation.width / 2, 
                     widget->allocation.height / 2, 0, 10) ;
   gdk_draw_polygon (widget->window, map->gc[MAP_GRAY], FALSE, points, 9) ;
}

static gint
cxc_map_geometry (GtkWidget *widget, CxcMap *map, gdouble x, gdouble y)
{
   gdouble dx, dy, dr, ix ;

   if (!map->selected && *map->mode != FRAME_SINGLE && 
       *map->mode != FRAME_MERGE && *map->mode != FRAME_COINC)
      return TRUE ;
   switch (*map->pointer_action) {
      case PAN :
         map->panx -= (map->xcur-map->dragx) ;
         map->pany -= (map->ycur-map->dragy) ;
         break ;
      case ZOOM :
         if (ABS (map->dragd) < 1.0e-5)
            return FALSE ;
         dx = (gdouble)x - widget->allocation.width / 2.0 ;
         dy = widget->allocation.height / 2.0 -(gdouble)y ;
         dr = HYPOT (dx, dy) ;
         map->zoom = map->dragz * MAX ( ABS (dr / map->dragd), 1.0e-5) ;
         map->izoom = -1 ;
         break ;
      case ROTATE :
         dx = (gdouble)x - widget->allocation.width / 2.0 ;
         dy = widget->allocation.height / 2.0 - (gdouble)y ;
         if (ABS (dx) < 1.0e-15 && ABS (dy) < 1.0e-15) 
            return FALSE ;
         dr =  atan2 (dy, dx) ;
         ix = ((map->list->out_sph) ? 1.0 : -1.0) * map->swx * map->swy ;
         map->rot = ang_res_sym (map->draga - ix * (dr - map->dragr)) ;
         break ;
      default :
         return TRUE ;
   }
   return TRUE ;
}

static gint
cxc_map_timer_id (gpointer data)
{
   CxcMap *map = CXC_MAP (data) ;
   Star   *s ;


   CPKDEBUG5 ("cxc_map_timer_id", 4) ;

   GDK_THREADS_ENTER ();
   if (map->timer_id) {
      map->timer_id = 0 ;

      if (GTK_WIDGET_MAPPED (map->draw_area) && 
          GTK_WIDGET_VISIBLE (map->draw_area->parent)) {
         s = cxc_map_find_nearest (map, map->xcur, map->ycur) ;
         cxc_map_draw_mark (map->draw_area, map, s, TRUE) ;
      }
   }
   GDK_THREADS_LEAVE ();

   return FALSE ;
}

/* Returns index of nearest star */
static Star *
cxc_map_find_nearest (CxcMap *map, gdouble x, gdouble y)
{
   Star   *s = NULL ;
   gint    i,ii ;
   gdouble sqdist, near, xdist, ydist ;
   near = 30*30 ;

   if (!map->list)
      return s ;
   CPKDEBUG5 ("cxc_map_find_nearest", 7) ;
   for (i = 0; i < (int)map->list->nstar; i++) {
      xdist = map->list->data[i]->xrot - x ;
      ydist = map->list->data[i]->yrot - y ;
      sqdist = xdist*xdist + ydist*ydist ;
      if (sqdist <= near) {
         near = sqdist;
         ii = i ;
         s = map->list->data[i] ;
      }
   }
   return s ;
}

void
cxc_map_draw_mark (GtkWidget *widget, CxcMap *map, Star *s, gboolean emit)
{
   GdkRectangle  update_rect ;
   GdkPoint      point[10] ;
   gdouble       xd, yd ;

   g_return_if_fail (widget != NULL) ;
   g_return_if_fail (GTK_IS_DRAWING_AREA (widget)) ;
   if (s == map->s)
      return ;
   CPKDEBUG5 ("cxc_map_draw_mark", 7) ;

   /* Clear old mark */
   if (map->s) {
      world_to_draw (map, map->s->xrot, map->s->yrot, &xd, &yd) ;
      update_rect.x = (gint)xd - 15 ;
      update_rect.y = (gint)yd - 15 ;
      update_rect.width = update_rect.height = 31 ;
      gtk_widget_draw (widget, &update_rect) ;
   }
   /* Draw new mark */
   map->s = s ;
   if (s) {
      gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 1, GDK_LINE_SOLID,
                                  GDK_CAP_PROJECTING, GDK_JOIN_ROUND) ;
      world_to_draw (map, s->xrot, s->yrot, &xd, &yd) ;
      cpk_poly_plus (point, (gint)xd, (gint)yd, 0, 15) ;
      gdk_draw_polygon (widget->window, map->gc[MAP_HIGHLIGHT], FALSE, point, 9) ;
   }
   if (emit)
      gtk_signal_emit_by_name (GTK_OBJECT (map), "star-changed") ;
}


static GdkGC **
cxc_map_set_stack_gc (CxcMap *map)
{
   GdkGC    **gc = NULL ;
   guint      i, nf ;
   gfloat     f ;

   nf = map->fnorm ;
   gc = (GdkGC **) calloc ((size_t)nf, sizeof(GdkGC *)) ;

   for (i = 0; i < nf; i++) {
      gc[i] = gdk_gc_new (map->draw_area->window) ;
      if (nf < 2)
         f = 1.0 ;
      else
         f = i / (gfloat) (map->fnorm-1) ;
      cpk_set_ns_color (map->draw_area, gc[i], f) ;
   }
   map->ngc = nf ;

   return gc ;
}

static void
cxc_map_free_stack_gc (CxcMap *map)
{
   guint  i = 0 ;
   for (i = 0; i < map->ngc; i++) 
      gdk_gc_unref (map->stack_gc[i]) ;
   free (map->stack_gc) ;
   map->stack_gc = NULL ;
}

static void
cxc_map_check_fnorm (CxcMap *map)
{
   if (!map->color)
      return ;
   if (map->stack_gc && map->ngc != map->fnorm)
      cxc_map_free_stack_gc (map) ;
   if (!map->stack_gc)
      map->stack_gc = cxc_map_set_stack_gc (map) ;
}

static void
cxc_map_draw (GdkPixmap *pixmap, CxcMap *map, gint mode)
{
   gint       i, nplot ;
   guint      j ;
   gdouble    x, y, r, rmx, mag, x0, y0 ;
   gdouble    dx1, dx2, dy1, dy2 ;
   Stack     *tab ;
   gboolean   displacement = (*map->mode != FRAME_BLINK && 
                               map->model->status > HIGH_ALIGN) ;

   if (!map->list)
      return ;
   CPKDEBUG ("cxc_map_draw");

   cxc_map_set_const (map) ;

   dx1 = dy1 = 1.0e10 ;
   dx2 = dy2 = -1.0e10 ;

   /* Set GCs for STACK MAP only */
   cxc_map_check_fnorm (map) ;

   rmx = map->radius * map->zoom ;
   nplot = (map->timer) ? MIN (MAP_NPLOT, map->list->nstar) : map->list->nstar ;
   for (i = nplot-1; i >= 0; i--) {
   /* for (i = 0; i < nplot; i++) { */
      world_to_draw (map, map->list->data[i]->xrot, map->list->data[i]->yrot, 
                     &x, &y) ;
      if (RESCALING) {
         if (x < dx1)
            dx1 = x ;
         if (x > dx2)
            dx2 = x ;
         if (y < dy1)
            dy1 = y ;
         if (y > dy2)
            dy2 = y ;
      }
      if (x > -rmx && x < map->nx + rmx && y > -rmx && y < map->ny + rmx) {
         mag = map->list->data[i]->mcal ;
         mag = (mag > map->mcut2) ? map->mcut2 : 
               (mag < map->mcut1) ? map->mcut1 : mag ;
         r = rmx * (map->mcut2 - mag) / map->mrange + 0.5 ;
         switch (mode) {
         case 0 :
            if (displacement) {
               world_to_draw (map, 
                              (map->list->data[i]->xprj * map->model->pc[0][0] + 
                               map->list->data[i]->yprj * map->model->pc[0][1]) * 
                               map->model->cdelt[0],
                              (map->list->data[i]->xprj * map->model->pc[1][0] + 
                               map->list->data[i]->yprj * map->model->pc[1][1]) *
                               map->model->cdelt[1], 
                              &x0, &y0) ;
               cpk_plot_seg (pixmap, map->gc[MAP_DISPL], x0, y0, x, y) ;
            }
            if (map->list->data[i]->flag == 0) 
               cpk_plot_symb (pixmap, map->gc[MAP_NORMAL], CPK_CIRCLE, x, y, 0., r, 25) ;
            else
               cpk_plot_symb (pixmap, map->gc[MAP_FLAGGED], CPK_CIRCLE, x, y, 0., r, 25) ;
            break ;
         case 1 :
            /* cpk_plot_symb (pixmap, map->gc[MAP_SHADE], CPK_CROSS, x, y, 0., r, 9) ; */
            cpk_plot_symb (pixmap, map->gc[MAP_SHADE], CPK_CIRCLE, x, y, 45., r, 4) ;
            break ;
         case 2 :
            tab = (Stack*)map->list->data[i]->p ;
            j = bin_get_count (tab->ns) + bin_get_count (tab->ns2) - 1 ;            
            cpk_plot_symb (pixmap, map->stack_gc[j], CPK_CIRCLE, x, y, 0., r, 25) ;
            break ;
         }
      }
   }
   if (RESCALING) {
      map->dx = dx2 - dx1 ; 
      map->dy = dy2 - dy1 ;
   }

   /* map->origin[0] = map->model->crpix[0] ;
   map->origin[1] = map->model->crpix[1] ; */
   map->origin[0] = map->list->origin[0] ;
   map->origin[1] = map->list->origin[1] ;
}

void
cxc_map_get_ref_wcs (CxcMap *map)
{
   guint      i ;
   gdouble    xd, yd, xw, yw ;
   for (i = 0; i < (guint)map->list->nstar; i++) {
      world_to_draw (map, map->list->data[i]->xrot, map->list->data[i]->yrot, 
                     &xd, &yd) ;
      draw_to_world (map->reference, xd, yd, &xw, &yw) ;
      map->list->data[i]->xrot = xw ;
      map->list->data[i]->yrot = yw ;
   }
}

void
world_to_draw (CxcMap *map, gdouble x, gdouble y, gdouble *xd, gdouble *yd)
{
   gdouble dx, dy ;
   dx = x - map->panx ;
   dy = y - map->pany ;
   *xd = (map->cos_rot*dx - map->sin_rot*dy) * map->map_scale * map->swx *
         map->swxx + map->nx/2.0 ;
   *yd = (-map->sin_rot*dx - map->cos_rot*dy) * map->map_scale * map->swy +
          map->ny/2.0 ;
}

void
draw_to_world (CxcMap *map, gdouble xd, gdouble yd, gdouble *x, gdouble *y)
{
   gdouble dx, dy ;
   dx = (map->nx / 2.0 - xd) * map->swxx * map->swx ;
   dy = (map->ny / 2.0 - yd) * map->swy ;
   *x = - ( map->cos_rot*dx - map->sin_rot*dy) / map->map_scale + map->panx ;
   *y = - (-map->sin_rot*dx - map->cos_rot*dy) / map->map_scale + map->pany ;
}

/*----------------------------------------------------------------------------
 * 							Paner
 *--------------------------------------------------------------------------*/

  /* Create a new backing pixmap of the appropriate size */
static gint
cxc_pan_configure_event (GtkWidget *widget, GdkEventConfigure *event, CxcMap *map)
{
   if (map->pixmap_pan) 
      if (map->npx == widget->allocation.width &&
          map->npy == widget->allocation.height )
         return TRUE ;

   /* CPKDEBUG ("cxc_pan_configure_event"); */

   /* Allocate pixmap */
   if (map->pixmap_pan) 
      gdk_pixmap_unref (map->pixmap_pan) ;

   map->pixmap_pan = gdk_pixmap_new (widget->window,
                                     widget->allocation.width,
                                     widget->allocation.height, -1) ;
   map->npx = widget->allocation.width ;
   map->npy = widget->allocation.height ;

   cxc_pan_remap (widget, map) ;

   return TRUE;
}  

gint
cxc_pan_remap (GtkWidget *widget, CxcMap *map)
{
   if (!map->pixmap_pan)
      return TRUE ;
   if (!GTK_WIDGET_REALIZED (map->draw_area))
      return TRUE ;
   CPKDEBUG5 ("cxc_pan_remap", 6) ;

   cxc_map_set_pan_scale (map) ;

   /* Set pixmap backgound to white */
   gdk_draw_rectangle (map->pixmap_pan, map->paner->style->white_gc, TRUE, 0, 0,
         map->paner->allocation.width, map->paner->allocation.height) ;

   /* draw paner */
   cxc_pan_draw (map) ;

   /* Settare solo l'area esposta */
   gdk_draw_pixmap (map->paner->window,
         map->paner->style->fg_gc[GTK_WIDGET_STATE (map->paner)], 
         map->pixmap_pan, 0, 0, 0, 0, -1,-1) ;

   return TRUE ;
}

/* Redraw the screen from the backing pixmap */
static gint
cxc_pan_map_event (GtkWidget *widget, GdkEvent *event, CxcMap *map)
{
   /* CPKDEBUG ("cxc_pan_map_event") ; */
   return FALSE ;
}

static gint
cxc_pan_expose_event (GtkWidget *widget, GdkEventExpose *event, CxcMap *map)
{
   CPKDEBUG5 ("cxc_pan_expose_event", 7);
   gdk_draw_pixmap (widget->window, 
         widget->style->fg_gc[GTK_WIDGET_STATE (widget)], map->pixmap_pan,
         event->area.x, event->area.y, event->area.x, event->area.y,
         event->area.width, event->area.height) ;
  return FALSE ;
}

static void
cxc_pan_draw (CxcMap *map)
{
   guint     i, j ;
   gdouble   x, y, r, x1, y1, x2, y2, mcut1, mcut2, range ;
   Stack    *tab ;

   /* Set GCs for STACK MAP only */
   cxc_map_check_fnorm (map) ;

   /* Plot back map */
   mcut1 = map->list->data[0]->mcal ;
   mcut2 = MIN (map->mcut2, map->list->data[map->nbright]->mcal) ;
   range = mcut2 - mcut1 + 0.15 ;

   for (i = 0; i < map->nbright; i++) {
      pan_world_to_draw (map, map->list->data[i]->xrot, map->list->data[i]->yrot, 
                         &x, &y) ;
      r = 4 * (mcut2 - map->list->data[i]->mcal + 0.15) / range ;
      if (r >= 0) {
         if (map->color && SHOW_COLOR) {
            tab = (Stack*)map->list->data[i]->p ;
            j = bin_get_count (tab->ns) + bin_get_count (tab->ns2) - 1 ;
            draw_circle (map->pixmap_pan, map->stack_gc[j], x, y, r, 10) ;
         }
         else {
            if (map->list->data[i]->flag == 0) 
               draw_circle (map->pixmap_pan, map->gc[MAP_NORMAL], x, y, r, 10) ;
            else
               draw_circle (map->pixmap_pan, map->gc[MAP_FLAGGED], x, y, r, 10) ;
         }
      }
   }

   /* Plot frame border */
   draw_to_world (map, 0.0, 0.0, &x, &y) ;
   pan_world_to_draw (map, x, y, &x1, &y1) ;
   draw_to_world (map, (gdouble)map->nx-1, (gdouble)map->ny-1, &x, &y) ;
   pan_world_to_draw (map, x, y, &x2, &y2) ;
   
   if (map->paner_active) 
      gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 2,
            GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER) ;
   else
      gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 1,
            GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER) ;

   gdk_draw_rectangle (map->pixmap_pan, map->gc[MAP_HIGHLIGHT], FALSE,
         (gint)x1+1, (gint)y1+1, (gint)(x2-x1), (gint)(y2-y1)) ;
}

static void
cxc_pan_draw_border (CxcMap *map)
{
   gdouble    x, y, x1, y1, x2, y2 ;

   /* Plot frame border */
   draw_to_world (map, 0.0, 0.0, &x, &y) ;
   pan_world_to_draw (map, x, y, &x1, &y1) ;
   draw_to_world (map, (gdouble)map->nx-1, (gdouble)map->ny-1, &x, &y) ;
   pan_world_to_draw (map, x, y, &x2, &y2) ;
   
   if (map->paner_active) 
      gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 2,
            GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER) ;
   else
      gdk_gc_set_line_attributes (map->gc[MAP_HIGHLIGHT], 1,
            GDK_LINE_SOLID, GDK_CAP_BUTT, GDK_JOIN_MITER) ;

   gdk_draw_rectangle (map->paner->window, map->gc[MAP_HIGHLIGHT], FALSE, 
         (gint)x1+1, (gint)y1+1, (gint)(x2-x1), (gint)(y2-y1)) ;
}

static gint
cxc_pan_button_press_event (GtkWidget *widget, GdkEventButton *event, CxcMap *map)
{
   GdkModifierType state;
   gint            x, y ;
   gdouble         xw, yw, xd, yd ;
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (GTK_IS_DRAWING_AREA (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;

   if (*map->mode == FRAME_BLINK)
      return FALSE ;

   gdk_window_get_pointer (event->window, &x, &y, &state);

   if (state & GDK_BUTTON3_MASK)
      return FALSE ;

   pan_draw_to_world (map, event->x, event->y, &xw, &yw) ;
   world_to_draw (map, xw, yw, &xd, &yd) ;

   if (state & GDK_BUTTON2_MASK) {
      map->panx = xw ;
      map->pany = yw ;
      map->dragx = xw-map->panx ;
      map->dragy = yw-map->pany ;
      map->paner_drag = TRUE ;
      if (*map->mode == FRAME_MERGE) {
         draw_to_world (map->reference, xd, yd, &xw, &yw) ;
         map->reference->panx = xw ;
         map->reference->pany = yw ;
         map->reference->dragx = 0 ;
         map->reference->dragy = 0 ;
      }
      cxc_map_remap (NULL, map) ;
   }
   else {
 
      if (xd < 0 || xd > map->nx || yd < 0 || yd > map->ny)
         return FALSE ;

      map->dragx = xw-map->panx ;
      map->dragy = yw-map->pany ;
      map->paner_drag = TRUE ;
      if (*map->mode == FRAME_MERGE) {
         draw_to_world (map->reference, xd, yd, &xw, &yw) ;
         map->reference->dragx = xw-map->reference->panx ;
         map->reference->dragy = yw-map->reference->pany ;
      }
   }
   return TRUE ;
}

static gint
cxc_pan_button_release (GtkWidget *widget, GdkEventButton *event, CxcMap *map)
{
   g_return_val_if_fail (widget != NULL, FALSE) ;
   g_return_val_if_fail (GTK_IS_DRAWING_AREA (widget), FALSE) ;
   g_return_val_if_fail (event != NULL, FALSE) ;
   if (map->paner_active) {
      map->paner_drag = map->paner_active = FALSE ;
      cxc_pan_remap (NULL, map) ;
      map->paner_active = TRUE ;
      cxc_pan_draw_border (map) ;
   }
   if (map->timer) {
      gtk_timeout_remove (map->timer) ;
      map->timer = 0 ;
      cxc_map_remap (NULL, map) ;
   }
   return FALSE ;
}

static gint
cxc_pan_motion_notify_event (GtkWidget *widget, GdkEventMotion *event, CxcMap *map)
{
   GdkModifierType state;
   gint            ix, iy ;
   gdouble         x, y, xw, yw, xd, yd ;
   gboolean        active ;

   if (event->is_hint) {
      gdk_window_get_pointer (event->window, &ix, &iy, &state) ;
      x = (gdouble)ix ;
      y = (gdouble)iy ;
   }
   else {
      x = event->x ;
      y = event->y ;
      state = event->state ;
   }
   pan_draw_to_world (map, x, y, &xw, &yw) ;
   world_to_draw (map, xw, yw, &xd, &yd) ;

   if (xd < 0 || xd > map->nx || yd < 0 || yd > map->ny) 
      active = FALSE ;
   else 
      active = TRUE ;

   if (active != map->paner_active) {
      map->paner_active = active ;
      gtk_widget_draw (map->paner, NULL) ;
      cxc_pan_draw_border (map) ;
   }

   if (!map->paner_drag)
      return FALSE ;

   if (state & (GDK_BUTTON1_MASK | GDK_BUTTON2_MASK) && map->pixmap != NULL) {
      if (!map->selected && *map->mode != FRAME_SINGLE && 
          *map->mode != FRAME_MERGE && *map->mode  != FRAME_COINC)
         return TRUE ;
      map->panx = (xw-map->dragx) ;
      map->pany = (yw-map->dragy) ;
      if (*map->mode == FRAME_MERGE) {
         draw_to_world (map->reference, xd, yd, &xw, &yw) ;
         map->reference->panx = (xw-map->reference->dragx) ;
         map->reference->pany = (yw-map->reference->dragy) ;
      }

      if (map->timer)
         gtk_timeout_remove (map->timer) ;
      map->timer = gtk_timeout_add (MAP_DELAY, cxc_map_timer, map) ;
      cxc_map_remap (NULL, map) ;
   }
   return TRUE ;
}

static void
pan_world_to_draw (CxcMap *map, gdouble x, gdouble y, gdouble *xd, gdouble *yd)
{
   gdouble dx, dy ;

   dx = x - (map->fx1 + map->fx2) / 2.0 ;
   dy = y - (map->fy1 + map->fy2) / 2.0 ;

   *xd = ( map->cos_rot*dx - map->sin_rot*dy)*map->pan_scale*map->swx*map->swxx + 
         map->paner->allocation.width / 2.0 ;
   *yd = (-map->sin_rot*dx - map->cos_rot*dy)*map->pan_scale*map->swy + 
         map->paner->allocation.height / 2.0 ;
}

static void
pan_draw_to_world (CxcMap *map, gdouble xd, gdouble yd, gdouble *x, gdouble *y)
{
   gdouble dx, dy ;

   dx = (map->paner->allocation.width  / 2.0 - xd) * map->swx * map->swxx ;
   dy = (map->paner->allocation.height / 2.0 - yd) * map->swy ;

   *x = - ( map->cos_rot*dx - map->sin_rot*dy) / map->pan_scale + 
          (map->fx1 + map->fx2) / 2.0 ;
   *y = - (-map->sin_rot*dx - map->cos_rot*dy) / map->pan_scale + 
          (map->fy1 + map->fy2) / 2.0 ;
}

static void
cxc_map_set_const (CxcMap *map)
{
   gdouble scale ;

   map->swxx = (map->list->out_sph) ? -1.0 : 1.0 ;

   scale = (map->list->out_sph) ? sph_scale : lin_scale ;
   scale *= map->zoom ;

   if (map->map_scale != scale || 
       map->cos_rot != cos (map->rot) ||
       map->sin_rot != sin (map->rot)) {

      CXC_MAP_SET_RESCALING (map) ;

      map->map_scale = scale ;
      map->cos_rot = cos (map->rot) ;
      map->sin_rot = sin (map->rot) ;
   }
}

static void
cxc_map_set_pan_scale (CxcMap *map)
{
   gdouble scale = map->map_scale ;
   scale *= MIN (map->paner->allocation.width, map->paner->allocation.height) / 
            MAX (map->dx, map->dy) ;
   scale *= 0.9 ;
   map->pan_scale = scale ;
}

