/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <math.h>

#include <gtk/gtk.h>

#include "catalib.h"
#include "cataxcorr_p.h"
#include "cp_messages.h"
#include "cxc_gtk_main.h"

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static void        cxc_res_close_window     (GtkWidget         *window, 
                                             gpointer           data) ;
static void        cxc_res_save             (cxc_display       *display) ;

static void        cxc_res_init             (CxcRes            *residuals) ;

static gint        cxc_res_configure        (GtkWidget         *widget, 
                                             GdkEventConfigure *event, 
                                             CxcRes            *residuals) ;
static gint        cxc_res_expose           (GtkWidget         *widget, 
                                             GdkEventExpose    *event, 
                                             CxcRes            *residuals) ;
static gint        cxc_res_remap            (GtkWidget         *widget, 
                                             CxcRes            *residuals) ;
static void        cxc_res_draw             (CxcRes            *residuals) ;
static void        cxc_res_draw_axes        (CxcRes            *residuals) ;
static void        cxc_res_draw_stdev       (CxcRes            *residuals, 
                                             gdouble            xm, 
                                             gdouble            ym, 
                                             gdouble            dxdev, 
                                             gdouble            dydev) ;

static void        res_word_to_draw         (CxcRes            *residuals, 
                                             gdouble            x, 
                                             gdouble            y, 
                                             gdouble           *xd, 
                                             gdouble           *yd) ;
static void        res_draw_to_word         (CxcRes            *residuals, 
                                             gdouble            xd, 
                                             gdouble            yd, 
                                             gdouble           *x, 
                                             gdouble           *y) ;

/*----------------------------------------------------------------------------
 * 							Private types
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/



/*----------------------------------------------------------------------------
 * 							Residuals
 *--------------------------------------------------------------------------*/

static void
cxc_res_close_window (GtkWidget *window, gpointer data)
{
   gint i ;
   cxc_display *display = data ;
   if (display->residuals->pixmap) 
      gdk_pixmap_unref (display->residuals->pixmap) ;
   for (i = 0; i < 6; i++) 
      gdk_gc_unref (display->residuals->gc[i]) ;
   gtk_signal_disconnect_by_func (GTK_OBJECT (display->residuals->data), 
            GTK_SIGNAL_FUNC (cxc_res_remap), display->residuals) ;
   display->residuals->dialog = NULL ;
}

static void
cxc_res_save (cxc_display *display)
{
   FILE     *out ;
   PlateMod *model = display->residuals->map->model ;
   char      lab[40], lab1[40] ;
   int       i, f1, f2 ;
   double    dx, dy ;
   char     *filename ;

   if (model->nmatch == 0)
      return ;

   gtk_widget_set_sensitive (display->residuals->b1, FALSE) ;
   gtk_widget_set_sensitive (display->residuals->b2, FALSE) ;
   gtk_widget_set_sensitive (display->residuals->b3, FALSE) ;

   filename = cpk_get_file_name (display->residuals->dialog, NULL) ;

   gtk_widget_set_sensitive (display->residuals->b1, TRUE) ;
   gtk_widget_set_sensitive (display->residuals->b2, TRUE) ;
   gtk_widget_set_sensitive (display->residuals->b3, TRUE) ;

   if (filename == NULL)
      return ;
   
   /* Open the file */
   f1 = display->residuals->data->nref ;
   f2 = display->residuals->data->frm2 ;
   CXC_MESSAGE (0, 2, "Saving residuals for '%s'\n", 
         display->residuals->data->f_set->info[f2]->filename) ;
   out = fopen (filename, "w") ;

   /* Write header lines */
   fprintf (out, "#%-8s  %-69s\n", "COMMENT", 
            "Generated by CataXcorr - Version " VERSION) ;

   strncpy (lab, display->residuals->data->f_set->info[f1]->filename, 38) ;
   sprintf (lab1, "'%s'", lab) ;
   fprintf (out, "#%-8s= %-20s / %-46s\n", "REFCAT", lab1, 
         "Name of reference file") ;
   strncpy (lab, display->residuals->data->f_set->info[f2]->filename, 38) ;
   sprintf (lab1, "'%s'", lab) ;
   fprintf (out, "#%-8s= %-20s / %-46s\n", "PARENT", lab1, 
         "Name of parent file") ;
   switch (model->fitmodel) {
      case LIN:       sprintf (lab, "'LIN'") ;       break ;
      case SAO:       sprintf (lab, "'SAO'") ;       break ;
      case POLY:      sprintf (lab, "'POLY'") ;      break ;
      case POLYRAD:   sprintf (lab, "'POLYRAD'") ;   break ;
      case LEGENDRE:  sprintf (lab, "'LEGENDRE'") ;  break ;
      case NUM_MODEL: sprintf (lab, "''") ;          break ;
      default: break ;
   }
   fprintf (out, "#%-8s= %-20s / %-46s\n", "PLTMODEL", lab, 
         "Plate Model Function") ;
   fprintf (out, "#%-8s= %-20d / %-46s\n", "PLTORDER", model->order, 
         "Plate Model Order") ;
   fprintf (out, "#%-8s= %-20d / %-46s\n", "NMATCH", model->nmatch, 
         "Number of matched points") ;
   fprintf (out, "#%-8s= %#20.14g / %-46s\n", "X_RMS", 
         ABS(model->cdelt[0]) * model->dxdev, 
         "X-axis model Root Mean Square deviation") ;
   fprintf (out, "#%-8s= %#20.14g / %-46s\n", "Y_RMS", 
         ABS(model->cdelt[1]) * model->dydev, 
         "Y-axis model Root Mean Square deviation") ;
   if (model->spheric) sprintf (lab, "'%s'", "deg") ;
   else sprintf (lab, "'%s'", "pixel") ;
   fprintf (out, "#%-8s= %-20s / %-46s\n", "CUNIT1", lab, 
            "Units of coordinate values") ;
   fprintf (out, "#%-8s= %-20s / %-46s\n", "CUNIT2", lab, 
            "Units of coordinate values") ;
   fprintf (out, "#%-8s= %-20s / %-46s\n", "DATE", 
            iso_8601_date_format (NULL, FALSE), 
            "Date Residuals file was generated") ;
   fprintf (out, "#\n") ;

   /* Write residuals */
   for (i = 0; i < model->nmatch; i++) {
      dx = model->match[i].s2->xrot - model->match[i].s1->xrot ;
      dy = model->match[i].s2->yrot - model->match[i].s1->yrot ;
      fprintf (out, "   % #.9g    % #.9g\n", dx, dy) ;
   }
   fclose (out) ;
}

void
cxc_res_frame (GtkWidget *window, cxc_display *display)
{
   GtkWidget *dialog, *box, *frame, *button, *table, *label, *hbox ;
   guint      current = cxc_second_frame (display) ;

   CPKDEBUG ("cxc_res_frame") ;

   if ( display->residuals->dialog ) {
      /* Window is already opened */
      cxc_res_remap (NULL, display->residuals) ;
      gtk_window_set_transient_for (GTK_WINDOW (display->residuals->dialog), 
            GTK_WINDOW (display->window)) ;
      return ;
   }

   display->residuals->dialog = dialog = gtk_dialog_new () ;
   gtk_window_set_title (GTK_WINDOW (dialog), "CataXcorr: Fit Residuals") ;
   gtk_window_set_position (GTK_WINDOW (dialog), GTK_WIN_POS_MOUSE) ;
   gtk_signal_connect (GTK_OBJECT (dialog), "destroy",
         GTK_SIGNAL_FUNC (cxc_res_close_window), (gpointer)display) ;

   cxc_res_init (display->residuals) ;

   display->residuals->data = CXC_DATA (display->data) ;
   display->residuals->map = CXC_MAP (display->map[current]) ;

   gtk_signal_connect (GTK_OBJECT (display->residuals->data), "model-fit",
         (GtkSignalFunc) cxc_res_remap, display->residuals) ;

   box = gtk_vbox_new (FALSE, 0) ;
   frame = gtk_frame_new (NULL) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 8) ;
   gtk_frame_set_shadow_type (GTK_FRAME (frame), GTK_SHADOW_IN) ;
   gtk_widget_set_usize (frame, 250, 250) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 0) ;
   gtk_container_add (GTK_CONTAINER (box), frame) ;

   gtk_container_add (GTK_CONTAINER (frame), display->residuals->draw_area) ;   

   frame = gtk_frame_new (NULL) ;
   gtk_widget_set_usize (frame, 250, -1) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 8) ;
   /* gtk_container_add (GTK_CONTAINER (box), frame) ; */
   gtk_box_pack_start (GTK_BOX (box), frame, FALSE, FALSE, 8) ;
   table = gtk_table_new (3, 5, FALSE) ;
   gtk_table_set_col_spacings (GTK_TABLE (table), 2) ;
   gtk_table_set_row_spacings (GTK_TABLE (table), 1) ;
   gtk_container_set_border_width (GTK_CONTAINER (frame), 2) ;
   gtk_container_add (GTK_CONTAINER (frame), table) ;

   label = gtk_label_new ("X rms") ;
   hbox = gtk_hbox_new (FALSE, 0) ;
   gtk_widget_set_usize (hbox, 50, 0);
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 4) ;
   gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, 0, 1, GTK_FILL, GTK_FILL, 1, 1) ;
   frame = gtk_frame_new (NULL) ;
   gtk_table_attach (GTK_TABLE (table), frame, 2, 3, 0, 1, GTK_FILL|GTK_EXPAND, GTK_FILL, 1, 1) ;
   hbox = gtk_hbox_new (FALSE, 0) ;
   display->residuals->xrms = label = gtk_label_new(NULL) ;
   gtk_container_add (GTK_CONTAINER (frame), hbox) ;
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2) ;
   display->residuals->xlab = label = gtk_label_new(NULL) ;
   gtk_widget_set_usize (label, 40, 0);
   gtk_table_attach (GTK_TABLE (table), label, 3, 4, 0, 1, GTK_FILL, GTK_FILL, 1, 1) ;

   label = gtk_label_new ("Y rms") ;
   hbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 4) ;
   gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, 1, 2, GTK_FILL, GTK_FILL, 1, 1) ;
   frame = gtk_frame_new (NULL) ;
   gtk_table_attach (GTK_TABLE (table), frame, 2, 3, 1, 2, GTK_FILL|GTK_EXPAND, GTK_FILL, 1, 1) ;
   hbox = gtk_hbox_new (FALSE, 0) ;
   display->residuals->yrms = label = gtk_label_new (NULL) ;
   gtk_container_add (GTK_CONTAINER (frame), hbox) ;
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2) ;
   display->residuals->ylab = label = gtk_label_new(NULL) ;
   gtk_table_attach (GTK_TABLE (table), label, 3, 4, 1, 2, GTK_FILL, GTK_FILL, 1, 1) ;

   label = gtk_label_new ("N matches") ;
   hbox = gtk_hbox_new (FALSE, 0) ;
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 4) ;
   gtk_table_attach (GTK_TABLE (table), hbox, 0, 1, 2, 3, GTK_FILL, GTK_FILL, 1, 1) ;
   frame = gtk_frame_new (NULL) ;
   gtk_table_attach (GTK_TABLE (table), frame, 2, 3, 2, 3, GTK_FILL|GTK_EXPAND, GTK_FILL, 1, 1) ;
   hbox = gtk_hbox_new (FALSE, 0) ;
   display->residuals->nmatch = label = gtk_label_new(NULL) ;
   gtk_container_add (GTK_CONTAINER (frame), hbox) ;
   gtk_box_pack_start (GTK_BOX (hbox), label, FALSE, FALSE, 2) ;

   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->vbox), box, TRUE, TRUE, 0) ;

   display->residuals->b1 = button = gtk_button_new_with_label ("Print") ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_set_sensitive (button, FALSE) ;

   display->residuals->b2 = button = gtk_button_new_with_label ("Save") ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (cxc_res_save), (gpointer) display) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG (dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;

   display->residuals->b3 = button = gtk_button_new_with_label ("Dismissis") ;
   gtk_signal_connect_object (GTK_OBJECT (button), "clicked",
         GTK_SIGNAL_FUNC (gtk_widget_destroy), (gpointer) dialog) ;
   gtk_box_pack_start (GTK_BOX (GTK_DIALOG(dialog)->action_area),
         button, TRUE, TRUE, 0) ;
   GTK_WIDGET_SET_FLAGS (button, GTK_CAN_DEFAULT) ;
   gtk_widget_grab_default (button) ;

   gtk_widget_show_all (dialog) ;
   
   return ;
}

static void
cxc_res_init(CxcRes *residuals)
{
   CPKDEBUG ("cxc_res_init") ;

   residuals->draw_area = gtk_drawing_area_new () ;
   gtk_signal_connect (GTK_OBJECT (residuals->draw_area),"configure_event",
         (GtkSignalFunc)cxc_res_configure, residuals) ;
   gtk_signal_connect (GTK_OBJECT (residuals->draw_area), "expose_event",
         (GtkSignalFunc)cxc_res_expose, residuals) ;
   gtk_widget_set_events (residuals->draw_area, GDK_EXPOSURE_MASK) ;
   residuals->gc[0] = NULL ;
   residuals->pixmap = NULL ;

   /* gtk_widget_set_visual (GTK_WIDGET (residuals->draw_area), gdk_rgb_get_visual ()) ;
   gtk_widget_set_colormap (GTK_WIDGET (residuals->draw_area), gdk_rgb_get_cmap ()) ; */


}

static gint
cxc_res_configure(GtkWidget *widget, GdkEventConfigure *event, CxcRes *residuals)
{
   CPKDEBUG("cxc_res_configure") ;

   /* Allocate pixmap */
   if (residuals->pixmap) 
      gdk_pixmap_unref (residuals->pixmap) ;

   residuals->pixmap = gdk_pixmap_new (widget->window,
         widget->allocation.width, widget->allocation.height, -1) ;

   if (!residuals->gc[0]) {
      gint i ;
      for (i=0; i<6; i++) 
         residuals->gc[i] = gdk_gc_new (residuals->draw_area->window) ;
      cpk_set_fg_color (residuals->draw_area, residuals->gc[0], "black") ;
      cpk_set_fg_color (residuals->draw_area, residuals->gc[1], "gray75") ;
      cpk_set_fg_color (residuals->draw_area, residuals->gc[2], "red") ;
      cpk_set_fg_color (residuals->draw_area, residuals->gc[3], "grey70") ;
      cpk_set_fg_color (residuals->draw_area, residuals->gc[4], "brown") ;
      cpk_set_fg_color (residuals->draw_area, residuals->gc[5], "blue") ;
   }

   cxc_res_remap(NULL, residuals) ;

   return TRUE;
}  

static gint
cxc_res_expose(GtkWidget *widget, GdkEventExpose *event, CxcRes *residuals)
{
   CPKDEBUG ("cxc_res_expose") ;
   gdk_draw_pixmap (widget->window,
         widget->style->fg_gc[GTK_WIDGET_STATE (widget)], 
         residuals->pixmap,
         event->area.x, event->area.y, event->area.x, event->area.y, 
         event->area.width, event->area.height) ;
  return FALSE;
}

static gint
cxc_res_remap (GtkWidget *widget, CxcRes *residuals)
{
   PlateMod *model = residuals->map->model ;
   gdouble xm, ym, dxdev, dydev ;
   gchar   buffer[50] ;

   CPKDEBUG ("cxc_res_remap") ;

   if (!residuals->pixmap)
      return TRUE ;

   /* Set pixmap backgound to white */
   gdk_draw_rectangle (residuals->pixmap,
         residuals->draw_area->style->white_gc, TRUE, 0, 0, 
         residuals->draw_area->allocation.width, 
         residuals->draw_area->allocation.height) ;

   model_stats (model->match, model->nmatch, model, &xm, &ym, &dxdev, &dydev) ;

   cxc_res_draw (residuals) ;

   cxc_res_draw_axes (residuals) ;
   cxc_res_draw_stdev (residuals, xm, ym, dxdev, dydev) ;

   /* Settare solo l'area esposta */
   gdk_draw_pixmap (residuals->draw_area->window,
         residuals->draw_area->style->fg_gc[GTK_WIDGET_STATE(residuals->draw_area)],
         residuals->pixmap, 0, 0, 0, 0, -1,-1) ;

   /* Set labels */
   if (residuals->data->list2->out_sph) {
      sprintf (buffer, "%#.6g", dxdev*3600.0) ;
      gtk_label_set_text (GTK_LABEL (residuals->xrms), buffer) ;
      sprintf (buffer, "%#.6g", dydev*3600.0) ;
      gtk_label_set_text (GTK_LABEL (residuals->yrms), buffer) ;
      sprintf (buffer, "[arcsec]") ;
      gtk_label_set_text (GTK_LABEL (residuals->xlab), buffer) ;
      gtk_label_set_text (GTK_LABEL (residuals->ylab), buffer) ;
   }
   else {
      sprintf (buffer, "%#.6g", dxdev) ;
      gtk_label_set_text (GTK_LABEL (residuals->xrms), buffer) ;
      sprintf (buffer, "%#.6g", dydev) ;
      gtk_label_set_text (GTK_LABEL (residuals->yrms), buffer) ;
      sprintf (buffer, "[pixel]") ;
      gtk_label_set_text (GTK_LABEL (residuals->xlab), buffer) ;
      gtk_label_set_text (GTK_LABEL (residuals->ylab), buffer) ;
   }
   sprintf (buffer, "%d", model->nmatch) ;
   gtk_label_set_text (GTK_LABEL (residuals->nmatch), buffer) ;

   return TRUE ;
}

static void
cxc_res_draw_stdev (CxcRes *residuals, 
      gdouble xm, gdouble ym, gdouble dxdev, gdouble dydev)
{
   GdkPoint     points[256] ;
   gint         n = 256, i ;
   gdouble      a, dx, dy, x, y ;

   for(i = 0; i < n; i++) {
      a = i*360.0/(gdouble)n ;
      dx = xm + cos (DEGTORAD (a)) * dxdev ;
      dy = ym + sin (DEGTORAD (a)) * dydev ;
      res_word_to_draw (residuals, dx, dy, &x, &y) ;
      points[i].x = (gint)x ;
      points[i].y = (gint)y ;
   }
   gdk_draw_polygon (residuals->pixmap, residuals->gc[2], FALSE, points, n) ;
}

static void
cxc_res_draw_axes (CxcRes *residuals)
{
   GdkSegment axes[2], mark ;
   gboolean   s = (CXC_DATA (residuals->data)->list2->out_sph) ;
   gdouble    l = (s) ? 1.0/3600.0 : 1.0 ;
   gdouble    l1, l2 ;
   gdouble    incr = (s) ? 0.5/3600.0 : 0.5 ;
   gint       i ;
   gint16     tk ;
   gdouble    x, y ;

   axes[0].x1 = 0 ;
   axes[0].y1 = (gint16) residuals->draw_area->allocation.height / 2.0 ;
   axes[0].x2 = (gint16) residuals->draw_area->allocation.width ;
   axes[0].y2 = (gint16) residuals->draw_area->allocation.height / 2.0 ;

   axes[1].x1 = (gint16) residuals->draw_area->allocation.width / 2.0 ;
   axes[1].y1 = 0 ;
   axes[1].x2 = (gint16) residuals->draw_area->allocation.width / 2.0 ;
   axes[1].y2 = (gint16) residuals->draw_area->allocation.height ;

   gdk_draw_segments (residuals->pixmap, residuals->gc[0], axes, 2) ;

   res_draw_to_word (residuals, (gdouble)axes[0].x1, (gdouble)axes[0].y1, &x, &y) ;
   l1 = (s) ? floor (x*3600)/3600.0 : floor (x) ;
   res_draw_to_word (residuals, (gdouble)axes[0].x2, (gdouble)axes[0].y2, &x, &y) ;
   l2 = (s) ? ceil (x*3600)/3600.0 : ceil (x) ;

   for (l = l1; l < l2; l+=incr) {
      gdouble ll = (s) ? l*3600 : l ;
      tk = (ABS (floor (ll) - ll) < 1e-10) ? 4.0 : 2.0 ;
      res_word_to_draw (residuals, l,  0.0, &x, &y) ;
      mark.x1 = (gint16)x ;
      mark.y1 = (gint16)y - tk ;
      mark.x2 = (gint16)x ;
      mark.y2 = (gint16)y + tk ;
      gdk_draw_segments (residuals->pixmap, residuals->gc[0], &mark, 1) ;
   }

   res_draw_to_word (residuals, (gdouble)axes[1].x1, (gdouble)axes[1].y1, &x, &y) ;
   l1 = (s) ? floor (y*3600)/3600.0 : floor (y) ;
   res_draw_to_word (residuals, (gdouble)axes[1].x2, (gdouble)axes[1].y2, &x, &y) ;
   l2 = (s) ? ceil (y*3600)/3600.0 : ceil (y) ;
   for (l = l1; l < l2; l+=incr) {
      gdouble ll = (s) ? l*3600 : l ;
      tk = (ABS (floor (ll) - ll) < 1e-10) ? 4.0 : 2.0 ;
      res_word_to_draw (residuals, 0.0, l, &x, &y) ;
      mark.x1 = (gint16)x - tk ;
      mark.y1 = (gint16)y ;
      mark.x2 = (gint16)x + tk ;
      mark.y2 = (gint16)y ;
      gdk_draw_segments (residuals->pixmap, residuals->gc[0], &mark, 1) ;
   }
}

static void
cxc_res_draw (CxcRes *residuals)
{
   CxcMap    *map ;
   CxcData   *data ;
   PlateMod  *model = residuals->map->model ;
   guint      k ;
   gdouble    x, y, dx, dy ;

   CPKDEBUG ("cxc_res_draw") ;

   map = residuals->map ;
   data = residuals->data ;
   for (k = 0; k < (guint) model->nmatch; k++) {
      dx = model->match[k].s2->xrot - model->match[k].s1->xrot ;
      dy = model->match[k].s2->yrot - model->match[k].s1->yrot ;
      res_word_to_draw (residuals, dx, dy, &x, &y) ;
      cpk_plot_symb (residuals->pixmap, residuals->gc[3], CPK_CROSS, x, y, 0., 2, 25) ;
   }
}

static void
res_word_to_draw (CxcRes *residuals, gdouble x, gdouble y, gdouble *xd, gdouble *yd)
{
   gdouble scale ;

   scale = (CXC_DATA (residuals->data)->list2->out_sph) ? 3600.0 : 1.0 ;
   scale *= 60 ;
   *xd = x * scale + residuals->draw_area->allocation.width / 2.0 ;
   *yd = y * scale + residuals->draw_area->allocation.height / 2.0 ;
}

static void
res_draw_to_word (CxcRes *residuals, gdouble xd, gdouble yd, gdouble *x, gdouble *y)
{
   gdouble scale ;

   scale = (CXC_DATA (residuals->data)->list2->out_sph) ? 3600.0 : 1.0 ;
   scale *= 60 ;

  *x = (xd - residuals->draw_area->allocation.width / 2.0) / scale ;
  *y = (yd - residuals->draw_area->allocation.height / 2.0) / scale ;
}

