/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: CataComb.c
 * Author 	: Paolo Montegriffo
 * Created on	: July, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0 - October 6, 2000
 * Version      : 2.1.2 - September 4, 2001
 * Version      : 2.1.3 - September 26, 2001
 * Version      :         November 13, 2001 - Run lint on the code
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <ctype.h>
#include <math.h>

#include "catalib.h"
#include "fileinfo.h"
#include "cp_messages.h"
#include "expand.h"
#include "parser.h"
#include "vector_arith.h"
#include "coinc_rw.h"

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

extern item    *ParseLogicalExpression   (char *expr) ;
static void     usage                    () ;
static void     HelloWorld               () ;

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef min
#define min(a,b) (((a)<=(b)) ? (a) : (b))
#endif
#ifndef max
#define max(a,b) (((a)>=(b)) ? (a) : (b))
#endif

#define UNDEF    0
#define IS_COINC 1
#define IS_LIST  2
#define ACTIVE   7
#define DISPLAY  8
#define INFO     9

#define MAXLENINPUT 2048

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

char     *module = "CataComb" ;
char     *version = VERSION" - September 26, 2001" ;
int       verbose = 1 ;
CoincId  *coinc ;
FileSet  *f_set ;
int       sigma = 1 ;          /* 0 to turn-off sigma rejection */
int       max_iter = 1 ;       /* maximum number of sigma rejection iterations */
double    nsig_rej = 2.0 ;     /* rejection limit in stdev units */
int       n_min = 1 ;          /* minimum number of measures for a valid result */
int       N_NullVal = 2 ;      /* number of elements in Null values list */
double    NullValues[5] = { 99.999, -99.999, 0.0, 0.0, 0.0 } ;
                               /* Input Null values */

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

static void 
usage()
{
   (void)fprintf(stderr,"\nUSAGE:\n\n CataComb -h\n");
   (void)fprintf(stderr,"\n CataComb [-c coinc_tab | -l file[...]]  -f \"control_string\"  [-fmt \"string\"]\n");
   (void)fprintf(stderr,"%10s[-filter \"expr\"]  [-o filename | -display]  [-null val[,val[,...]]]\n"," ");
   (void)fprintf(stderr,"%10s[-o_null val]  [-n_min val]  [-norej | -nsig val -maxiter val]\n"," ");
   (void)fprintf(stderr,"%10s[-v [n]]\n"," ");
   (void)fprintf(stderr,"\n CataComb [-c coinc_tab | -l file[...]]  ");
   (void)fprintf(stderr,"-info\n\n");
   (void)fprintf(stderr,"\t-h:\t  show this help\n");
   (void)fprintf(stderr,"\t-c:\t  coincidence table file name\t\t[Def: \"coinc.tab\"]\n");
   (void)fprintf(stderr,"\t-l:\t  list of input file names\n");
   (void)fprintf(stderr,"\t-f:\t  output control string\n");
   (void)fprintf(stderr,"\t-fmt:\t  character string combining literal characters with\n");
   (void)fprintf(stderr,"\t\t  conversion specifications\n");
   (void)fprintf(stderr,"\t-filter:  logical filter on output records\n");
   (void)fprintf(stderr,"\t-o:\t  output file name\t\t\t[Def: STDOUT]\n");
   (void)fprintf(stderr,"\t-display: pipes data to XGobi for a fast interactive plot\n");
   (void)fprintf(stderr,"\t-null:\t  list of null values in input files\t[Def: -99.999,99.999]\n");
   (void)fprintf(stderr,"\t-o_null:  output null value \t\t\t[Def: 0.000]\n");
   (void)fprintf(stderr,"\t-v:\t  verbose level\t\t\t\t[Def: 1]\n\n");
   (void)fprintf(stderr,"    Averaging related parameters:\n");
   (void)fprintf(stderr,"\t-n_min:\t  minimum number of measures for a valid result\t[Def: 1]\n");
   (void)fprintf(stderr,"\t-norej:\t  doesn't apply sigma rejection in average computations\n");
   (void)fprintf(stderr,"\t-nsig:\t  rejection limit in stdev units \t\t[Def: 2.0]\n");
   (void)fprintf(stderr,"\t-maxiter: maximum number of sigma rejection iterations \t[Def: 1]\n\n");
   (void)fprintf(stderr,"\t-info:\t  do-nothing mode: show only files content\n");
   (void)fprintf(stderr,"\n See 'man CataComb' for a more detailed description of command usage\n") ;
   exit(1) ;
}
static void 
HelloWorld()
{
   cp_message(0,1,"\n\
\t _______________________________________________________________ \n\
\to\t\t\t\t\t\t\t\to\n\
\to  CataComb: Catalogues Combiner \t\t\t\to\n\
\to  Version:  %s\t\t\t\to\n\
\to  Author:   P. Montegriffo - Bologna Astronomical Observatory\to\n\
\to_______________________________________________________________o\n\
\n",version) ;
}

int
main(int ac, char **av)
{
   char    command[1024] = "\0" ;
   char    outfile[1024] = "\0" ;
   char    fstring[1024] = "\0" ;
   char    filter[80] = "\0" ;
   char    format[1024] = "\0" ;
   char  **pform = (char**)NULL ;
   int     nform = 0 ;
   int    *ndig = (int*)NULL ;
   char    OutNull[20] = "0.000" ;      /* Output Null value */

   char  **files = (char**)NULL ;
   int     nfiles ;
   char  **f_cols = (char**)NULL ;
   int     n_cols ;
   int     i, j, k ;
   int     status = 1 ;

   char  * tok_exp = (char*)NULL ;
   char  * expanded ;
   int     arithm = ARITHM_STANDARD ; 
   item  **result = (item**)NULL ;
   item  * flag = (item*)NULL ;
   char    fmt[15] ;
   char    line[MAX_RECORD_LENGTH] = "\n" ;
   char    outline[MAX_RECORD_LENGTH] = "\0" ;
   
   char    infile[MAXLENINPUT] = "coinc.tab" ;
   int     filetype = UNDEF ;
   int     mode = UNDEF ;
   int     filter_is = UNDEF ;
   char    yynn[5] = "\0" ;
   FILE   *out, *rej = (FILE*)NULL ;
   FILE   *outcol, *outraw = (FILE*)NULL, *outX11 ;
   char    tmpfile_dat[80] = "\0" ;
   char    tmpfile_col[80] = "\0" ;
   char    tmpfile_raw[80] = "\0" ;
   char    tmpfile_X11[80] = "\0" ;
   pid_t   proc_num ;
   Finfo **f_info = NULL ;

   /* Decode arguments list */

   (void) strcpy (command,*av) ;
   for (i = 1; i < ac; i++) 
      (void) sprintf (command, "%s %s", command, *(av+i)) ;
   for (av++; --ac>0; av++) {
      char * str ;
      /* sprintf(command,"%s %s",command,*av) ; */
      if (*(str = *av) == '-') {
         char c ;
         c = *++str ;
         switch (c) {
	 case 'h':
	    usage () ;
            break ;
	 case 'c':
	    if (filetype == IS_LIST) {
	       cp_error ("cannot specify both a coinc. table and a file list: exiting") ;
	       usage () ;
	    }
	    if (ac <= 1 || *(str = *(av+1)) == '-') {
	       cp_error ("missing file name: exiting") ;
	       usage () ;
	    }
	    filetype = IS_COINC ;
	    (void) strcpy (infile,*++av) ;
	    ac--;
            break ;
	 case 'l':
	    if (filetype == IS_COINC) {
	       cp_error ("cannot specify both a coinc. table and a file list: exiting") ;
	       usage () ;
	    }
	    if (ac<=1 || (*(str = *(av+1)) == '-' && strcmp(*(av+1),"-"))) {
	       cp_error("missing file name: exiting") ;
	       usage() ;
	    }
            (void) strcpy (infile,"") ;
	    if (!strcmp (*(av+1),"-")) {
	       av++ ;
	       ac-- ;
	       /* Read file list from STDIN */
	       cp_warning ("Reading file list from STDIN") ;
	       while (fgets (line, MAX_RECORD_LENGTH, stdin)!=NULL) {
	          strip_newline (line) ;
	          (void) strcat (infile,line) ;
		  (void) strcat (infile," ") ;
	       }
	    }
	    while (ac > 1 && *(str = *(av+1)) != '-')
	    {
               (void) sprintf (infile, "%s %s", infile, *++av) ;
	       /* strcat(infile,*++av) ;
	       strcat(infile," ") ; */
	       ac--;
	    }
	    if (strlen (infile) >= MAXLENINPUT) {
	       cp_error ("input file list too long(%d characters): exiting",
	       MAXLENINPUT) ;
	       exit (1) ;
	    }
	    filetype = IS_LIST ;
            break ;
	 case 'i':
	    if (strcmp (*av, "-info") == 0) mode = INFO ;
	    else {
	       cp_error ("Argument '%s' not recognized",*av) ;
	       usage () ;
	    }
	    break ;
	 case 'd':
	    if (strcmp (*av, "-display") == 0) {
	       if (mode != INFO) mode = DISPLAY ;
	    }
	    else {
	       cp_error ("Argument '%s' not recognized",*av);
	       usage () ;
	    }
	    break ;
	 case 'o':
	    if(strcmp (*av, "-o_null")==0) {
	       if(ac <= 1) {
	          cp_error ("output null value unspecified: exiting") ;
	          usage () ;
	       }
	       (void) strcpy (OutNull,*++av) ;
	       ac-- ;
	       break ;
            }
	    if (ac <= 1 || *(str = *(av+1)) == '-') {
	       cp_error ("missing file name: exiting") ;
	       usage () ;
	    }
	    (void) strcpy (outfile, *++av) ;
            ac-- ;
            break ;
         case 'f':
	    if (strcmp (*av, "-fmt") == 0) {
	       if (ac <= 1) {
	          cp_error ("missing filter specification: exiting") ;
	          usage () ;
	       }
	       (void) strcpy (format,*++av) ;
	       ac-- ;
	       break ;
            }
	    if (strcmp (*av, "-filter") == 0) {
	       if (ac <= 1) {
	          cp_error("missing filter specification: exiting") ;
	          usage() ;
	       }
	       (void) strcpy (filter,*++av) ;
	       ac-- ;
               filter_is = ACTIVE ;
	       break ;
            }
	    if (ac <= 1) {
	       cp_error ("missing output control string: exiting") ;
	       usage () ;
	    }
	    (void) strcpy (fstring,*++av) ;
            ac-- ;
            break ;
	 case 'm':
	    if (strcmp (*av,"-maxiter") == 0) {
	       if (ac <= 1) {
	          cp_error ("max iteration no. unspecified: exiting") ;
	          usage () ;
	       }
	       max_iter = atoi (*++av) ;
	       ac-- ;
	    }
	    else {
	       cp_error ("Argument '%s' not recognized",*av);
	       usage () ;
	    }
	    break ;
	 case 'n':
	    if (strcmp (*av,"-n_min") == 0) {
	       if (ac <= 1) {
	          cp_error ("minimum number of measures for a valid result unspecified: exiting") ;
	          usage () ;
	       }
	       n_min = atoi (*++av) ;
	       ac-- ;
	       break ;
            }
	    if (strcmp (*av,"-norej") == 0) {
	       sigma = 0 ;
	       break ;
            }
	    if (strcmp (*av,"-nsig") == 0) {
	       if (ac <= 1) {
	          cp_error("rejection limit unspecified: exiting") ;
	          usage() ;
	       }
	       nsig_rej = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    if (strcmp (*av,"-null") == 0) {
               char nullstring[80] = "\0" ;
	       char ** nulls ;
	       if (ac <= 1) {
	          cp_error ("Null Values list unspecified: exiting") ;
	          usage () ;
	       }
	       (void) strcpy (nullstring, *++av) ;
	       nulls = tokenize (nullstring, ",", &N_NullVal) ;
	       if (N_NullVal > 5) {
	          cp_error ("cannot specify more than 5 null values: exiting") ;
                  (void) free_carray (nulls, N_NullVal) ;
		  goto exit ;
	       }
	       for (i = 0; i < N_NullVal; i++) NullValues[i] = atof (nulls[i]) ;
               (void) free_carray (nulls, N_NullVal) ;
	       ac-- ;
	       break ;
	    }
	    cp_error ("Argument '%s' not recognized",*av) ;
	    usage () ;
	    break ;
	 case 'v':
	    if (ac <= 1 || *(str = *(av+1)) == '-') verbose = 0 ;
	    else {
	       verbose = atoi (*++av) ;
	       ac-- ;
	    }
	    break ;
	 default:
            cp_error ("Argument '%s' not recognized",*av) ;
	    usage () ;
            break ;
	 }
      }
      else {
         cp_error ("Argument '%s' not recognized",*av) ;
	 usage () ;
      }
   }

   /* Checks some input */
   if (filetype == UNDEF) filetype = IS_COINC ;
   if (strlen (fstring)==0 && mode != INFO) {
      cp_error ("missing output control string: exiting") ;
      usage () ;
   }

   /* Keep silent if output goes on STDOUT */
   if (strlen (outfile) == 0 && mode == UNDEF) verbose = 0 ;
   if (mode == INFO) verbose = 1 ;

   /* Output some infos */
   HelloWorld () ;
   cp_message (1,1,"Current parameter settings:\n") ;
   if (filetype == IS_COINC) 
      cp_message (0,1,"Input coincidence table: %s\n", infile) ;
   else 
      cp_message (0,1,"Input files: %s\n", infile) ;
   if(mode != INFO) {
      char message[512] = "Input Null Values list: ";
      if (mode == UNDEF) 
         cp_message(0,1,"Output file: %s\n",outfile) ;
      (void) sprintf (message,"%s %g",message,NullValues[0]) ;
      for (i = 1; i < N_NullVal; i++) 
         (void) sprintf (message,"%s,%g",message,NullValues[i]) ;
      cp_message (0,1,"%s\n",message) ;
      cp_message (0,1,"Output Null Value: %s\n\n",OutNull) ;
      cp_message (0,1,"Minimum number of measures for a valid result: %d\n",n_min) ;
      if (sigma) {
	 cp_message (0,1,"Applying sigma-rejection in average computations\n") ;
	 cp_message (0,1,"Sigma-rejection limit: %g\n",nsig_rej);
	 cp_message (0,1,"Maximum number of iterations: %d\n",max_iter) ;
      }
      else
         cp_message (0,1,"Sigma-rejection in average computations: none\n") ;
   }

   /* Read or build coincidence table from listed files */
   if (filetype == IS_COINC) {
      FILE  *in ;
      char  *p1, *p2 ;
      /* Checks the existance of input file */
      if (access (infile, F_OK)) {
         cp_error ("File '%s' does not exist: aborting",infile) ;
         goto exit ;
      }
      /* Open Coinc Table */
      in = fopen (infile, "r") ;

      /* Gets file names, close input file */
      (void) fgets (line, MAX_RECORD_LENGTH, in) ;
      (void) fclose (in) ;
      if ((p1 = strstr (line,"ID")) == (char*)NULL ||
          (p2 = strstr (line,"NS")) == (char*)NULL) {
         cp_error ("Irregular coincidence table: cannot extract file names: aborting") ;
         goto exit ;
      }
      p1 += 3 ;
      *p2 = '\0' ;

      /* Maps input files */
      cp_message (1, 2, "Mapping input files\n") ;
      files = tokenize (p1, " ", &nfiles) ;
      files = (char**) realloc (files, (size_t)(nfiles+1)*sizeof(char*)) ;
      files[nfiles] = (char)NULL ;
      f_set = file_info (files) ;
      if (f_set == NULL) goto exit ;
      f_info = f_set->info ;

      /* Read coincidence table */
      cp_message (1,2,"Reading coincidence table\n") ;
      coinc = read_coinc (infile) ;
   } else {
      /* Maps input files */
      cp_message (1, 2, "Mapping input files\n") ;
      files = tokenize (infile, " ", &nfiles) ;
      files = (char**) realloc (files,(size_t)(nfiles+1)*sizeof(char*)) ;
      files[nfiles] = (char)NULL ;
      f_set = file_info (files) ;
      if (f_set == NULL) goto exit ;
      f_info = f_set->info ;

      /* Build coincidence table */
      cp_message (1, 2, "Building coincidence table\n") ;
      coinc = build_coinc () ;
   }

   if (coinc == (CoincId*)NULL ) goto exit ;
   for (i = 0; i < coinc->nfiles; i++)
      cp_message (0, 3, "File: %10s; \n", coinc->files[i]) ;
   for (j = 0; j < min (50, coinc->nrec); j++) {
      (void) sprintf (line, "  %5d", coinc->id[j]) ;
      for (i = 0; i < coinc->nfiles; i++) 
         (void) sprintf (line,"%s  %5d",line,coinc->rank[i][j]) ;
      (void) sprintf (line, "%s  %5u", line, coinc->ns[j]) ;
      if(coinc->nfiles>32) 
         (void) sprintf (line, "%s:%u", line, coinc->ns2[j]) ;
      cp_message (0,4,"%s\n",line) ;
   }

   /* Do-nothing mode: show file content, then exit */
   if (mode == INFO) {
      (void) printf ("\n") ;
      cp_message (1, 0, "*** Info mode: showing files content ***\n") ;
      for (i = 0; i < coinc->nfiles; i++) {
         if (f_info[i]->fl_lab) (void) strcpy (yynn, "Yes\0") ;
	 else (void)strcpy (yynn, "No\0") ;
         cp_message (0,0,"\nFile: %10s; \
%d header lines; %d columns; %d data rows; \
descriptors: %s\nContent:\n  ",
coinc->files[i], f_info[i]->n_head, f_info[i]->n_cols, f_info[i]->n_rows, yynn) ;
         for (j = 0; j < f_info[i]->n_cols; j++) {
  	    if (j && fmod ((double)(j),6.0) == 0.0) (void) printf ("\n  ") ;
  	    (void) printf (" %-12s", f_info[i]->labels[j]) ;
	 }
	 (void) printf ("\n") ;
	 for (j = 0; j < 80; j++) (void) printf ("_") ;
      }
      cp_message (0, 0, "\n");
      status = 0 ;
      goto exit;
   }

   /* Checks output format string */
   if (strlen (format) != 0) {
      pform = getFormat(format, &nform) ;
      if(pform == (char **)NULL) {
         cp_error("illegal format string: aborting") ;
         goto exit;
      }
      /* Counts output conversion strings */
      for(j=i=0; j<nform; j++) 
         if(strcmp(pform[j],"%%")!=0&&strncmp(pform[j],"%",(size_t)1)==0) i++ ;
   }

   /* Tokenize output control string into single-column control strings */
   f_cols = tokenize(fstring,";",&n_cols) ;
   if(strlen(format)!=0 && i!=n_cols) {
      if(i<n_cols) cp_error("too few arguments for format string: aborting") ;
      else cp_error("too many arguments for format string: aborting") ;
      goto exit;
   }

   /* Allocate memory for outcoming results */
   result = (item**) calloc( (size_t)n_cols, sizeof(item*) ) ;
   if (result == (item**)NULL) {
      cp_error("cannot alloc memory: aborting") ;
      goto exit;
   }

   /* Loop over output columns */
   cp_message(1,2,"Start Loop over output columns:\n");
   for(i=0; i<n_cols; i++) {
      cp_message(0,2,"Column # %d: '%s'\n",i+1,f_cols[i]);

      /* Pre-process input column string */
      cp_message(0,2,"Pre-processing input column string: %s\n",f_cols[i]);
      expanded = pre_process(f_cols[i], f_set) ;
      cp_message(0,3,"Expanded string: %s\n",expanded);
      if(expanded == (char*)NULL) goto exit ;

      /* Tokenize expanded string into single items */
      cp_message(0,2,"Tokenize expanded string into single items\n");
      tok_exp = tokenizeExpression(expanded) ;
      cp_message(0,3,"Tokenized expression: %s\n",tok_exp);
      free(expanded) ;

      /* Parse the expression and compute it */
      cp_message(0,2,"Parsing the expression\n");
      result[i] = ParseExpression(tok_exp, arithm) ;
      if (result[i] == NullItem) {
	 cp_error("in computation: aborting") ;
	 goto exit;
      }
      if (result[i]->t == vector) 
         for(j=0; j<min(10,coinc->nrec); j++)
	    cp_message(0,3,"result[%d]: %g; type: %c; is_in: %d\n",i,
	    result[i]->v.vector->data[j],result[i]->v.vector->type,
	    result[i]->v.vector->is_in[j]) ;
      free(tok_exp) ;
      tok_exp = (char*)NULL ;
   }

   /* Parse the logical filter */
   if(filter_is==ACTIVE) {
      cp_message(1,2,"Parsing the logical filter expression \"%s\"\n",filter) ;
      flag = ParseLogicalExpression(filter) ;
      if (flag == NullItem) {
	 cp_error("in parsing logical filter expression: aborting") ;
	 goto exit ;
      }
   }

   /* Get format strings for each column */
   /* Insert a space between each column */
   cp_message(1,2,"Get format strings for each column\n") ;
   if(pform==(char**)NULL) {
      pform = (char**)calloc((size_t)2*n_cols, sizeof(char*)) ;
      ndig = (int*)calloc((size_t)2*n_cols, sizeof(int)) ;
      for(i=0; i<n_cols; i++) {
         pform[2*i]   = (char*)calloc((size_t)25, sizeof(char)) ;
         pform[2*i+1] = (char*)calloc((size_t)2, sizeof(char)) ;
         switch(result[i]->t) {
            case number:
               switch(result[i]->v.f.type) {
                  case 'f':
                  case 'e': case 'E':
                     (void)sprintf(pform[2*i],"%%#%d.%d%c", result[i]->v.f.ndig, 
                     result[i]->v.f.ndec, result[i]->v.f.type) ;
                     break;
                  case 'g':
                  case 'd':
                     (void)sprintf(pform[2*i],"%%%dg", result[i]->v.f.ndig) ;
                     break;
                  case 'o':
                     (void)sprintf(pform[2*i],"%%%d%c", result[i]->v.f.ndig, 
                     result[i]->v.f.type) ;
                     break;
                  default:
                     ;
               }
               ndig[2*i] = result[i]->v.f.ndig ;
               break;
            case vector:
               switch(result[i]->v.vector->type) {
                  case 'f':
                     (void)sprintf(pform[2*i],"%%%d.%df", result[i]->v.vector->width+3,
  	             result[i]->v.vector->ndec) ;
                     break ;
                  case 'g':
                     (void)sprintf(pform[2*i],"%%%d.%dg", result[i]->v.vector->width+3,
  	             result[i]->v.vector->ndec) ;
                     break ;
                  case 'd':
                     (void)sprintf(pform[2*i],"%%%dd", result[i]->v.vector->width+3) ;
                     break ;
                  default:
  		     cp_error("BOH ??: %c %s",result[i]->v.vector->type,result[i]->v.vector->data[j]) ;
               }
               ndig[2*i] = result[i]->v.vector->width+3 ;
               break ;
            case string:
               (void)sprintf(pform[2*i],"%%%ds", result[i]->v.vector->width+2) ;
               ndig[2*i] = result[i]->v.vector->width ;
               break ;
            default:
               ;
         }
         (void)sprintf(pform[2*i+1]," ") ;
         ndig[2*i+1] = 0 ;
      }
      nform = 2*n_cols ;
   } else {
      /* Get number of digits for each column */
      ndig = (int*)calloc((size_t)n_cols, sizeof(int)) ;
      for(j=i=0; j<nform; j++) {
         if(strcmp(pform[j],"%%")!=0&&strncmp(pform[j],"%",(size_t)1)==0) {
            char *ptr = pform[j] ;
            while(!isdigit((unsigned int)*ptr)&&*ptr) ptr++ ;
            if(*ptr) (void)sscanf(ptr,"%d",&ndig[i++]) ;
            else ndig[i++] = 0 ;
         }
      }
   }
   for(i=0; i<nform; i++) cp_message(0,3,"Field %d: format: %7s\n",i,pform[i]) ;

   /* Output data to the proper stream */
   if( mode!=DISPLAY ) {
      /* Write results to output file or to STDOUT if not specified */
      cp_message(1,2,"Writing output file\n");
      if(strlen(outfile)!=0)
	 out = fopen(outfile, "w");
      else
	 out = stdout ;
      /* Open file for rejected objects */
      rej = fopen("CataComb.rej", "w");

      /* Put header informations on output file */
      (void) sprintf (line, "#COMMAND= %s\n", command) ;
      (void) fputs (line, out) ;
      if (filetype == IS_COINC)
	 (void) sprintf (line, "#COINCIDENCE_TABLE= %s\n", infile) ;
      else
	 (void) sprintf (line, "#COINCIDENCE_TABLE= none\n") ;
      (void) fputs (line, out) ;
      (void) strcpy (line, "#PARENT_FILES= ") ;
      for (i = 0; i < coinc->nfiles; i++)
	 (void) sprintf (line, "%s %s", line, coinc->files[i]) ;
      (void) strcat (line, "\n") ;
      (void) fputs (line, out) ;
      (void) fputs ("#\n", out) ;
   }
   else {
      /*
       * Put data in temp files for XGobi analisys
       */ 

      /* Get the temporary file name */
      proc_num = getpid();
      if(!access("/tmp", W_OK))
	 (void)sprintf(tmpfile_dat,"/tmp/CataComb%d",proc_num);
      else {
	 if(!access("/usr/tmp", W_OK))
            (void)sprintf(tmpfile_dat,"/usr/tmp/CataComb%d",proc_num);
	 else
            (void)sprintf(tmpfile_dat,"%s/CataComb%d",getenv("HOME"),proc_num);
      }
      (void)sprintf(tmpfile_col,"%s.col",tmpfile_dat);
      (void)sprintf(tmpfile_raw,"%s.raw",tmpfile_dat);
      (void)sprintf(tmpfile_X11,"%s.resources",tmpfile_dat);

      /* Open temporary output files */
      out    = fopen(tmpfile_dat, "w");
      outcol = fopen(tmpfile_col, "w");
      outraw = fopen(tmpfile_raw, "w");
      outX11 = fopen(tmpfile_X11, "w");

      /* Fill outcol file with vector labels */
      for( i=0; i<n_cols; i++ ) {
         (void)strncpy(line,f_cols[i],(size_t)22) ;
	 /*
	  * 22 is the max lenght of XGobi's labels
	  */ 
         if(strlen(f_cols[i])>=22) {
	    int    nlb = 0 ;
	    char * ptr = line ;
	    for(j=0; j<20; j++, ptr++) {
	       if(*ptr=='(') nlb++ ;
	       if(*ptr==')') nlb-- ;
            }
	    if(nlb) (void)strcpy(--ptr,"..)") ;
	    else (void)strcpy(ptr,"..") ;
	 }
	 (void)strcat(line,"\n") ;
         (void)fputs(line, outcol) ;
      }

      /* Fill X11 resource file */
      /* See /xgobi/src/xgobitop.h */
      (void)fputs("*XGobi*PlotWindow.height: 500\n", outX11) ;
      (void)fputs("*XGobi*PlotWindow.width: 500\n", outX11) ;
      (void)fputs("*XGobi*VarPanel.width:   250\n", outX11) ;
      (void)fputs("*XGobi*VarWindow.width:   28\n", outX11) ;
      (void)fputs("*XGobi*VarLabel.width:   210\n", outX11) ;
      (void)fputs("*XGobi.title: CataComb Output Data\n", outX11) ;
      (void)fputs("*defaultPrintCmd: lpr -Pps1\n", outX11) ;
      (void)fputs("*glyphType:     5\n", outX11) ;
      (void)fputs("*glyphSize:     1\n", outX11) ;
      (void)fputs("*defaultGlyph: 21\n", outX11) ;
      (void)fputs("*plotFont: -*-times-bold-r-*-*-14-*-*-*-*-*-*-*\n", outX11) ;
      (void)fputs("*font:     -*-times-medium-r-*-*-13-*-*-*-*-*-*-*\n", outX11) ;

      /* Close files */
      (void)fclose(outcol) ;
      (void)fclose(outX11) ;

      /* Set output null value to xgobi coded */
      (void)strcpy(OutNull,"NA") ;
   }

   /* Loop over output records */
   for(j=0; j<coinc->nrec; j++) {
      int is_in = 0;
      (void)strcpy(outline," \0") ;
      (void)strcpy(line,"\0") ;
      /* Loop over output format strings */
      for(i=k=0; i<nform; i++) {
         is_in = 1 ;
         strip_newline(line) ;
         (void)sprintf(fmt,"%%s%s\n",pform[i]) ;
         if(strncmp(pform[i],"%",(size_t)1)==0) {
            /* Conversion Specifications */
            char cc = *(pform[i]+strlen(pform[i])-(size_t)1) ;
            switch(cc) {
               case 'd': case 'i':
               case 'u':
               case 'o':
               case 'x': case 'X':
               case 'c':
                  /* Accepts an integer value */
                  if (result[k]->t == number)
                     (void)sprintf(outline, fmt, line, (int)result[k]->v.f.val) ;
                  else {
                     if(result[k]->v.vector->is_in[j]) {
                        is_in = 1 ;
                        (void)sprintf(outline, fmt, line, (int)result[k]->v.vector->data[j]) ;
                     }
                     else
                        (void)sprintf(outline,"%s%*s\n",line,max(ndig[k],(int)strlen(OutNull)+1),OutNull) ;
                  }
                  k++ ;
                  break ;
               case 'f':
               case 'e': case 'E':
               case 'g': case 'G':
                  /* Accepts a float or double value */
                  if (result[k]->t == number)
                     (void)sprintf(outline, fmt, line, result[k]->v.f.val) ;
                  else {
                     if(result[k]->v.vector->is_in[j]) {
                        is_in = 1 ;
                        (void)sprintf(outline, fmt, line, result[k]->v.vector->data[j]) ;
                     }
                     else
                        (void)sprintf(outline,"%s%*s\n",line,max(ndig[k],(int)strlen(OutNull)+1),OutNull) ;
                  }
                  k++ ;
                  break ;
               case 's': case 'S':
                  /* Accepts a pointer to an array of char type */
                  if (result[k]->t == string) {
                     if(result[k]->v.vector->is_in[j]) {
                        is_in = 1 ;
                        (void)sprintf(outline, fmt, line, result[k]->v.vector->lab[j]) ;
                     }
                     else
                        (void)sprintf(outline,"%s%*s\n",line,max(ndig[k],(int)strlen(OutNull)+1),OutNull) ;
                  }
                  else
                     if(j==0) cp_warning("output conversion error: skip column %d",k+1) ;
                  k++ ;
                  break ;
               case '%':
               case 'n':
                  /* No argument is converted */
                  (void)sprintf(outline, fmt, line) ;
                  break ;
               default :
                  if(j==0) cp_warning("%c conversion format unsupported: skip column %d",cc,k+1) ;
                  k++ ;
            }
         }
         else
            (void)sprintf(outline, fmt, line) ;

         (void)strcpy(line,outline) ;
      }

      /* Write output record */
      if( is_in && ( filter_is==UNDEF || (int)flag->v.vector->data[j]==1 ) ) {
	 (void)fputs(line, out) ;
         if( mode==DISPLAY ) {
            /* Fill outraw file with ID vector */
            (void)sprintf(line,"%d\n",coinc->id[j]) ;
            (void)fputs(line, outraw) ;
         }
      }
      else 
	 if( mode!=DISPLAY ) (void)fputs(line, rej) ;
   }

   /* Close files */
   (void)fclose(out) ;
   if( mode!=DISPLAY )
      (void)fclose(rej) ;
   else {
      (void)fclose(outraw) ;
      /* Launch XGobi */
      cp_message(1,1,"Launching XGobi...\n");
      (void)sprintf(line,"xgobi -mono %s\n",tmpfile_dat) ;
      (void)system(line) ;
      
      /* Destroy temp file */
      (void)remove(tmpfile_dat) ;
      (void)remove(tmpfile_col) ;
      (void)remove(tmpfile_raw) ;
      (void)remove(tmpfile_X11) ;
   }

   /* If we are here, then everything was ok */
   status = 0 ;

   exit:

   /* Free memory */

   cp_message(1,2,"Deallocating memory\n");
   if (mode != INFO) {
      (void) free_carray (pform, nform) ;
      if (ndig != (int*)NULL) free (ndig) ;
      cp_message (0, 3, "Deallocating tok_exp\n") ;
      if (tok_exp != (char*)NULL) free (tok_exp) ;
      cp_message (0, 3, "Deallocating f_cols\n") ;
      (void) free_carray (f_cols, n_cols) ;
      cp_message (0, 3, "Deallocating result[]\n") ;
      if (result != (item**)NULL) {
         for (i = n_cols-1; i >= 0; i--) 
            if (result[i] != NullItem) {
	       if (result[i]->t == vector ||
	           result[i]->t == vector_list ||
		   result[i]->t == computed_vector ||
                   result[i]->t == string ) 
		   destroy_vector( result[i]->v.vector ) ;
	       free (result[i]) ;
               result[i] = NullItem ;
	    }
         free (result) ;
         result = (item**)NULL ;
      }
      cp_message (0, 3, "Deallocating logical flag\n") ;
      if (filter_is == ACTIVE && flag != NullItem) {
         destroy_vector (flag->v.vector) ;
	 free (flag) ;
      }
   }
   cp_message (0, 3, "Deallocating f_info\n") ;
   if (f_info) {
      for (i = 0; i < f_set->n; i++)
	 (void) free_carray (f_info[i]->labels, f_info[i]->n_cols) ;
      free (f_info) ;
      free (f_set) ;
   }
   (void) free_carray (files, nfiles) ;
   cp_message (0, 3, "Deallocating coinc\n") ;
   if (coinc != (CoincId*)NULL) destroy_coinc (coinc) ;

   if (status == 0) cp_message (1, 1, "Done...\n") ;
   else cp_message (1, 1, "Execution failed...\n") ;
   return  (status) ;
}

/*---------------------------------------------------------------------------*/
