/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: cxc_decode.c
 * Author 	: Paolo Montegriffo
 * Created on	: May, 2002	
 * Language	: ANSI C
 * Description	: Decode cataxcorr arguments list
 * Version      : 2.1.6 - May 10, 2002
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "catalib.h"
#include "cataxcorr.h"
#include "cp_messages.h"
#include "platemodel.h"
#include "cxc_decode.h"

#if HAVE_GTK
#include <gtk/gtk.h>
#include "cxc_init.h"
#endif

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

extern void      gui_update        (void) ;
static def_pars *file_info_def_set (FileSet  *f_set) ;

static void      usage             () ;
static void      husage            () ;
static void      HelloWorld        () ;
static char    **arg_file_name_get (int	     *argc, 
                                    char   ***argv) ;
static void      arg_feel_mode_set (int      *argc, 
                                    char   ***argv) ;
static void      arg_fset_mpos_set (int      *argc, 
                                    char   ***argv) ;
static void      arg_fset_parm_set (int      *argc, 
                                    char   ***argv) ;
static void      arg_fset_parm_scl (int      *argc, 
                                    char   ***argv) ;

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef MIN
#define MIN(a,b) (((a)<=(b)) ? (a) : (b))
#endif
#define MAXLENINPUT 2048

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern    char      *version ;

boolean   mag_flag  = FALSE ; /* Magnitude list is specified? */
boolean   spheric   = FALSE ;
int       nrif = 0 ;
int       verbose   = 2 ;

FileSet  *f_set ;
def_pars *param     = (def_pars*)NULL ;
g_pars    gparam = {
   4,                               /* number of elements in Null values list */
   { 99.999, -99.999, 99.90, 0.000, 0.000 },              /* Null values list */
   2.0,                          /* controls selection criteria for triangles */
   "coinc.tab\0",                                 /* Output coincidence table */
   2.5,                           /* tolerance in magnitude for align section */
   2.0, 2.0, 500.0,                                 /* tolerance in magnitude */
   2.5, 2.5, 3.0,                           /* rejection limit in stdev units */
   5, 5, 1,                                           /* rejection iterations */
   0.1, 0.1, 0.1,                                           /* Merging radius */
   50,                                        /* maximum number of iterations */
   1,                                         /* maximum transformation order */
   1,                                /* Internal maximum transformation order */
   0,                                        /* feel mode: 0=RELAXED, 1=RIGID */
   0.0,0.0,0.0,0.0,0.0,                                            /* min_tol */
   MATCH,                                                        /* Find mode */
   FALSE,                                                      /* Mosaic mode */
   FALSE,                                                    /* Recenter mode */
   FALSE,                                /* Save computed model for each file */
   TRUE,                           /* Append/substitute coords in output file */
   TRUE,                          /* Out WCS format: FALSE=hh:mm:ss; TRUE=deg */
   ".rot",                         /* extension rot */
   ".plm",                         /* extension plm */
   POLY,                                   /* Fit a plate model for each file */
   { DBL_MAX, DBL_MAX },                                             /* CRVAL */
   { DBL_MAX, DBL_MAX }                                              /* CRPIX */
} ;

/*----------------------------------------------------------------------------
 *							Functions code
 *--------------------------------------------------------------------------*/
static void 
husage ()
{
   fprintf (stderr,"\n Type 'CataXcorr -h' for a brief description of command usage\n") ;
   fprintf (stderr,"\n See 'man CataXcorr' for a more detailed description of command usage\n") ;
   exit (1) ;
}

static void 
usage ()
{
   fprintf (stderr,"\nUSAGE:\n\n CataXcorr -h\n");
   fprintf (stderr,"\n CataXcorr [-x vec_list]  [-y vec_list]  [-mag vec_list]  [-null val[,...]]\n");
   fprintf (stderr,"%11s[-eps val]  [-crow_rad val]  [-tol_mag[_123] val]\n"," ");
   fprintf (stderr,"%11s[-feel  RELAXED | RIGID ]  [-mix val]  [-sr[_123] val]\n"," ");
   fprintf (stderr,"%11s[-find  MATCH  [-max_side val]  [-dm val]  [-s val,val[,...]] | ...\n"," ");
   fprintf (stderr,"%11s   ...SHIFT  [-max_shift val]  |  MOSAIC  [-max_shift val]  |  NONE ]\n"," ");
   fprintf (stderr,"%11s[-maskn  CIRCLE,x,y,r0,r1 | RECT,x0,x1,y0,y1 ]\n"," ");
   fprintf (stderr,"%11s[-fit  [ SAO | POLY | POLYRAD | NONE ]]  [-max_ord n]  [-max_iter n]\n"," ");
   fprintf (stderr,"%11s[-save]  [-v [n]]  [-nodisplay]  [-o filename]  file1 file2 [...]  \n"," "); 
   fprintf (stderr,"%11s\n"," ");

   fprintf (stderr,"\t-h:        show this help\n\n");
   fprintf (stderr,"\t-o:        output X-correlation table                 [Def: 'coinc.tab']\n");
   fprintf (stderr,"\t-x:        comma separated list of X-coordinate vector names\n");
   fprintf (stderr,"\t-y:        comma separated list of Y-coordinate vector names\n");
   fprintf (stderr,"\t-mag:      comma separated list of comparison magnitudes vector names\n");
   fprintf (stderr,"\t-null:     list of mag null values in input files  [Def: -99.999,99.999]\n");

   fprintf (stderr,"\t-save:     save plate models in output '.plm' files\n");
   fprintf (stderr,"\t-nodisplay: batch mode\n");
   fprintf (stderr,"\t-v:        verbose level                                        [Def: 2]\n\n");

   fprintf (stderr,"\t-eps:      error on coordinates measure                       [Def: 0.1]\n");
   fprintf (stderr,"\t-crow_rad: crowding radius                                    [Def: 8.0]\n") ;
   fprintf (stderr,"\t-tol_mag:  tolerance in relative magnitude differences        [Def: 1.0]\n") ;
   fprintf (stderr,"\t-mix:      merging radius for xcorrelation search             [Def: 0.7]\n");
   fprintf (stderr,"\t-nsig:     sigma rejection factor for xcorrelation search     [Def: 3.0]\n\n");

   fprintf (stderr,"\t-find:     kind of initial transformation:                  [Def: MATCH]\n");
   fprintf (stderr,"\t             MATCH:  looks for a 1st order transformation\n");
   fprintf (stderr,"\t             SHIFT:  looks for a 0th order transformation\n");
   fprintf (stderr,"\t             MOSAIC: attach each field to a border of the previous one\n");
   fprintf (stderr,"\t             NONE:   doesn't look for any transformation\n");
   fprintf (stderr,"\t-max_shift: maximum shift for SHIFT and MOSAIC modes        [Def:  50.0]\n");
   fprintf (stderr,"\t-min_side: minimum side length for triangles selection       [Def: 10.0]\n");
   fprintf (stderr,"\t-max_side: maximum side length for triangles selection      [Def: 250.0]\n");
   fprintf (stderr,"\t-dm:       tuning parameter for triangles selection           [Def: 2.0]\n");
   fprintf (stderr,"\t-s:        user specified scale ratio between frames\n\n");

   fprintf (stderr,"\t-fit:      fit fields with a given distortion function        [Def: SAO]\n");
   fprintf (stderr,"\t             SAO:     model each field with a SAO polynomial\n");
   fprintf (stderr,"\t             POLY:    model each field with a simple polynomial\n");
   fprintf (stderr,"\t             POLYRAD: polynomial with odd radial terms\n");
   fprintf (stderr,"\t             NONE:    no fit is made and '-find' flag is turn to 'NONE'\n");
   fprintf (stderr,"\t-max_iter: maximum number of iterations for plate solution     [Def: 50]\n") ;
   fprintf (stderr,"\t-max_ord:  maximum transformation order for plate model         [Def: 4]\n");

   fprintf (stderr,"*** ATTENZIONE : LISTA NON AGGIORNATA !!!\n");

   fprintf (stderr,"\n See 'man CataXcorr' for a more detailed description of command usage\n") ;
   exit (1) ;
}

static void 
HelloWorld()
{
   cp_message(0,1,"\n\
\t _______________________________________________________________ \n\
\to\t\t\t\t\t\t\t\to\n\
\to  \33[1;34mCata\33[31mXcorr\33[0m: Catalogues Cross Correlator \t\t\to\n\
\to  Version:  %s\t\t\t\to\n\
\to  Author:   P. Montegriffo - Bologna Astronomical Observatory\to\n\
\to_______________________________________________________________o\n\
\n",version) ;
}

void
cxc_arg_check_help (int argc, char **argv)
{
   int i ;
   for (i = 1; i < argc; i++) {
      if (strcmp ("-h", argv[i]) == 0 || 
          strcmp ("--help", argv[i]) == 0) 
         usage () ;
   }
   return ;
}

int
cxc_arg_check_display (int *argc, char ***argv)
{
   int status = 0 ;
   int i, j, k;
   if (argc && argv) {
      for (i = 1; i < *argc; i++) {
         if (strcmp ("-nodisplay", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            status = 1 ;
         }
      }
      for (i = 1; i < *argc; i++) {
         for (k = i; k < *argc; k++)
            if ((*argv)[k] != NULL)
	      break ;
         if (k > i) {
            k -= i ;
            for (j = i + k; j < *argc; j++)
               (*argv)[j-k] = (*argv)[j];
            *argc -= k;
         }
      }
   }
   return status ;
}

void
cxc_arg_list_decode (int ac, char **av)
{
   int       i ;
   char    **files = (char**)NULL ;
   char      line[MAX_RECORD_LENGTH] = "\n" ;

   /* Maps input files */
   files = arg_file_name_get (&ac, &av) ;
   for (i = 0; files[i]; i++)
      ;

   HelloWorld () ;
   cp_message (1, 2, "Mapping input files\n") ;
   f_set = file_info (files) ;
   gui_update () ;
   if (f_set == NULL) exit (1) ;
   if (f_set->info[0]->spheric) 
      spheric = TRUE ;

   (void) free_carray (files, i) ;

   /* Set defaults accordling to coordinate units */
   CXC_MESSAGE (0, 3, "Set defaults accordling to coordinate units") ;
   param = file_info_def_set (f_set) ;
   gparam.m1_rad = (f_set->info[nrif]->spheric) ? 0.05/3600.0 : 0.1;
 ;
   gparam.m2_rad = (f_set->info[nrif]->spheric) ? 0.05/3600.0 : 0.1;
 ;
   gparam.m3_rad = (f_set->info[nrif]->spheric) ? 0.25/3600.0 : 1.0 ;
   gparam.min_tol = (f_set->info[nrif]->spheric) ? 0.5/3600.0 : 2.0 ;
   gparam.min_tol0 = gparam.min_tol1 = gparam.min_tol2 = gparam.min_tol3 = 
         gparam.min_tol ;

   /* Process input column strings */
   CXC_MESSAGE (0, 3, "Process input column strings") ;
   arg_fset_mpos_set (&ac, &av) ;

   /* Set feel mode if given */
   CXC_MESSAGE (0, 3, "Set feel mode if given") ;
   arg_feel_mode_set (&ac, &av) ;

   /* Set params for generic files */
   CXC_MESSAGE (0, 3, "Set params for generic files") ;
   arg_fset_parm_set (&ac, &av) ;

   /* Scale params for scale ratios specifications */
   CXC_MESSAGE (0, 3, "Scale params for scale ratios specifications") ;
   arg_fset_parm_scl (&ac, &av) ;

   /* Decode remaining arguments list */
   CXC_MESSAGE (0, 3, "Decode remaining arguments list") ;
   for (av++; --ac > 0; av++) {
      char * str ;
      /* sprintf(command,"%s %s",command,*av) ; */
      if (*(str = *av) == '-') {
         char c ;
         c = *++str ;
         switch (c) {
	 case 'c':
	    if (strncmp (*av, "-crow_rad", (size_t)9) == 0) {
	       if (ac <= 1) {
	          cp_error ("crowding radius unspecified: exiting") ;
	          husage () ;
	       }
               sscanf (*av, "-crow_rad%d", &i) ;
               if (i < 1 || i > f_set->n) {
                  cp_error ("invalid option '%s': exiting", *av) ;
                  exit (1) ;
               }
               param[i-1].crow_rad = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    if (strcmp (*av,"-crval") == 0) {
	       char **tmp = (char**)NULL ;
	       int    ncrval = 0 ;
	       if (ac <= 1) {
	          cp_error ("missing CRVAL values: exiting") ;
	          husage () ;
	       }
	       tmp = tokenize (*++av,",",&ncrval) ;
	       if (ncrval == 2) {
		  gparam.CRVAL[0] = atof (tmp[0]) ;
		  gparam.CRVAL[1] = atof (tmp[1]) ;
                  (void) free_carray (tmp, ncrval) ;
                  gparam.recenter = FALSE ;
	       }
	       else {
	          cp_error ("must specify 2 comma separated CRVAL values: exiting") ;
                  (void) free_carray (tmp, ncrval) ;
	          husage () ;
	       }
               ac-- ;
	       break ;
	    }
	    if (strcmp (*av,"-crpix") == 0) {
	       char **tmp = (char**)NULL ;
	       int    ncrpix = 0 ;
	       if (ac <= 1) {
	          cp_error ("missing CRPIX values: exiting") ;
	          husage () ;
	       }
	       tmp = tokenize (*++av, ",", &ncrpix) ;
	       if (ncrpix == 2) {
		  gparam.CRPIX[0] = atof (tmp[0]) ;
		  gparam.CRPIX[1] = atof (tmp[1]) ;
                  (void) free_carray (tmp, ncrpix) ;
                  gparam.recenter = TRUE ;
	       }
	       else {
	          cp_error ("must specify 2 comma separated CRPIX values: exiting") ;
                  (void) free_carray (tmp, ncrpix) ;
	          husage () ;
	       }
               ac-- ;
	       break ;
	    }
	    else {
	       cp_error ("Argument '%s' not recognized", *av) ;
	       husage () ;
	    }
	    ac-- ;
	    break ;
	 case 'f':
	    if (strcmp (*av,"-fit") == 0) {
	       if (!strcmp (*(av+1), "NONE"))
	          gparam.fitmodel = LIN ;
	       else if (!strcmp (*(av+1), "POLY")) {
		  gparam.fitmodel = POLY ;
	       }
	       else if (!strcmp (*(av+1), "POLYRAD")) {
		  gparam.fitmodel = POLYRAD ;
	       }
	       else if (!strcmp (*(av+1), "SAO")) {
		  gparam.fitmodel = SAO ;
	       }
	       else if (!strcmp (*(av+1), "LIN")) {
		  gparam.fitmodel = LIN ;
	       }
               else if (!strcmp (*(av+1), "LEGENDRE")) {
		  gparam.fitmodel = LEGENDRE ;
	       }
	       else {
	          cp_error ("Argument '%s' not recognized",*++av) ;
		  husage () ;
	       }
	       av++ ;
	       ac-- ;
	       break ;
	    }
	    else {
	       cp_error ("Argument '%s' not recognized",*av) ;
	       husage () ;
	    }
	    ac-- ;
	    break ;
         case 'e':
	    if (strncmp (*av, "-eps", (size_t)4) == 0) {
	       if (ac <= 1) {
	          cp_error ("position error unspecified: exiting") ;
	          husage () ;
	       }
               sscanf (*av, "-eps%d", &i) ;
               if (i < 1 || i > f_set->n) {
                  cp_error ("invalid option '%s': exiting", *av) ;
                  exit (1) ;
               }
               param[i-1].eps = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    cp_error("Argument '%s' not recognized",*av);
	    husage() ;
	    break ;
	 case 'm':
	    if (strcmp (*av, "-max_ord") == 0) {
	       if (ac <= 1) {
	          cp_error ("max transformation order unspecified: exiting") ;
	          husage () ;
	       }
	       gparam.max_ord = atoi (*++av) ;
	       gparam.Imax_ord = gparam.max_ord ;
	       ac-- ;
	       break ;
	    }
	    if (strcmp (*av, "-max_iter") == 0) {
	       if (ac <= 1) {
	          cp_error ("max iteration no. unspecified: exiting") ;
	          husage () ;
	       }
	       gparam.max_iter = atoi (*++av) ;
	       ac-- ;
	       break ;
	    }
	    if (strncmp (*av, "-min_side", (size_t)9) == 0) {
	       if (ac <= 1) {
	          cp_error ("minimum side length unspecified: exiting") ;
	          husage () ;
	       }
               sscanf (*av, "-min_side%d", &i) ;
               if (i < 1 || i > f_set->n) {
                  cp_error ("invalid option '%s': exiting", *av) ;
                  exit (1) ;
               }
               param[i-1].min_side = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    if (strncmp (*av,"-max_side",(size_t)9) == 0) {
	       if (ac <= 1) {
	          cp_error ("maximum side length unspecified: exiting") ;
	          husage () ;
	       }
               sscanf (*av, "-max_side%d", &i) ;
               if (i < 1 || i > f_set->n) {
                  cp_error ("invalid option '%s': exiting", *av) ;
                  exit (1) ;
               }
               param[i-1].max_side = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    if (strncmp (*av,"-max_shift",(size_t)10) == 0) {
	       if (ac <= 1) {
	          cp_error ("max shift unspecified: exiting") ;
	          husage () ;
	       }
               sscanf (*av, "-max_shift%d", &i) ;
               if (i < 1 || i > f_set->n) {
                  cp_error ("invalid option '%s': exiting", *av) ;
                  exit (1) ;
               }
               param[i-1].max_shift = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    if (strcmp (*av,"-mix") == 0) {
	       if (ac <= 1) {
	          cp_error ("mixing radius unspecified: exiting") ;
	          husage () ;
	       }
               gparam.m3_rad = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
            if (strncmp (*av,"-mask",(size_t)5) == 0) {
	       if (ac <= 1) {
	          cp_error ("mask unspecified: exiting") ;
	          husage () ;
	       }
               if (mask_decode_list (*av, *++av))
                  exit (1) ;
	       ac-- ;
	       break ;
	    }
	    cp_error ("Argument '%s' not recognized",*av);
	    husage () ;
	    break ;
	 case 'n':
            if (strncmp (*av,"-nsr", (size_t)4) == 0) {
	       if (ac <= 1) {
	          cp_error ("SR iteration number unspecified: exiting") ;
	          husage () ;
	       }
               if (strcmp (*av,"-nsr") == 0) 
                  gparam.nsr_1 = gparam.nsr_2 = gparam.nsr_3 = atoi (*++av) ;
               if (strcmp (*av,"-nsr_1") == 0) 
	          gparam.nsr_1 = atoi (*++av) ;
               if (strcmp (*av,"-nsr_2") == 0) 
	          gparam.nsr_2 = atoi (*++av) ;
               if (strcmp (*av,"-nsr_3") == 0) 
	          gparam.nsr_3 = atoi (*++av) ;
	       ac-- ;
	       break ;
	    }
	    if (strcmp (*av,"-null")==0) {
	       char **tmp = (char**)NULL ;
	       int    n = 0 ;
	       if (ac <= 1) {
	          cp_error ("Input magnitude Null Value unspecified: exiting") ;
	          husage () ;
	       }
	       tmp = tokenize (*++av, ",", &n) ;
               if (n >= 5) cp_warning ("Cannot specify more than 5 Null Values") ;
               gparam.N_NullVal = MIN (5, n) ;
               for (i = 0; i < 5; i++)
                  if (i < n) gparam.NullValues[i] = atof (tmp[i]) ;
               else
                  gparam.NullValues[i] = 0.0 ;
               (void) free_carray (tmp, n) ;
	       ac-- ;
	       break ;
	    }
	    cp_error ("Argument '%s' not recognized", *av) ;
	    husage () ;
	    break ;
	 case 't':
	    if (strncmp (*av, "-tol_mag", (size_t)8) == 0) {
	       if (ac <= 1) {
	          cp_error ("tollerance in magnitude unspecified: exiting") ;
	          husage () ;
	       }
               if (strcmp (*av, "-tol_mag") == 0)
	          gparam.t0_mag = gparam.t1_mag = gparam.t2_mag = atof (*++av) ;
               if (strcmp (*av, "-tol_mag_0") == 0)
	          gparam.t0_mag = atof (*++av) ;
               if (strcmp (*av, "-tol_mag_1") == 0)
	          gparam.t1_mag = atof (*++av) ;
               if (strcmp (*av, "-tol_mag_2") == 0)
	          gparam.t2_mag = atof (*++av) ;
               if (strcmp (*av, "-tol_mag_3") == 0)
	          gparam.t3_mag = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    cp_error ("Argument '%s' not recognized",*av) ;
	    husage () ;
	    break ;
	 case 'd':
	    if (strcmp (*av,"-dm") == 0) {
	       if (ac <= 1) {
	          cp_error ("tollerance in magnitude unspecified: exiting") ;
	          husage () ;
	       }
	       gparam.tune = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    cp_error ("Argument '%s' not recognized",*av) ;
	    husage () ;
	    break ;
	 case 's':
	    if (strcmp (*av,"-save") == 0) {
	       gparam.savetrans = TRUE ;
	       break ;
	    }
	    if (strncmp (*av,"-scale",(size_t)6) == 0) {
	       if (ac <= 1) {
	          cp_error ("scale unspecified: exiting") ;
	          husage () ;
	       }
               sscanf (*av, "-scale%d", &i) ;
               if (i < 1 || i > f_set->n) {
                  cp_error ("invalid option '%s': exiting", *av) ;
                  exit (1) ;
               }
               param[i-1].scale = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
	    if (strncmp (*av, "-sr", (size_t)3) == 0) {
	       if (ac <= 1) {
	          cp_error ("rejection limit unspecified: exiting") ;
	          husage () ;
	       }
               if (strcmp (*av, "-sr") == 0) 
	          gparam.sr_3 = atof(*++av) ;
               if (strcmp (*av, "-sr_1") == 0) 
	          gparam.sr_1 = atof(*++av) ;
               if (strcmp (*av, "-sr_2") == 0) 
	          gparam.sr_2 = atof (*++av) ;
               if (strcmp (*av, "-sr_3") == 0) 
	          gparam.sr_3 = atof (*++av) ;
               if (strcmp (*av, "-sr_12") == 0) 
	          gparam.sr_1 = gparam.sr_2 = atof (*++av) ;
	       ac-- ;
	       break ;
	    }
            if (strcmp (*av,"-s") == 0) {
               char **tmp = (char**)NULL ;
               int    n = 0 ;
               tmp = tokenize (*++av,",",&n) ;
               if (n != f_set->n) {
                  cp_error ("scale ratio factors numbers differs from files number: exiting") ;
                  exit (1) ;
               }
               for (i = 0; i < n; i++) param[i].scale = atof (tmp[i]) ;
               (void) free_carray (tmp, n) ;
               ac-- ;
               break ;
            }
	    cp_error ("Argument '%s' not recognized",*av) ;
	    husage () ;
	    break ;
	 default:
            cp_error ("Argument '%s' not recognized",*av) ;
	    husage () ;
            break ;
	 }
      }
      else {
         if (ac > 0 && *(str = *av) != '-') {
            cp_error ("Argument '%s' not recognized",*av) ;
	    husage () ;
         }
         ac++ ;
         av-- ;
      }
   }

   /* Output some infos */
   cp_message (1, 1, "Current parameter settings:\n");
   cp_message (0, 1, "Output file: %s\n", gparam.outfile) ;
   strcpy (line," ") ;
   sprintf (line, "%g", gparam.NullValues[0]) ;
   for (i = 1; i < gparam.N_NullVal; i++) 
      sprintf (line, "%s,%g", line, gparam.NullValues[i]) ;
   cp_message (0, 1, "Input Null Values: %s\n",line) ;

   if (gparam.find == MATCH)
      cp_message (0, 1, "Search for starting 1st order transformation\n") ;
   else if (gparam.find == SHIFT)
      cp_message (0, 1, "Search for starting 0th order transformation\n") ;
   else
      cp_message (0, 1, "Search for starting transformation: NONE\n") ;

   cp_message(0,1,"Sigma-rejection limit: %g\n", gparam.sr_3) ;
   cp_message(0,1,"Maximum number of iterations: %d\n", gparam.max_iter) ;

   if (gparam.mosaic == TRUE && gparam.max_ord > 1) {
      cp_warning ("Selected MOSAIC mode with a transformation order %d", gparam.max_ord) ;
      cp_warning ("You may want to decrease transformation order") ;
   }

   return ;
}


/*-----------------------------------------------------------------------------
 * Function  : arg_file_name_get
 * In        : standard argument list
 * Out       : NULL terminated file names array
 * Purpouse  : 
 * Notes     : file names are stripped from input argument list
 *             Intercept also varbose level
 *             
 *---------------------------------------------------------------------------*/

static char **
arg_file_name_get (int *argc, char ***argv)
{
   char **ret = (char**)NULL ;
   char   infile[MAXLENINPUT] = "\0" ;
   char   line[MAX_RECORD_LENGTH] = "\n" ;
   int    nf ;
   if (argc && argv) {
      int i, j, k;
      for (i = 1; i < *argc;) {
         if (strcmp ("-l", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            if (strcmp ("-", (*argv)[i+1]) == 0) {
               i += 1 ;
               (*argv)[i] = NULL ;
               /* Read file list from STDIN */
	       cp_warning ("Reading file list from STDIN") ;
	       while (fgets (line, MAX_RECORD_LENGTH, stdin)!=NULL) {
	          strip_newline (line) ;
	          strcat (infile, line) ;
		  strcat (infile, " ") ;
	       }
            }
         }
         else if (strcmp ("-o", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i==*argc || !strncmp ("-", (*argv)[i], (size_t)1)) {
               cp_error ("missing output file name: exiting") ;
               exit (1) ;
            }
	    strcpy (gparam.outfile, (*argv)[i]) ;
            (*argv)[i] = NULL ;
         }
         else if (!access ((*argv)[i], F_OK)) {
            strcat (infile, (*argv)[i]) ;
	    strcat (infile, " ") ;
            (*argv)[i] = NULL ;
         }
         i += 1 ;
      }
      for (i = 1; i < *argc; i++) {
         for (k = i; k < *argc; k++)
            if ((*argv)[k] != NULL)
	      break ;
         if (k > i) {
            k -= i ;
            for (j = i + k; j < *argc; j++)
               (*argv)[j-k] = (*argv)[j];
            *argc -= k;
         }
      }
      /* Intercept verbose level */
      for (i = 1; i < *argc;) {
         if (strcmp ("-v", (*argv)[i]) == 0 ||
             strcmp ("--verbose", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
	    if (i+1 < *argc) {
               if (is_integer ((*argv)[i+1]) == 1) {
                  i += 1 ;
	          verbose = atoi ((*argv)[i]) ;
                  (*argv)[i] = NULL ;
               }
               else verbose = 0 ;
	    }
            else
               verbose = 0 ;
         }
         i += 1 ;
      }
      for (i = 1; i < *argc; i++) {
         for (k = i; k < *argc; k++)
            if ((*argv)[k] != NULL)
	      break ;
         if (k > i) {
            k -= i ;
            for (j = i + k; j < *argc; j++)
               (*argv)[j-k] = (*argv)[j];
            *argc -= k;
         }
      }
   }
   if (strlen (infile) >= MAXLENINPUT) {
      cp_error ("input file list too long(%d characters): exiting",
      MAXLENINPUT) ;
      exit (1) ;
   }
   if (strlen (infile) == 0) {
      cp_error ("missing input file list: exiting") ;
      exit (1) ;
   }
   ret = tokenize (infile," ",&nf) ;
   ret = (char**) realloc (ret,(size_t)(nf+1)*sizeof(char*)) ;
   ret[nf] = (char)NULL ;

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : arg_fset_mpos_set
 * In        : standard argument list
 * Out       : 
 * Purpouse  : Process input column strings and set value in f_info structure
 * Notes     : input column strings args are stripped from input argument list
 *             
 *---------------------------------------------------------------------------*/

static void
arg_fset_mpos_set (int *argc, char ***argv)
{
   char      xlist[1024] = "\0" ;
   char      ylist[1024] = "\0" ;
   char      mlist[1024] = "\0" ;
   if (argc && argv) {
      int i, j, k;
      for (i = 1; i < *argc;) {
         if (strcmp ("-mag", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i==*argc) {
               cp_error("comparison magnitudes unspecified: exiting") ;
               exit (1) ;
            }
	    strcpy(mlist, (*argv)[i]) ;
	    mag_flag = TRUE ;
            (*argv)[i] = NULL ;
         }
         else if (strcmp ("-x", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i==*argc) {
               cp_error("missing X coordinate list: exiting") ;
               exit (1) ;
            }
	    strcpy(xlist, (*argv)[i]) ;
            (*argv)[i] = NULL ;
         }
         else if (strcmp ("-y", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i==*argc) {
               cp_error("missing Y coordinate list: exiting") ;
               exit (1) ;
            }
	    strcpy(ylist, (*argv)[i]) ;
            (*argv)[i] = NULL ;
         }
         i += 1 ;
      }
      for (i = 1; i < *argc; i++) {
         for (k = i; k < *argc; k++)
            if ((*argv)[k] != NULL)
	      break ;
         if (k > i) {
            k -= i ;
            for (j = i + k; j < *argc; j++)
               (*argv)[j-k] = (*argv)[j];
            *argc -= k;
         }
      }
   }
   if (file_info_col_set (f_set, mlist, xlist, ylist)) exit (1) ;
   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : arg_feel_mode_set
 * In        : standard argument list
 * Out       : 
 * Purpouse  : set default values depending on feel and find modes
 * Notes     : feel and find args are stripped from input argument list
 *             
 *---------------------------------------------------------------------------*/

static void
arg_feel_mode_set(int *argc, char ***argv)
{
   if (argc && argv) {
      int i, j, k;
      for (i = 1; i < *argc;) {
         if (strcmp ("-feel", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i==*argc) {
               cp_error("missing -feel argument: exiting") ;
               exit (1) ;
            }
	    if ( strcmp((*argv)[i], "R") == 0 ||
                 strcmp((*argv)[i], "RE") == 0 ||
                 strcmp((*argv)[i], "REL") == 0 ||
                 strcmp((*argv)[i], "RELA") == 0 ||
                 strcmp((*argv)[i], "RELAX") == 0 ||
                 strcmp((*argv)[i], "RELAXE") == 0 ||
                 strcmp((*argv)[i], "RELAXED") == 0 ) {
               gparam.sr_3 = 3.5 ;
               gparam.sr_2 = 2.5 ;
               gparam.sr_1 = 2.5 ;
               gparam.nsr_3 = 1 ;
               gparam.nsr_2 = 5 ;
               gparam.nsr_1 = 5 ;
               gparam.t1_mag = gparam.t2_mag = 2.0 ;
               gparam.feel = 0 ;
            }
	    else if ( strcmp((*argv)[i], "R") == 0 ||
                      strcmp((*argv)[i], "RI") == 0 ||
                      strcmp((*argv)[i], "RIG") == 0 ||
                      strcmp((*argv)[i], "RIGI") == 0 ||
                      strcmp((*argv)[i], "RIGID") == 0 ) {
               gparam.sr_3 = 2.5 ;
               gparam.sr_2 = 1.8 ;
               gparam.sr_1 = 1.8 ;
               gparam.nsr_3 = 3 ;
               gparam.nsr_2 = 10 ;
               gparam.nsr_1 = 20 ;
               gparam.t1_mag = gparam.t2_mag = gparam.t3_mag ;
               gparam.feel = 1 ;
            }
	    else {
	       cp_error("Argument '%s' not recognized: exiting", (*argv)[i]) ;
	       exit (1) ;
	    }
            (*argv)[i] = NULL ;
         }
         else if (strcmp ("-find", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i==*argc) {
               cp_error("missing -find argument: exiting") ;
               exit (1) ;
            }
            gparam.max_ord = 5 ;
	    if ( strcmp((*argv)[i], "MA") == 0 ||
                 strcmp((*argv)[i], "MAT") == 0 ||
                 strcmp((*argv)[i], "MATC") == 0 ||
                 strcmp((*argv)[i], "MATCH") == 0 ) {
	       gparam.find = MATCH ;
               gparam.mosaic = FALSE ;
            }
	    else if ( strcmp((*argv)[i], "S") == 0 ||
                      strcmp((*argv)[i], "SH") == 0 ||
                      strcmp((*argv)[i], "SHI") == 0 ||
                      strcmp((*argv)[i], "SHIF") == 0 ||
                      strcmp((*argv)[i], "SHIFT") == 0 ) {
	       gparam.find = SHIFT ;
               gparam.mosaic = FALSE ;
            }
            else if ( strcmp((*argv)[i], "MO") == 0 ||
                      strcmp((*argv)[i], "MOS") == 0 ||
                      strcmp((*argv)[i], "MOSA") == 0 ||
                      strcmp((*argv)[i], "MOSAI") == 0 ||
                      strcmp((*argv)[i], "MOSAIC") == 0 ) {
	       gparam.find = SHIFT ;
               gparam.mosaic = TRUE ;
               gparam.max_ord = 1 ;
            }
	    else if ( strcmp((*argv)[i], "N") == 0 ||
                      strcmp((*argv)[i], "NO") == 0 ||
                      strcmp((*argv)[i], "NON") == 0 ||
                      strcmp((*argv)[i], "NONE") == 0 )
	       gparam.find = NONE ;
	    else {
	       cp_error("Argument '%s' not recognized: exiting", (*argv)[i]) ;
	       exit (1) ;
	    }
            (*argv)[i] = NULL ;
         }
         i += 1 ;
      }
      for (i = 1; i < *argc; i++) {
         for (k = i; k < *argc; k++)
            if ((*argv)[k] != NULL)
	      break ;
         if (k > i) {
            k -= i ;
            for (j = i + k; j < *argc; j++)
               (*argv)[j-k] = (*argv)[j];
            *argc -= k;
         }
      }
   }
   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : arg_fset_parm_set
 * In        : 
 * Out       : 
 * Purpouse  : Set params for generic files
 * Note      : 
 *             
 *---------------------------------------------------------------------------*/

static void
arg_fset_parm_set (int *argc, char ***argv) 
{
   int l, sph = 0 ;
   for (l = 0; l < f_set->n; l++) 
      if (f_set->info[l]->spheric!=TRUE) sph = 1 ;

   if (argc && argv) {
      int i, j, k;
      for (i = 1; i < *argc;) {
         if (strcmp ("-eps", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i==*argc) {
               cp_error ("missing -eps argument: exiting") ;
               exit (1) ;
            }
            for (j = 0; j < f_set->n; j++) 
               if (f_set->info[j]->spheric == FALSE || sph == 0)
                  param[j].eps = atof ((*argv)[i]) ;
            (*argv)[i] = NULL ;
         }
         else if (strcmp ("-crow_rad", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i == *argc) {
               cp_error ("missing -crow_rad argument: exiting") ;
               exit (1) ;
            }
            for (j = 0; j < f_set->n; j++) 
               if (f_set->info[j]->spheric == FALSE || sph == 0)
                  param[j].crow_rad = atof ((*argv)[i]) ;
            (*argv)[i] = NULL ;
         }
         else if (strcmp ("-min_side", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i == *argc) {
               cp_error ("missing -min_side argument: exiting") ;
               exit (1) ;
            }
            for (j = 0; j < f_set->n; j++) 
               if (f_set->info[j]->spheric == FALSE || sph == 0)
                  param[j].min_side = atof ((*argv)[i]) ;
            (*argv)[i] = NULL ;
            if (gparam.find != MATCH) 
               cp_warning ("-min_side usefull only in MATCH mode") ; 
         }
         else if (strcmp ("-max_side", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i == *argc) {
               cp_error ("missing -max_side argument: exiting") ;
               exit (1) ;
            }
            for (j = 0; j < f_set->n; j++) {
               if (f_set->info[j]->spheric == FALSE || sph == 0)
                  param[j].max_side = atof ((*argv)[i]) ;
	    }
            (*argv)[i] = NULL ;
            if (gparam.find != MATCH) 
               cp_warning ("-max_side usefull only in MATCH mode") ; 
         }
         else if (strcmp ("-max_shift", (*argv)[i]) == 0) {
            (*argv)[i] = NULL ;
            i += 1 ;
            if (i == *argc) {
               cp_error ("missing -max_shift argument: exiting") ;
               exit (1) ;
            }
            for (j = 0; j < f_set->n; j++) 
               if (f_set->info[j]->spheric == FALSE || sph == 0)
                  param[j].max_shift = atof ((*argv)[i]) ;
            (*argv)[i] = NULL ;
            if (gparam.find != SHIFT) 
               cp_warning ("-max_shift usefull only in SHIFT OR MOSAIC mode") ; 
         }
         i += 1 ;
      }
      for (i = 1; i < *argc; i++) {
         for (k = i; k < *argc; k++)
            if ((*argv)[k] != NULL)
	      break ;
         if (k > i) {
            k -= i ;
            for (j = i + k; j < *argc; j++)
               (*argv)[j-k] = (*argv)[j];
            *argc -= k;
         }
      }
   }
   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : arg_fset_parm_scl
 * In        : 
 * Out       : 
 * Purpouse  : Scale params value accordingly to scale ratio specification
 * Note      : 
 *             
 *---------------------------------------------------------------------------*/

static void
arg_fset_parm_scl (int *argc, char ***argv) 
{
   if (argc && argv) {
      int i, j, k, scale = 0 ;
      for (i = 1; i < *argc;) {
	 if (strncmp ("-scale", (*argv)[i], (size_t)6) == 0) {
            scale = 1 ;
            sscanf ((*argv)[i], "-scale%d", &j) ;
            if (j < 1 || j > f_set->n) {
               cp_error ("invalid option '%s': exiting", (*argv)[i]) ;
               exit (1) ;
            }
            if (i == *argc-1) {
               cp_error ("missing '%s' argument: exiting", (*argv)[i]) ;
               exit (1) ;
            }
            (*argv)[i] = NULL ;
            i += 1 ;
            param[j-1].scale = atof ((*argv)[i]) ;
            (*argv)[i] = NULL ;
         }
         i += 1 ;
      }
      for (i = 1; i < *argc; i++) {
         for (k = i; k < *argc; k++)
            if ((*argv)[k] != NULL)
	      break ;
         if (k > i) {
            k -= i ;
            for (j = i + k; j < *argc; j++)
               (*argv)[j-k] = (*argv)[j];
            *argc -= k;
         }
      }
      /* Adiust params sets for given scale ratio */
      if (scale) {
         for (i = 0; i < f_set->n; i++)
            if (param[i].scale == 1) break ;
         for (j = 0; j < f_set->n; j++) {
            param[j].eps =      param[j].scale * param[i].eps ;
            param[j].crow_rad = param[j].scale * param[i].crow_rad ;
            param[j].min_side = param[j].scale * param[i].min_side ;
            param[j].max_side = param[j].scale * param[i].max_side ;
         }
      }
   }
   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : file_info_def_set
 * In        : 
 * Out       : 
 * Purpouse  : Set default params value accordingly to reference frame units
 * Note      : 
 *             
 *---------------------------------------------------------------------------*/

static def_pars  *
file_info_def_set (FileSet *f_set)
{
   int        i ;
   def_pars  *ret = (def_pars*) malloc ((size_t)f_set->n * sizeof(def_pars)) ;

   if (ret == (def_pars*)NULL) 
      return ret ;
   for (i = 0; i < f_set->n; i++) {
      if (f_set->info[i]->spheric == TRUE) {
         ret[i].eps = 0.05/3600.0 ;
	 ret[i].m1_rad = 0.05/3600.0 ;
	 ret[i].m2_rad = 0.05/3600.0 ;
	 ret[i].m3_rad = 0.25/3600.0 ;
         ret[i].crow_rad = 2.5/3600.0 ;
         ret[i].min_side = 0.0 ;
         ret[i].max_side =  70/3600.0 ;
         ret[i].max_shift = 10/3600.0 ;
      }
      else {
         ret[i].eps = 0.1 ;
	 ret[i].m1_rad = 0.1 ;
	 ret[i].m2_rad = 0.1 ;
	 ret[i].m3_rad = 0.5 ;
         ret[i].crow_rad = 10.0 ;
         ret[i].min_side = 0.0 ;
         ret[i].max_side = 200.0 ;
         ret[i].max_shift = 50.0 ;
      }
      ret[i].scale = 1.0 ;
      ret[i].m0 = -1.0e10 ;
      ret[i].m1 = 1.0e10 ;
      strcpy(ret[i].mask, "") ;
      ret[i].mask_shape = SHAPE ;
      ret[i].mask_p[0] = ret[i].mask_p[1] = ret[i].mask_p[2] = ret[i].mask_p[3] = 0.0 ;
   }
   return ret ;
}

def_pars  *
file_info_set_default (boolean spheric)
{
   int        i ;
   def_pars  *ret = (def_pars*) malloc (sizeof(def_pars)) ;

   if (ret == NULL) 
      return ret ;
   if (spheric == TRUE) {
      ret->eps = 0.05/3600.0 ;
      ret->m1_rad = 0.05/3600.0 ;
      ret->m2_rad = 0.05/3600.0 ;
      ret->m3_rad = 0.25/3600.0 ;
      ret->crow_rad = 2.5/3600.0 ;
      ret->min_side = 0.0 ;
      ret->max_side =  70/3600.0 ;
      ret->max_shift = 10/3600.0 ;
   }
   else {
      ret->eps = 0.1 ;
      ret->m1_rad = 0.1 ;
      ret->m2_rad = 0.1 ;
      ret->m3_rad = 0.5 ;
      ret->crow_rad = 10.0 ;
      ret->min_side = 0.0 ;
      ret->max_side = 200.0 ;
      ret->max_shift = 50.0 ;
   }
   ret->scale = 1.0 ;
   ret->m0 = -1.0e10 ;
   ret->m1 = 1.0e10 ;
   strcpy (ret->mask, "") ;
   ret->mask_shape = SHAPE ;
   ret->mask_p[0] = ret->mask_p[1] = ret->mask_p[2] = ret->mask_p[3] = 0.0 ;
   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : mask_decode_list
 * In        : 
 * Out       : 
 * Purpouse  : 
 * Note      : 
 *             
 *---------------------------------------------------------------------------*/

int
mask_decode_list (char *par1, char *par2)
{
   int        i, j, n ;
   char     **tok = (char**)NULL ;
   shape      s ;

   sscanf (par1, "-mask%d", &i) ;
   if (i < 1 || i > f_set->n) {
      cp_error ("invalid option '%s': exiting", par1) ;
      exit (1) ;
   }
   tok = tokenize (par2, ",", &n) ;
   if ( strcmp (tok[0], "C") == 0 ||
        strcmp (tok[0], "CI") == 0 ||
        strcmp (tok[0], "CIR") == 0 ||
        strcmp (tok[0], "CIRC") == 0 ||
        strcmp (tok[0], "CIRCL") == 0 ||
        strcmp (tok[0], "CIRCLE") == 0 ) 
      param[i-1].mask_shape = s = CIRCLE ;
   else if ( strcmp (tok[0], "R") == 0 ||
             strcmp (tok[0], "RE") == 0 ||
             strcmp (tok[0], "REC") == 0 ||
             strcmp (tok[0], "RECT") == 0 ||
             strcmp (tok[0], "RECTA") == 0 ||
             strcmp (tok[0], "RECTAN") == 0 ||
             strcmp (tok[0], "RECTANG") == 0 ||
             strcmp (tok[0], "RECTANGL") == 0 ||
             strcmp (tok[0], "RECTANGLE") == 0 ) 
      param[i-1].mask_shape = s = RECT ;
   else if ( strcmp (tok[0], "M") == 0 ||
             strcmp (tok[0], "MA") == 0 ||
             strcmp (tok[0], "MAG") == 0 ) 
      s = MAG ;
   else {
      cp_error ("Invalid mask shape: %s %s: aborting", par1, par2) ;
      (void) free_carray (tok, n) ;
      return (1) ;
   }
   if (param[i-1].mask_shape == CIRCLE || param[i-1].mask_shape == RECT) {
      strcpy (param[i-1].mask, par2) ;
      if (n != 5) {
         cp_error ("Invalid mask specification: %s %s: aborting", par1, par2) ;
         exit (1) ;
      }
      for (j = 0; j < 4; j++) {
         if (strcmp (tok[j+1], "*") == 0 || strcmp (tok[j+1], "-") == 0)
            param[i-1].mask_p[j] = -1.0e10 ;
         else
            sscanf (tok[j+1], "%lg", &param[i-1].mask_p[j]) ;
      }
   }
   else if (s == MAG) {
      if (n != 3) {
         cp_error ("Invalid mask specification: %s %s: aborting", par1, par2) ;
         exit (1) ;
      }
      param[i-1].m0 = -1.0e10 ;
      param[i-1].m1 =  1.0e10 ;
      if (strcmp (tok[1], "*") != 0)
         sscanf (tok[1], "%lg", &param[i-1].m0) ;
      if (strcmp (tok[2], "*") != 0)
         sscanf (tok[2], "%lg", &param[i-1].m1) ;
      /* for (j=0; j<2; j++) 
         if (strcmp(tok[j+1], "-")!=0)
            sscanf(tok[j+1], "%lg", &param[i-1].mask_p[j]) ; */
   }
   else {
      cp_error ("Invalid mask specification: %s %s: aborting", par1, par2) ;
      (void) free_carray (tok, n) ;
      return (1) ;
   }
   (void) free_carray (tok, n) ;
   return 0 ;
}

