/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: expand.c
 * Author 	: Paolo Montegriffo
 * Created on	: July, 2000	
 * Language	: ANSI C
 * Description	: Substitutes columns names to vector labels
 * Version      : 2.0 - October 1, 2000
 *                2.1.2 - September 4, 2001
 *                2.1.3 - September 7, 2001
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <ctype.h>
#include <math.h>
#include <regex.h>

#include "cp_messages.h"
#include "fileinfo.h"
#include "catalib.h"

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static char   *expandLab1     (char *expr, Finfo **f_info, int n_files) ;
static char   *expandLab2     (char *expr, Finfo **f_info, int n_files) ;
static char   *expandLab3     (char *expr, Finfo **f_info, int n_files) ;
static char   *expand1        (char *expr) ;
static char   *expand2        (char *expr, int n_el) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

char *pre_process (char *arexp, FileSet *f_set)
{
   Finfo **f_info = f_set->info ;
   int     nfiles = f_set->n ;
   int     sz ;
   int     i, j ;
   char   *returning = (char*)NULL ;
   char    ret[1024] = "\0" ;
   char   *cur = (char*)NULL ;
   char   *expanded ;
   char   *truncexp = (char*)NULL ;
   int     beg ;
   int     len ;
   boolean matched ;
   regex_t re_not1, re_not2, re_not3 ;
   regex_t re_lab1, re_lab2, re_lab3, re_lst0, re_lst1, re_lst2 ;

   /*
    * Let's make use of regexp
    * we define four groups of possible objects to be recognized:
    * two groups of objects to be never expanded,
    * labeled objects and columns objects
    */

   char dont_exp1[] = "^(ID|NS)$" ;
   char dont_exp2[] = "^(AVE|WAVE|STDEV|STERR|WSTERR)\\(+$" ;
   char dont_exp3[] = "^(LOG|LOG2|LOG10|EXP|SQRT|ABS)\\(+$" ;

   char *labels ;
   char *labels1 ;
   char lista0[] = "^C[0-9]+_[0-9]+$" ;
   char lista1[] = "^C[0-9]+_\\[[0-9]+(([-]|,)[0-9]+)+(,[0-9]+([-][0-9]+)?)*\\]$" ;
   char lista2[] = "^C[0-9]+$" ;

   /* Allocate memory for and initialize Labels object */
   len = 0;
   for (i = 0; i < nfiles; i++) {
      if (f_info[i]->fl_lab) 
         for (j = 0; j < f_info[i]->n_cols; j++)
	    len += (int) strlen (f_info[i]->labels[j])+1;
   }
   if ((labels = (char*) calloc ((size_t)(len+4), sizeof(char))) == (char*)NULL)
      goto exit ;
   if ((labels1 = (char*) calloc ((size_t)(len+60), sizeof(char))) == (char*)NULL) 
      goto exit ;

   (void)strcpy(labels,"^(") ;
   for (i = 0; i < nfiles; i++) {
      if (f_info[i]->fl_lab) 
         for (j = 0; j < f_info[i]->n_cols; j++)
	    (void) sprintf (labels, "%s%s|", labels, f_info[i]->labels[j]) ;
   }
   strip_newline (labels) ;

   /*
    *    Compile rule for labels: three kind of labels:
    *       Mag_3       : column 'Mag' from file N.3
    *       Mag_[1,3-5] : list of columns 'Mag' from files N.1,3,4,5
    *       Mag         : list of columns 'Mag' from all files
    */

   (void) strcpy (labels1,"\0") ;
   (void) strncat (labels1, labels, strlen (labels)-(size_t)1) ;
   (void) strcat (labels1, ")_[0-9]+$\0") ;
   if (regcomp (&re_lab1, &labels1[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error ("cannot compile rule for 1st kind of labels: aborting") ;
      free (labels) ;
      free (labels1) ;
      return (char*)NULL ;
   }

   (void) strcpy (labels1,"\0") ;
   (void) strncat (labels1, labels, strlen (labels)-(size_t)1) ;
/*
 *    strcat(labels1,")_\\[[0-9]+(,[0-9]+)*\\]$\0") ;
 */
   (void)strcat (labels1, ")_\\[[0-9]+(([-]|,)[0-9]+)+(,[0-9]+([-][0-9]+)?)*\\]$\0") ;
   if (regcomp (&re_lab2, &labels1[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error ("cannot compile rule for 2nd kind of labels: aborting") ;
      free (labels) ;
      free (labels1) ;
      regfree (&re_lab1) ;
      return (char*)NULL ;
   }

   (void) strcpy (labels1, "\0") ;
   (void) strncat (labels1, labels, strlen (labels)-(size_t)1) ;
   (void) strcat (labels1, ")$\0") ;
   if (regcomp (&re_lab3, &labels1[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error ("cannot compile rule for 3rd kind of labels: aborting") ;
      free (labels) ;
      free (labels1) ;
      regfree (&re_lab1) ;
      regfree (&re_lab2) ;
      return (char*)NULL ;
   }

   free (labels);
   free (labels1);

   /* Compile rule for objects to not be expanded */
   if (regcomp (&re_not1, &dont_exp1[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error ("cannot compile rule for objects to not be expanded : aborting") ;
      goto exit ;
   }
   if (regcomp (&re_not2, &dont_exp2[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error ("cannot compile rule for objects to not be expanded : aborting") ;
      goto exit ;
   }
   if (regcomp (&re_not3, &dont_exp3[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) != 0) {
      cp_error ("cannot compile rule for objects to not be expanded : aborting") ;
      goto exit ;
   }

   /* Compile rule for columns objects */
   if ((regcomp (&re_lst0, &lista0[0], REG_EXTENDED|REG_NOSUB) != 0) ||
       (regcomp (&re_lst1, &lista1[0], REG_EXTENDED|REG_NOSUB) != 0) ||
       (regcomp (&re_lst2, &lista2[0], REG_EXTENDED|REG_NOSUB) != 0)) {
      cp_error ("cannot compile rule for columns objects : aborting") ;
      goto exit ;
   }

   /* allocate all work strings	*/
   sz = (int) strlen (arexp) ;
   if ((cur = (char*) calloc ((size_t)(sz+1), sizeof(char))) == (char*)NULL) 
      goto exit ;
   if ((truncexp = (char*) calloc ((size_t)(sz+1), sizeof(char))) == (char*)NULL) 
      goto exit;

   /*
    * The pattern identification is done in the following way:
    * try to match the entire string. if it matches, store it
    * as an identified token, else remove the rightmost character and
    * try again to match it. 
    */

   beg = 0 ;
   while (beg < sz) {
      (void) memset ((char*)truncexp, (int)0, (size_t)sz) ;
      (void) memcpy ((char*)truncexp, (char*)arexp+beg, (size_t)(sz-beg+1)) ;
      matched = FALSE ;
      for (i = (int)strlen (truncexp) ; i>0 ; i--) {
         (void) memset ((char*)cur, (int)0, (size_t)sz) ;
         (void) strncpy ((char*)cur, (char*)truncexp, (size_t)i) ;

	 /* Try to match objects to not be expanded */

	 if ((regexec (&re_not1, cur, (size_t) 0, NULL, 0) == 0) ||
	     (regexec (&re_not2, cur, (size_t) 0, NULL, 0) == 0) ||
	     (regexec (&re_not3, cur, (size_t) 0, NULL, 0) == 0)) {
            (void) sprintf (ret, "%s%s", ret, cur) ;
            beg += (int) strlen (cur) ;
            matched = TRUE ;
            break ;
         }

         /* Try to match the three different kind of labels objects */

         if (regexec (&re_lab1, cur, (size_t) 0, NULL, 0) == 0) {
            if ((expanded = expandLab1 (cur, f_info, nfiles)) == (char*)NULL)
	       goto exit ;
            (void) sprintf (ret, "%s%s", ret, expanded) ;
	    free (expanded) ;
            beg += (int) strlen (cur) ;
            matched = TRUE ;
            break ;
         }
	 if (regexec (&re_lab2, cur, (size_t) 0, NULL, 0) == 0) {
            if ((expanded = expandLab2 (cur, f_info, nfiles)) == (char*)NULL)
	       goto exit ;
            (void) sprintf (ret, "%s%s", ret, expanded) ;
	    free (expanded) ;
            beg += (int) strlen (cur) ;
            matched = TRUE ;
            break ;
         }
         if (regexec (&re_lab3, cur, (size_t) 0, NULL, 0) == 0) {
            if ((expanded = expandLab3 (cur, f_info, nfiles)) == (char*)NULL) 
	       goto exit ;
            (void) sprintf (ret, "%s%s", ret, expanded) ;
	    free (expanded) ;
            beg += (int) strlen (cur) ;
            matched = TRUE ;
            break ;
         }

	 /* Try to match the three different kind of generic column name */

	 if (regexec (&re_lst0, cur, (size_t) 0, NULL, 0) == 0) {
            (void)sprintf (ret, "%s%s", ret, cur) ;
            beg += (int) strlen (cur) ;
            matched = TRUE ;
            break ;
         }
         if (regexec (&re_lst1, cur, (size_t) 0, NULL, 0) == 0) {
            if ((expanded = expand1 (cur)) == (char*)NULL) goto exit ;
            (void) sprintf (ret, "%s%s", ret, expanded) ;
	    free (expanded) ;
            beg += (int) strlen (cur) ;
            matched = TRUE ;
            break ;
         }
	 if (regexec (&re_lst2, cur, (size_t) 0, NULL, 0) == 0) {
            if ((expanded = expand2 (cur, nfiles)) == (char*)NULL) goto exit;
            (void) sprintf (ret, "%s%s", ret, expanded) ;
	    free (expanded);
            beg += (int) strlen (cur) ;
            matched = TRUE ;
            break ;
         }
      }

      if (matched == FALSE) {
         /*
	 * The whole expression has been reduced to zero characters
	 * without match found: append the 1st character to ret string.
         */
	 (void) sprintf (ret, "%s%s", ret, cur) ;
         beg += (int) strlen (cur) ;
      }
   }

   /* Allocate returning array */
   if ((returning = (char*)calloc(strlen(ret)+1, sizeof(char))) == (char*)NULL)
      goto exit ;
   (void)strcpy(returning,ret);

   exit:

   /* free regexp rules	*/
   regfree(&re_not1) ;
   regfree(&re_not2) ;
   regfree(&re_not3) ;
   regfree(&re_lab1) ;
   regfree(&re_lab2) ;
   regfree(&re_lab3) ;
   regfree(&re_lst0) ;
   regfree(&re_lst1) ;
   regfree(&re_lst2) ;
   /* free memory */
   free(cur) ;
   free(truncexp) ;

   return(returning) ;
}

/*----------------------------------------------------------------------------
 * Function :  expand1
 * In       :  character string containing an implicit vectors list
 * Out      :  newly allocated character string, containing the explicit
 *             vectors list
 * Job      :  expand an implicit vector list to the explicit form
 * Notice   :  returned string has to be freed by caller
 * Example  :  ret = expand1("C3_[1,3-5,7]")
 *             ret contains then: "C3_1,C3_3,C3_4,C3_5,C3_7"
 *--------------------------------------------------------------------------*/

static char *
expand1(char *expr)
{
   char * ret = (char*)NULL ;
   char * ptr;
   int    i, len0, n_el, llen;
   int  * list = (int*)NULL ;

   /* Get number of indexes in list */
   ptr = expr;
   len0 = 0;
   while(*ptr++!='[')
      len0++;
   list = getNumbers(ptr,&n_el) ;
   if (list == (int*)NULL) {
      cp_error("illegal range in expression: aborting") ;
      return (ret) ;
   }

   /* Allocate memory for returning string */
   llen = n_el*len0+(int)strlen(expr)-1;
   if((ret = (char*)calloc((size_t)llen, sizeof(char)))==(char*)NULL) {
      return (char*)NULL ;
   }

   /* Build the returning string */
   ptr = expr+len0+1 ;
   for(i=0;i<n_el;i++) {
      (void)strncat(ret,expr,(size_t)len0) ;
      (void)sprintf(ret,"%s%d",ret,list[i]) ;
      if(i!=n_el-1) (void)strncat(ret,",",(size_t)1) ;
   }
   free(list) ;

   return (ret) ;
}

/*----------------------------------------------------------------------------
 * Function :  expand2
 * In       :  character string containing an implicit vectors list
 * Out      :  newly allocated character string, containing the explicit
 *             vectors list
 * Job      :  expand an implicit vector list to the explicit form
 * Notice   :  returned string has to be freed by caller
 * Example  :  ret = expand1("C3",4)
 *             ret contains then: "C3_1,C3_2,C3_3,C3_4"
 *--------------------------------------------------------------------------*/

static char *
expand2(char *expr, int n_el)
{
   char * ret ;
   int    i, len0 ;

   len0 = (strlen(expr)+3)*n_el-1 ;
   if(n_el>9) len0 += n_el-9 ;
   if(n_el>99) len0 += n_el-99 ;

   /* Allocate memory for returning string */
   if((ret = (char*)calloc((size_t)len0, sizeof(char)))==(char*)NULL) {
      return (char*)NULL ;
   }

   /* Build the returning string */
   for(i=1;i<n_el;i++) {
      (void)sprintf(ret,"%s%s_%d,",ret,expr,i) ;
   }
   (void)sprintf(ret,"%s%s_%d",ret,expr,n_el) ;
   
   return (ret) ;
}
/*----------------------------------------------------------------------------
 * Function : 
 * In       :  character string containing an implicit vectors list
 * Out      :  newly allocated character string, containing the explicit
 *             vectors list
 * Job      :  expand an implicit vector list to the explicit form
 * Notice   :  returned string has to be freed by caller
 * Example  :  ret = expand1("C3",4)
 *             ret contains then: "C3_1,C3_2,C3_3,C3_4"
 *--------------------------------------------------------------------------*/

static char *
expandLab1 (char *expr, Finfo **f_info, int n_files)
{
   char  * ret = (char*)NULL ;
   int     i, j ;
   char    col[10] ;
   char    fmt[10] ;
   char  * tmp ;
   boolean matched ;

   if ((tmp = (char*) calloc (strlen (expr)+1, sizeof(char))) == (char*)NULL) {
      return (char*)NULL ;
   }
   
   /* Read label and file number from input string */
   (void) sprintf (fmt, "%%%ds_%%d", (int)((long int) rindex (expr,'_') - (long int)expr)) ;
   (void) sscanf (expr, fmt, tmp, &i) ;
   i -=1 ;
   if (i >= n_files) {
      cp_error ("illegal file number in %s: aborting", expr) ;
      goto exit ;
   }

   /* Get column number */
   matched = FALSE ;
   if (f_info[i]->fl_lab) {
      for (j = 0; j < f_info[i]->n_cols; j++) {
	 if (strcmp (tmp, f_info[i]->labels[j]) == 0) {
	    (void) sprintf (col, "C%d_%d", j+1, i+1) ;
	    matched = TRUE ;
	    break ;
	 }
      }
   }
   if (matched == FALSE) {
      cp_error ("label '%s' not present in file %d: aborting",tmp,i+1) ;
      free (tmp) ;
      return (char*)NULL ;
   }
   if ((ret = (char*) calloc (strlen(col)+1, sizeof(char)))==(char*)NULL) {
      free (tmp) ;
      return (char*)NULL ;
   }
   (void) strcpy (ret,col) ;

   exit:

   free(tmp) ;
   
   return (ret) ;
}

/*----------------------------------------------------------------------------
 * Function : 
 * In       :  character string containing an implicit vectors list
 * Out      :  newly allocated character string, containing the explicit
 *             vectors list
 * Job      :  expand an implicit vector list to the explicit form
 * Notice   :  returned string has to be freed by caller
 * Example  :  ret = expandLab2("mag_[1,3-5]")
 *             ret contains then: "C3_1,C3_3,C3_4,C3_5"
 *--------------------------------------------------------------------------*/

static char *
expandLab2 (char *expr, Finfo **f_info, int n_files)
{
   char  * returning = (char*)NULL ;
   char    ret[1024] = "\0";
   int     i, j, n_el ;
   char    col[8] ;
   char    fmt[12] ;
   char  * tmp ;
   char  * tmp1 = (char*)NULL ;
   char  * ptr ;
   int     ii ;
   int   * list = (int*)NULL ;
   boolean matched ;

   if ((tmp  = (char*) calloc (strlen (expr)+(size_t)1, sizeof(char))) == (char*)NULL ||
       (tmp1 = (char*) calloc (strlen (expr)+(size_t)1, sizeof(char))) == (char*)NULL )
      goto exit ;

   /* Read label and file number from input string */
   (void) sprintf (fmt, "%%%ds_%%s", (int)((long int) rindex (expr,'_') - (long int)expr)) ;
   (void) sscanf (expr, fmt, tmp, tmp1) ;

   /* Get number of indexes in list */
   ptr = tmp1 ;
   while (*ptr++!='[')
      ;
   list = getNumbers(ptr,&n_el) ;
   if (list == (int*)NULL) {
      cp_error("illegal range in expression: aborting") ;
      goto exit ;
   }

   for (i = 0; i < n_el; i++) {

      /* Get file number */
      ii = list[i] ;
      if (ii > n_files) {
         cp_error ("illegal file number in %s: aborting",expr) ;
	 goto exit ;
      }

      /* Get column number */
      matched = FALSE ;
      if (f_info[ii-1]->fl_lab) {
	 for (j = 0; j < f_info[ii-1]->n_cols; j++) {
	    if (strcmp (tmp, f_info[ii-1]->labels[j]) == 0) {
	       (void) sprintf (col, "C%d_%d", j+1, ii) ;
	       matched = TRUE ;
	       break ;
	    }
	 }
      }
      if (matched == FALSE) {
	 cp_error ("label '%s' not present in file %d: aborting", tmp, ii) ;
	 goto exit ;
      }
      /* Append the current column name */
      (void) strcat (ret,col) ;
      (void) strncat (ret, ",", (size_t)1) ;
   }

   /* Allocate returning array */
   if ((returning = (char*) calloc (strlen (ret), sizeof(char))) == (char*)NULL)
       goto exit ;
   (void) strncat (returning, ret, strlen (ret)-1) ;

   exit:

   /* Free memory */
   if (list != (int*)NULL) free (list) ;
   free (tmp) ;
   free (tmp1) ;
   
   return (returning) ;
}

/*----------------------------------------------------------------------------
 * Function : 
 * In       :  character string containing an implicit vectors list
 * Out      :  newly allocated character string, containing the explicit
 *             vectors list
 * Job      :  expand an implicit vector list to the explicit form
 * Notice   :  returned string has to be freed by caller
 * Example  :  ret = expand1("C3",4)
 *             ret contains then: "C3_1,C3_2,C3_3,C3_4"
 *--------------------------------------------------------------------------*/

static char *
expandLab3 (char *expr, Finfo **f_info, int n_files)
{
   char  * ret ;
   int     i, j ;
   char    col[8] ;
   char  * tmp ;
   boolean matched ;

   if ((tmp = (char*) calloc ((size_t)(10*n_files), sizeof(char))) == (char*)NULL) {
      return (char*)NULL ;
   }
   for (i = 0; i < n_files; i++) {
      matched = FALSE ;
      if (f_info[i]->fl_lab) {
         for (j = 0; j < f_info[i]->n_cols; j++) {
	    if (strcmp (expr, f_info[i]->labels[j]) ==0) {
	       (void) sprintf (col, "C%d_%d", j+1, i+1) ;
	       matched = TRUE ;
	       break ;
	    }
	 }
	 if (matched == TRUE)
	    (void) sprintf (tmp, "%s%s,", tmp, col) ;
      }
   }
   if ((ret = (char*) calloc (strlen (tmp), sizeof(char))) == (char*)NULL) {
      free (tmp) ;
      return (char*)NULL ;
   }
   (void) strncat (ret, tmp, strlen (tmp)-1) ;
   free (tmp) ;
   
   return (ret) ;
}

