/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name   : file_info.c
 * Author      : Paolo Montegriffo
 * Created on  : September, 2001	
 * Language    : ANSI C
 * Description : Determines files content
 * Function    : file_info
 * In          : file names list
 * Out         : FileSet structure
 * Version     : 13-10-01: Run lint on the code
 * Version     : 2.1.9: 26/09/02: used strdup instead of strcpy for labels
 *             
 *---------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <ctype.h>
#include <regex.h>

#include "catalib.h"
#include "fileinfo.h"
#include "cp_messages.h"
#include "expand.h"


#define LINE_LEN 2048


/*----------------------------------------------------------------------------
 * 							Function prototypes	
 *--------------------------------------------------------------------------*/


/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

FileSet *
file_info (char **file_names)
{
   FileSet *ret ;
   int      i ;

   CPKDEBUG ("file_info") ;

   /* Allocate memory for output array of structures */
   ret = (FileSet*) calloc ((size_t)1 , sizeof(FileSet)) ;
   ret->n = 0 ;
   for (i = 0; file_names[i] != NULL; i++) 
      if (file_info_add (ret, file_names[i])) {
         file_info_destroy (ret) ;
         return NULL ;
      }
   return ret ;
}

int
file_info_add (FileSet *f_set, char *file_name)
{
   Finfo  *new = NULL ;
   int     j ;
   char    lab[10] ;

   /* Chech if file already in f_set ?? */

   new = file_info_new (file_name) ;
   if (new == NULL) {
      cp_warning ("file_info_add: Cannot load file '%s'", file_name) ;
      return 1 ;
   }
   if (!new->fl_lab) {
      for (j = 0; j < new->n_cols; j++) {
         (void) sprintf (lab, "C%d_%d", j+1, f_set->n+1) ;
         new->labels[j] = strdup (lab) ;
      }
   }

   f_set->n++ ;
   f_set->info = (Finfo**) realloc (f_set->info, (size_t)f_set->n * sizeof(Finfo*)) ;
   f_set->info[f_set->n-1] = new ;
   return 0 ;
}

void
file_info_remove_by_name (FileSet *f_set, char *file_name)
{
   int n ;
   /* Chech if file is in f_set */
   n = 0 ;
   while (n < f_set->n && strcmp (file_name, f_set->info[n]->filename))
      n++ ;
   printf("+++++++++++++++++++++++++++REMOVING %d\n",n) ;
   if (n < f_set->n)
      file_info_remove_by_number (f_set, n) ;
   else
      cp_warning ("file_info_remove_by_name: Invalid file '%s'", file_name) ;
}

void
file_info_remove_by_number (FileSet *f_set, int n)
{
   /* Chech if file already in f_set*/
   if (n >= f_set->n || n < 0)
      cp_warning ("file_info_remove_by_number: Invalid file '%d'", n) ;
   file_info_free (f_set->info[n]) ;

   f_set->n-- ;
   for ( ; n < f_set->n; n++)
      f_set->info[n] = f_set->info[n+1] ;
   f_set->info = (Finfo**) realloc (f_set->info, (size_t)f_set->n * sizeof(Finfo*)) ;
}

void
file_info_destroy (FileSet *f_set)
{
   int i ;
   for (i = 0; i < f_set->n; i++)
      if (f_set->info[i]) file_info_free (f_set->info[i]) ;
   if (f_set) free (f_set) ;
}

void
file_info_free (Finfo *info)
{
   int i ;
   for (i = 0; i < info->n_cols; i++) 
      if (info->labels[i]) free (info->labels[i]) ;
   if (info->labels)   free (info->labels) ;
   if (info->fmt)      free (info->fmt) ;
   if (info->filename) free (info->filename) ;
   if (info)           free (info) ;
}

Finfo *
file_info_new (char *file_name)
{
   Finfo   *f_info ;
   FILE    *in ;
   int      j, k, n, n_cols = 0 ;
   char     line[LINE_LEN], line1[LINE_LEN] ;
   char     lab[10] ;
   char   **toks ;
   int      fmt = 0, sex = 0, oldf = 0 ;
   char     inumber[] = "^([+-]?[0-9]+)$" ;
   char     rnumber[] = "^([+-]?[0-9]+[.]?[0-9]*|[+-]?[.][0-9]+)([eE][+-]?[0-9]+)?$" ;
   char     mname[] = "^((mag[_]?)?(v|i|b|u|r|j|h|k)([_]?mag)?|mag([_]?(aper|auto|best)?))$" ;
   char     xname[] = "^(x|x(_pos|_ima|pos|ima|_image|coord|_coord))$" ;
   char     yname[] = "^(y|y_pos|y_ima|ypos|yima|y_image|ycoord|y_coord)$" ;
   char     ra[] = "^(ra|ra[_]?(1950|2000))$" ;
   char     dec[] = "^(dec|dec[_]?(1950|2000))$" ;
   regex_t  re_inumb, re_rnumb, re_mname, re_xname, re_yname, re_ra, re_dec ;
   char    *p1 ;
   size_t   nc, nd ;

   CPKDEBUG ("file_info_s") ;

   /* Checks the existance of input files */
   if (access (file_name, F_OK)) {
      cp_error ("File '%s' does not exist: aborting", file_name) ;
      return NULL ;
   }

   /* Compile rules for integer and real numbers matching */
   if (regcomp (&re_inumb, &inumber[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error ("cannot compile rule for numbers: aborting") ;
      return NULL ;
   }
   if (regcomp (&re_rnumb, &rnumber[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error ("cannot compile rule for numbers: aborting") ;
      regfree (&re_inumb) ;
      return NULL ;
   }

   /* Allocate memory for output array of structures */
   f_info = (Finfo*) malloc (sizeof(Finfo)) ;

   /* Allocate memory for file name and copy it */
   f_info->filename = strdup (file_name) ;

   /* Open the file */
   in = fopen (file_name, "r");

   /* Counts file header lines and data records */
   /* White line at beginning are considered as header ones */
   f_info->n_head = 0 ;
   f_info->n_rows = 0 ;
   n = 0 ;
   while (fgets (line, LINE_LEN, in) != NULL) {
      toks = tokenize (line, " ", &n) ;
      if (!f_info->n_rows && (!n || !strncmp (toks[0], "#", (size_t)1))) 
          f_info->n_head++ ;
      else {
         /* Get number of data rows */
         if (f_info->n_rows == 0) {
            (void) strcpy (line1, line) ;
            n_cols = n ;
         }
         if (n == n_cols && strncmp (toks[0], "#", (size_t)1)) 
            f_info->n_rows++ ;
      }
      (void) free_carray (toks, n) ;
   }

   /* Get number of columns */
   toks = tokenize (line1, " ", &n) ;
   f_info->n_cols = n ;
   cp_message (1, 2, "Fileinfo: File %s: n_head %d, n_rows %d, n_cols %d\n", 
   file_name, f_info->n_head, f_info->n_rows, f_info->n_cols) ;

   /* Get data type and format for each column */
   f_info->fmt = (Dform*) calloc ((size_t)n, sizeof(Dform)) ;
   p1 = line1 ;
   for (j = 0; j < n; j++) {

      /* Uses RegExpr matching rules for data type */
      if (regexec (&re_inumb, toks[j], (size_t) 0, NULL, 0) == 0) 
         f_info->fmt[j].type = 'd' ;
      else {
         if (regexec (&re_rnumb, toks[j], (size_t) 0, NULL, 0) == 0)
            f_info->fmt[j].type = 'f' ;
         else
            f_info->fmt[j].type = 's' ;
      }

      /* Get column spacing */
      nc = (size_t)(strstr (p1, toks[j]) - p1) + strlen (toks[j]) ;
      f_info->fmt[j].nc = (int)nc ;

      /* Get number of decimals */
      f_info->fmt[j].nd = 0 ;
      if (f_info->fmt[j].type == 'f') {
         nd = strlen (toks[j]) - (size_t)(strstr (toks[j],".") - toks[j]) ;
         f_info->fmt[j].nd = (int)nd - 1 ;
      }
      p1 += nc ;
   }
   (void) free_carray (toks, n) ;

   /* Allocate memory for labels array */
   f_info->labels = (char**) calloc ((size_t)f_info->n_cols, sizeof(char*)) ;

   /* Looks for identification labels */
   rewind (in) ;
   fmt = 0 ;
   sex = -1 ;
   f_info->fl_lab = 1 ;
   for (j = 0; j < f_info->n_head; j++) {
      (void) fgets (line, LINE_LEN, in) ;
      toks = tokenize (line, " ", &n) ;
      if (!strcmp (toks[0],"#N") && n != f_info->n_cols+1)
         cp_warning ("File %s: number of labels [%d] differs from column "
               "number [%d]: skipping", file_name, n, f_info->n_cols+1) ;
      if (!strcmp (toks[0],"#N") && n == f_info->n_cols+1) {
         /* Record #N with identification labels - Highest priority */
	 fmt = 1 ;
         for (k = 0; k < f_info->n_cols; k++) 
            f_info->labels[k] = strdup (toks[k+1]) ;

	 (void) free_carray (toks, n) ;
	 break ;
      }
      (void) sprintf (lab, "%d", j+1) ;
      if (n > 1) {
	 if (!strcmp (toks[0],"#") && !strcmp (toks[1], lab) && sex) {
            if (sex < 0) sex = 1 ;
            /* Sextractor identification labels */
            f_info->labels[j] = strdup (toks[2]) ;
	    sex++ ;
	 }
	 else sex = 0 ;
      }
      (void) free_carray (toks, n) ;
   }
   if (sex > 0 && sex != f_info->n_cols+1) {
      for (k = 0; k < sex; k++)
         if (f_info->labels[k])
            free (f_info->labels[k]) ;
      sex = 0 ;
   }

   /* If no #N and Sextractor, look for generic identification labels */
   if (!fmt && sex <= 0) {
      rewind (in) ;
      for (j = 0; j < f_info->n_head; j++) {
         (void) fgets (line, LINE_LEN, in) ;
	 toks = tokenize (line, " ", &n) ;
         if (!strcmp (toks[0], "#") && n == f_info->n_cols+1) {
            fmt = 1 ;
            for (k = 0; k < f_info->n_cols; k++)
               f_info->labels[k] = strdup (toks[k+1]) ;
            (void) free_carray (toks, n) ;
            break ;
         }
         (void) free_carray (toks, n) ;
      }
   }

   /* If no id label was found, put in the standard Ci_j id string */
   if (!fmt && sex <= 0) {
      f_info->fl_lab = 0 ;
      /*** FIXME: OKKIO ***/
      /* for (j = 0; j < f_info->n_cols; j++) {
         (void) sprintf (lab, "C%d_%d", j+1, i+1) ;
         f_info->labels[j] = strdup (lab) ;
      } */
   }
   f_info->m_pos = f_info->x_pos = f_info->y_pos = -1 ;

   /* Strip newlines from labels */
   for (j = 0; j < f_info->n_cols; j++)
      strip_newline (f_info->labels[j]) ;

   /* Close the file */
   fclose (in) ;

   /* Free regexp rules	*/
   regfree (&re_inumb) ;
   regfree (&re_rnumb) ;

   /* Looks for reference mag and x&y coordinates positions using regexp rules */
   if (regcomp (&re_mname, &mname[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) == 0 &&
       regcomp (&re_xname, &xname[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) == 0 &&
       regcomp (&re_yname, &yname[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) == 0 &&
       regcomp (&re_ra,       &ra[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) == 0 &&
       regcomp (&re_dec,     &dec[0], REG_EXTENDED|REG_NOSUB|REG_ICASE) == 0 ) {

      f_info->spheric = FALSE ;
      if (f_info->fl_lab != 0) {
         for (j = 0; j < f_info->n_cols; j++) {
            if (regexec (&re_mname, f_info->labels[j], (size_t) 0, NULL, 0) == 0) {
               f_info->m_pos = j ;
               break ;
            }
         }
         for (j = 0; j < f_info->n_cols; j++) {
            if (regexec (&re_xname, f_info->labels[j], (size_t) 0, NULL, 0) == 0) 
               if (!f_info->spheric)
                  f_info->x_pos = j ;
            if (regexec (&re_yname, f_info->labels[j], (size_t) 0, NULL, 0) == 0) 
               if (!f_info->spheric)
                  f_info->y_pos = j ;
            if (regexec (&re_ra, f_info->labels[j], (size_t) 0, NULL, 0) == 0) {
               f_info->spheric = TRUE ;
               f_info->x_pos = j ;
            }
            if (regexec (&re_dec, f_info->labels[j], (size_t) 0, NULL, 0) == 0) {
               f_info->spheric = TRUE ;
               f_info->y_pos = j ;
            }
         }
         cp_message (0,2,"File %12s: Coordinates on columns %d,%d; ref mag on column %d\n",
               f_info->filename, f_info->x_pos+1, f_info->y_pos+1, f_info->m_pos+1) ;
         if (f_info->spheric) 
            cp_message (0,2,"%18s Spherical Coordinates\n", " ") ;
      }
      else {
         /* Checks for old-style format */
         if (f_info->n_cols > 5) {
            oldf = 0 ;
            for (j = 1; j < f_info->n_cols-2; j++)
               if (f_info->fmt[j].type != 'f') oldf = 1 ;
            if (f_info->fmt[0].type == 'd' &&
                f_info->fmt[f_info->n_cols-2].type == 'd' &&
                f_info->fmt[f_info->n_cols-1].type == 'd' && oldf == 0 ) {
               f_info->m_pos = 1 ;
               f_info->x_pos = f_info->n_cols-4 ;
               f_info->y_pos = f_info->n_cols-3 ;
            }
         }
      }
      /* Free regexp rules */
      regfree (&re_mname) ;
      regfree (&re_xname) ;
      regfree (&re_yname) ;
      regfree (&re_ra) ;
      regfree (&re_dec) ;
   }

   return f_info ;
}

/*-----------------------------------------------------------------------------
 * Function  : file_info_col_set
 * In        : 
 * Out       : 
 * Purpouse  : 
 * Note      : 
 *             
 *---------------------------------------------------------------------------*/

int 
file_info_col_set (FileSet *f_set, char *mlist, char *xlist, char *ylist)
{
   int      i, n_col, n_file, n_vecs, status = 0 ;
   char   **vec_name ;
   char     rule[] = "^C[0-9]+_[0-9]+$" ;
   regex_t  re_col ;
   Finfo  **f_info = f_set->info ; 
   char     *xexp = (char*)NULL ;
   char     *yexp = (char*)NULL ;
   char     *mexp = (char*)NULL ;

   /* Pre process input column strings */
   if (strlen (xlist) != 0 || strlen (ylist) != 0 || strlen (mlist) != 0 ) {
      cp_message (1,2,"Pre-processing input column strings\n") ;
      if (strlen (xlist) != 0) {
         if ((xexp = pre_process (xlist, f_set)) == (char*)NULL) goto exit ;
         cp_message (0,3,"Expanded X string: %s\n",xexp) ;
      }
      if (strlen (ylist) != 0) {
         if ((yexp = pre_process (ylist, f_set)) == (char*)NULL) goto exit ;
         cp_message (0,3,"Expanded Y string: %s\n",yexp) ;
      }
      if (strlen (mlist)!=0) {
         if ((mexp = pre_process (mlist, f_set)) == (char*)NULL) goto exit ;
         cp_message (0,3,"Expanded Mag string: %s\n",mexp) ;
      }
   }

   /*    Compile rule for column names */
   if (regcomp (&re_col, &rule[0], REG_EXTENDED|REG_NOSUB) != 0) {
      cp_error ("cannot compile rule for columns objects : aborting") ;
      return 3 ;
   }
   /* Get vector names */
   if (mexp != (char*)NULL) {
      vec_name = tokenize (mexp, ",", &n_vecs) ;
      /* Scan for column and file numbers */
      for (i = 0; i < n_vecs; i++) {
         if (regexec (&re_col, vec_name[i], (size_t) 0, NULL, 0) == 0) {
            sscanf (vec_name[i], "C%d_%d", &n_col, &n_file) ;
            f_info[n_file-1]->m_pos = n_col-1 ;
         } 
         else {
            cp_warning ("'%s' is not a regular column name: skipping", vec_name[i]) ;
            status = 1 ;
            /* goto exit ; */
         }
      }
      (void) free_carray (vec_name, n_vecs) ;
   }
   if (xexp != (char*)NULL) {
      vec_name = tokenize (xexp, ",", &n_vecs) ;
      /* Scan for column and file numbers */
      for (i = 0; i < n_vecs; i++) {
         if (regexec (&re_col, vec_name[i], (size_t) 0, NULL, 0) == 0) {
            sscanf (vec_name[i], "C%d_%d", &n_col, &n_file) ;
            f_info[n_file-1]->x_pos = n_col-1 ;
         } 
         else {
            cp_warning ("'%s' is not a regular column name: skipping", vec_name[i]) ;
            status = 1 ;
           /*  goto exit ; */
         }
      }
      (void) free_carray (vec_name, n_vecs) ;
   }
   if (yexp != (char*)NULL) {
      vec_name = tokenize (yexp, ",", &n_vecs) ;
      /* Scan for column and file numbers */
      for (i = 0; i < n_vecs; i++) {
         if (regexec (&re_col, vec_name[i], (size_t) 0, NULL, 0) == 0) {
            sscanf (vec_name[i], "C%d_%d", &n_col, &n_file) ;
            f_info[n_file-1]->y_pos = n_col-1 ;
         } 
         else {
            cp_warning ("'%s' is not a regular column name: skipping", vec_name[i]) ;
            status = 1 ;
            /* goto exit ; */
         }
      }
      (void) free_carray (vec_name, n_vecs) ;
   }

   /* Checks coordinate specifications */
   for (i = 0; i < f_set->n; i++) {
      if (f_info[i]->x_pos < 0 || f_info[i]->y_pos < 0) {
         cp_error ("List %s: missing coordinates specifications: aborting", 
                  f_set->info[i]->filename) ;
         status = 2 ;
      }
   }

   exit:
   /* free regexp rules	*/
   regfree(&re_col) ;
   if (xexp != (char*)NULL) free (xexp) ;
   if (yexp != (char*)NULL) free (yexp) ;
   if (mexp != (char*)NULL) free (mexp) ;

   return status ;
}

