/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: findmatch.c
 * Author 	: Paolo Montegriffo
 * Created on	: October, 2000	
 * Language	: ANSI C
 * Description	: Given two input star lists, looks for pairs of matching stars
 *                independently from any translation, rotation, magnification
 *                or inversion between the two coordinate system.
 *                Main parameters are:
 *                eps      : precision in coordinate measures: determines also
 *                           the tolerance for the starting cross-correlation
 *                tune     : may be used to modify the selection criteria for 
 *                           triangles
 *                scale    : user specified scale ratio between frames: it 
 *                           facilitates the process for failing situations
 *                min_side : selection criteria for triangles
 *                max_side : selection criteria for triangles
 *
 *                Based on paper "Groth,E.J. 1986, AJ, 91, 1244"
 *
 * Version      : 2.0.0 - November 9, 2000
 * Version      : 2.1.0 - July 02, 2001
 * Version      : 2.1.1 - July 09, 2001
 * Version      : 2.1.3 - October 22, 2001
 * Version      : 2.1.3 - November 2, 2001: run lint on the code
 * Version      : 2.1.4 - January 18, 2002: added support for spheric coords
 * Version      : 2.1.5 - March 8, 2002: highly optimized
 * Version      : 2.1.6 - May 16, 2002: corrected many bugs
 * Version      : 2.1.7 - Jun 13, 2002: added control for GUI
 * Version      : 2.1.13 - Feb 27, 2003: fixed bugs in triangles selection 
 * Version      : 2.1.14 - Mar 21, 2003: fixed bug in triangles match
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include <ctype.h>
#include <math.h>

#include "cataxcorr.h"
#include "cataxcorr_p.h"
#include "cp_messages.h"
#include "fitplate.h"
#include "starlist.h"

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern g_pars    gparam ;

extern int       verbose ;
extern boolean   suspend ;

def_pars *cparam ;
def_pars *l_param1, *l_param2 ;

/*----------------------------------------------------------------------------
 * 							Private Types
 *--------------------------------------------------------------------------*/

typedef enum orType_tag {
   CLOCKWISE,
   COUNTERCLOCKWISE
} orType ;

typedef struct Triangle_tag {
   int    ind[3] ;          /* index of point of vertex 1,2,3 */
   double p ;               /* logarithm of the perimeter */
   double dp ;              /* tolerance in the logarithm */
   double r ;               /* ratio longest/shortest side */
   double dr ;              /* tolerance in the ratio */
   double c ;               /* cosine of angle 1 */
   double dc ;              /* tolerance in the cosine */
   orType ot ;              /* orientation */
   int    f ;               /* flag */
} Triangle ;

typedef struct Trasf_tag {
   int    inv ;
   double xfit_coeff[3] ;
   double yfit_coeff[3] ;
   double dmag ;
} Trasf ;

typedef struct Point_tag {
   double x ;
   double y ;
   int    i ;
   int    j ;
} Point ;

typedef struct Pair_tag {
   int    i ;
   int    j ;
   double d ;
   int    f ;
} Pair ;

struct t {
   double  side ;
   Star   *s1 ;
   Star   *s2 ;
} ;

/*----------------------------------------------------------------------------
 * 							Private functions
 *--------------------------------------------------------------------------*/

static int sort_t(const struct t *d1, const struct t *d2)
{
   if ( d1->side > d2->side )
      return 1 ; 
   else
      return -1 ;
}
static int sort_tr_r(const Triangle **d1, const Triangle **d2)
{ 
   if ( (*d1)->r > (*d2)->r ) 
      return 1 ; 
   else 
      return -1 ;
}

static int sort_pair(const Pair *d1, const Pair *d2)
{
   if ( d1->d > d2->d ) 
      return 1 ; 
   else {
      if ( d1->d == d2->d )
         return ( d1->i - d2->i ) ;
      else
         return -1 ;
   }
}

static int sort_match_j(const Match *d1, const Match *d2)
{
   if ( d1->j > d2->j ) 
      return 1 ; 
   else 
      return -1 ;
}

static int sort_match_i(const Match *d1, const Match *d2)
{
   if ( d1->i > d2->i ) 
      return 1 ; 
   else {
      if ( d1->i == d2->i )
         return sort_match_j(d1, d2) ;
      else
         return -1 ;
   }
}

static int sort_match_r(const Match *d1, const Match *d2)
{
   if ( d1->f < d2->f ) 
      return 1 ; 
   else {
      if ( d1->f == d2->f ) {
         return sort_match_i(d1, d2) ;
      }
      else
         return -1 ;
   }
}

static int sort_p_y(const Point *d1, const Point *d2)
{
   if ( d1->y > d2->y ) 
      return 1 ; 
   else 
      return -1 ;
}

static int sort_star_yprj(const Star **d1, const Star **d2)
{
   if ((*d1)->yprj > (*d2)->yprj) 
      return 1 ; 
   else 
      return -1 ;
}

static int sort_star_yrot(const Star **d1, const Star **d2)
{
   if ((*d1)->yrot > (*d2)->yrot) 
      return 1 ; 
   else 
      return -1 ;
}

static int sort_star_m(const Star **d1, const Star **d2)
{ 
   if ((*d1)->mag > (*d2)->mag) 
      return 1 ;
   else {
      if ((*d1)->mag == (*d2)->mag) 
         return (sort_star_yprj(d1, d2)) ;
      else
         return -1 ;
   }
}

static void FitFunc(double x, double y, double p[], int np)
{ 
   p[0] = 1.0 ; 
   p[1] = x ; 
   p[2] = y ; 
   return ;
}

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

StarMatch        *FindMatch      (StarList *sl1, StarList *sl2, int *pnmatch) ;

static Triangle **GetTrList      (Star **si, int nstar, 
                                  int *i0, int *pntr) ;
static Triangle  *GetTriangle    (Star *s1, Star *s2, Star *s3, int *code) ;

static double     GetRmax        (Triangle **t, int ntr) ;

static Match     *SearchPairs    (Star **s1, Triangle **t1, int ntr1, 
                                  Star **s2, Triangle **t2, int ntr2, 
                                  int *pnmatch) ;
static int        purge          (Triangle **t1, Triangle **t2,   
                                  Match *pair, int nmatch, 
                                  Triangle ***ts1, Triangle ***ts2, 
                                  int *npairs) ;
static Match     *check_transf   (Star **s1, int ns1, Triangle **t1, 
                                  Star **s2, int ns2, Triangle **t2, 
                                  int npairs, int *nout) ;
static Match     *rating         (Triangle **t1, Triangle **t2, 
                                  int nmatch, int *pnr) ;
static Match     *sel_angle      (Star **s1, Triangle **t1, 
                                  Star **s2, Triangle **t2,
                                  Match *pair, int nmatch, int *pn) ;
static Match     *voting         (Triangle **t1, Triangle **t2, 
                                  Match *rates, int nr, int *pnr) ;
static Match     *pedantic       (Star **s1, int ns1, Triangle **t1, 
                                  Star **s2, int ns2, Triangle **t2, 
                                  Match *group, int ng, int *pnr) ;
static Trasf     *MatchTriangle  (Star **s1, Triangle *t1, 
                                  Star **s2, Triangle *t2) ;
static double     comp_angle     (Star **s1, Triangle *t1, 
                                  Star **s2, Triangle *t2) ;
static double     get_angle      (double   xc[3], double yc[3]) ;

static Match     *uniq           (Match *ret, int *pnr) ;

static double     get_dmag       (Star **s1, Star **s2, Match *pair, 
                                  int nmatch) ;

/*----------------------------------------------------------------------------
 * 							Defines
 *--------------------------------------------------------------------------*/

#ifndef MIN
#define MIN(a,b) (((a) <= (b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b) (((a) >= (b)) ? (a) : (b))
#endif
#ifndef ABS
#define ABS(a) ((a) < 0 ? (-(a)) : (a))
#endif
#ifndef SQR
#define SQR(x) ((x)*(x))
#endif
#ifndef PI
#define PI 3.1415926535897932384626433832795028841971693993751
#endif
#ifndef RADTODEG
#define RADTODEG(x) (180.0*(x)/(double)PI)
#endif
#ifndef HYPOT
#define HYPOT(a,b) (sqrt((a)*(a)+(b)*(b)))
#endif

#define NullTriangle (Triangle*)NULL

#define N_MAX_TR 70000         /* Maximum number of triangles for each list */

#if HAVE_GSL
#define fit gslfit
#else
#define fit qrfit
#endif
#ifndef fit
#define fit svdfit
#endif

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

StarMatch *
FindMatch(StarList *sl1, StarList *sl2, int *pnmatch)
{
   StarMatch  *ret = NULL ;
   int         nmatch = 0 ;
   Triangle  **t1 = (Triangle**)NULL ;
   Triangle  **t2 = (Triangle**)NULL ;
   Triangle  **ts1 = (Triangle**)NULL ;
   Triangle  **ts2 = (Triangle**)NULL ;
   int         i, i1, i2 ;
   Match      *pairs = (Match*)NULL ;
   Match      *p_out = (Match*)NULL ;
   Match      *group = (Match*)NULL ;
   int         npairs = 0, nout, ng ;
   int         nt1 = 0, nt2 = 0, nreq1, nreq2 ;
   int         ntr1 = 40000 ;
   int         ntr2 = 10000 ;
   double      dmag ;
   StarList   *sld1, *sld2 ;
   Star      **s1, **s2 ;
   int         ns1, ns2 ;

   l_param1 = sl1->param ;
   l_param2 = sl2->param ;

   /* Duplicate star lists */
   sld1 = star_list_dup (sl1) ;
   sld2 = star_list_dup (sl2) ;
   s1 = sld1->data ;
   s2 = sld2->data ;
   ns1 = (int)sld1->nstar ;
   ns2 = (int)sld2->nstar ;

   /* Store actual star index in flag field */
   for (i = 0; i < ns1; i++) s1[i]->index = i ;
   for (i = 0; i < ns2; i++) s2[i]->index = i ;

   /* Sort star lists for magnitudes */
   cp_message (0,3,"Sorting star lists for magnitudes\n") ;
   qsort ((void*)s1, (size_t)ns1, sizeof(Star*), 
          (int (*)(const void *, const void *))sort_star_m) ;
   qsort ((void*)s2, (size_t)ns2, sizeof(Star*), 
          (int (*)(const void *, const void *))sort_star_m) ;

   /*Loop over first list to find good triangles */
   cp_message (0,2,"Loop over first list to find good triangles\n") ;
   i1 = 0 ;
   while (npairs < 40 && i1 < 2*ns1/3) {

      nt1 += ntr1 ;
      cparam = l_param1 ;
      nreq1 = ntr1 ;
      t1 = GetTrList (s1, ns1, &i1, &ntr1) ;

      IF_ABORT_EXIT ;

      CXC_MESSAGE (0,2,"Got %d triangles from list 1...\n",ntr1) ;
      i2 = nt2 = 0 ;
      while (nt2 < nt1 && i2 < 2*ns2/3) {
         cparam = l_param2 ;
         nreq2 = ntr2 ;
         /* Loop over second list to find good triangles */
         cp_message (0,2,"Loop over second list to find good triangles\n") ;
         t2 = GetTrList (s2, ns2, &i2, &ntr2) ;

         IF_ABORT_EXIT ;

         CXC_MESSAGE (0, 2, "Got %d triangles from list 2...       "
               "covered %6d/%6d: %4d pairs... ",
               ntr2, nt2, nt1, npairs) ;

         /*Loop over triangles to find matches */
         cp_message (0,2,"Looking for matching triangles\n") ;
         pairs = SearchPairs (s1, t1, ntr1, s2, t2, ntr2, &nmatch) ;

         IF_ABORT_EXIT ;

         if (nmatch > 1) {
            /* Purge triangles lists */
            cp_message (0,2,"Purge unmatched triangles\n") ;
            (void) purge (t1, t2, pairs, nmatch, &ts1, &ts2, &npairs) ;

            /* Checks transformation */
            p_out = check_transf (s1, ns1, ts1, s2, ns2, ts2, npairs, &nout) ;
         }

         /* Free memory */
         if (pairs != (Match*)NULL) free (pairs) ;
         pairs = (Match*)NULL ;
         nmatch = 0 ;
         for (i = 0; i < ntr2; i++) 
            free (t2[i]) ;
         free (t2) ;
	 t2 = (Triangle**)NULL ;
         nt2 += ntr2 ;

         if (p_out != (Match*)NULL) 
            break ;
         if (ntr2 < nreq2)
            break ;
      }

      /* Free memory */
      for (i = 0; i < ntr1; i++) 
            free (t1[i]) ;
      free (t1) ;
      t1 = (Triangle**)NULL ;
      if (p_out != (Match*)NULL)
         break ;
      if (ntr1 < nreq1)
         break ;
   }

   if (npairs == 0) {
      cp_error ("Cannot find a good transformation: aborting") ;
      goto exit ;
   }

   if (p_out == (Match*)NULL) {
      /* Checks transformation */
      CXC_MESSAGE (2, 3, "Rating sharpest peaks\n") ;
      group = rating (ts1, ts2, npairs, &ng) ;
      if (suspend)
         goto exit ;
      if (group != (Match*)NULL) {

        /* Try pedantic solution */
         CXC_MESSAGE (2, 3, "Try a pedantic solution on given groups\n") ;
         p_out = pedantic (s1, ns1, ts1, s2, ns2, ts2, group, ng, &nout) ;
         if (suspend)
            goto exit ;
         if (nout < 0) {
            nout = -nout ;
            cp_warning ("low number of matching pairs:") ;
            cp_warning ("    check carefully the resulting match") ;
            cp_warning ("    or try to restart with higher 'eps'") ;
         }
         free (group) ;
      }
      else {
         cp_error ("Cannot find a good transformation: aborting") ;
         goto exit ;
      }
   }

   /* Calibrate 2nd star list */
   dmag = get_dmag (s1, s2, p_out, nout) ;
   CXC_MESSAGE (0,3,"Calibrating 2nd star list: adding %g\n", dmag) ;
   for (i = 0; i < ns2; i++) 
      s2[i]->mcal = s2[i]->mag + dmag ;

   /* Create returning index table */
   CXC_MESSAGE (0,2,"Alignment achieved with %d matching pairs\n", nout) ;
   ret = (StarMatch*) malloc ((size_t)nout*sizeof(StarMatch)) ;
   for (i = 0; i < nout; i++) {
      ret[i].s1 = sl1->data[s1[p_out[i].i]->index] ;
      ret[i].s2 = sl2->data[s2[p_out[i].j]->index] ;
   }
   *pnmatch = nout ;

   exit:

   /* Free memory */
   cp_message (0,6,"Deallocating memory\n") ;
   if (p_out != (Match*)NULL) free (p_out) ;
   if (pairs != (Match*)NULL) free (pairs) ;
   if (t1) {
      for (i = 0; i < ntr1; i++) 
         free (t1[i]) ;
      free (t1) ;
   }
   if (t2) {
      for (i = 0; i < ntr2; i++)
         free (t2[i]) ;
      free (t2) ;
   }
   if (ts1) {
      for (i = 0; i < npairs; i++) 
         free (ts1[i]) ;
      free (ts1) ;
   }
   if (ts2) {
      for (i = 0; i < npairs; i++) 
         free (ts2[i]) ;
      free (ts2) ;
   }
   star_list_destroy (sld1) ;
   star_list_destroy (sld2) ;
   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : GetTrList
 * In        : Star list, indexes list to be scanned
 * Out       : Triangle list, Number of returning triangles
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static Triangle **
GetTrList (Star **s, int nstar, int *i0, int *pntr)
{
   Triangle **t = (Triangle**)NULL ;
   int        i, j = 0, k = 0, code, ntr0, npair ;
   int        ntr = 0 ;
   int        npmax = 12 ; /*** old value 12 ***/
   int        ntrmax = *pntr ;
   double     tolmag = gparam.tune ;
   double     tol = cparam->max_side ;

   /* Store actual star index in flag field */
   for (i = 0; i < nstar; i++) s[i]->flag = i ;

   /*Loop over list to find good triangles */
   cp_message (0,3,"Loop over list %d %d \n", *i0, nstar);
   t = (Triangle**) malloc ((size_t)ntrmax*sizeof(Triangle*)) ;

   i = *i0 ;
   while (i < nstar-3 && ntr < ntrmax) {

      IF_ABORT_EXIT ;

      j = i+1 ;
      npair = 0 ;
      while (j < nstar-2 && 
             ntr < ntrmax &&
             npair < npmax && 
             s[j]->mag-s[i]->mag < tolmag) {

         if (ABS (s[j]->xprj-s[i]->xprj) < tol && 
             ABS (s[j]->yprj-s[i]->yprj) < tol) {

            k = j+1 ;
            ntr0 = ntr ;
            while (k < nstar-1 && 
                   ntr < ntrmax && 
                   ntr-ntr0 < npmax && 
                   s[k]->mag-s[i]->mag < tolmag) {
               t[ntr] = GetTriangle (s[i], s[j], s[k], &code) ;
	       if (t[ntr]) 
	          ntr++ ;
               else if (code == 1) break ;
               k++ ;
            }
            if (ntr - ntr0) npair++ ;
         }
         j++ ;
      }
      i++ ;
   }

#ifdef debug
   printf("Reached indexes: %d %d %d;\n",i,j,k) ;
#endif

   exit:

   /* Save returning triangles */
   *i0 = i+1 ;
   *pntr = ntr ;
   t = (Triangle**) realloc (t, (size_t)ntr*sizeof(Triangle*)) ;

   return t ;
}

/*-----------------------------------------------------------------------------
 * Function  : GetTriangle
 * In        : 3 corner stars
 * Out       : filled Triangle struct
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static Triangle *
GetTriangle (Star *s1, Star *s2, Star *s3, int *code)
{
   Triangle *ret = NullTriangle ;
   double    a2, a3 ;
   double    r, c, dx3, dx2, dy3, dy2, r2, r3, ff, eps2 ;
   double    Rmin = 1.1, Rmax = 20.0, Cmax = 0.94 ;
   struct t  t[3] ;

   *code = 2 ;

   /* Get sides length */
   t[0].side = HYPOT ((s2->xprj-s1->xprj), (s2->yprj-s1->yprj)) ;
   t[1].side = HYPOT ((s3->xprj-s2->xprj), (s3->yprj-s2->yprj)) ;
   t[2].side = HYPOT ((s3->xprj-s1->xprj), (s3->yprj-s1->yprj)) ;
   t[0].s1 = s1 ;
   t[1].s1 = s2 ;
   t[2].s1 = s1 ;
   t[0].s2 = s2 ;
   t[1].s2 = s3 ;
   t[2].s2 = s3 ;

   /* Test 1st side length */
   if (t[0].side < cparam->min_side || t[0].side > cparam->max_side) {
      *code = 1 ;
      goto exit ;
   }
   /* Sort sides for length */
   qsort ((void*)t, (size_t)3, sizeof(struct t), 
          (int (*)(const void *, const void *))sort_t) ;

   /* Test minimum side length and  Ratio longest/shortest side */
   if (t[0].side < cparam->min_side || t[2].side > cparam->max_side) 
      goto exit ;
   if (t[0].side == 0.0) {
      cp_warning ("Duplicate star\n") ;
      goto exit ;
   }

   r = t[2].side/t[0].side ;
   if (r < Rmin || r > Rmax) goto exit ;
   if (t[1].side/t[0].side < Rmin) goto exit ;

   dx2 = t[0].s2->xprj - t[0].s1->xprj ;
   dy2 = t[0].s2->yprj - t[0].s1->yprj ;
   dx3 = t[2].s2->xprj - t[2].s1->xprj ;
   dy3 = t[2].s2->yprj - t[2].s1->yprj ;
   r2 = t[0].side ;
   r3 = t[2].side ;
   c = (dx3*dx2+dy3*dy2)/(r3*r2) ;

   /* Test minimum angle at vertex 1 */
   if (c > Cmax) goto exit ;

   eps2 = cparam->eps * cparam->eps ;
   ff = 1.0/(r3*r3) - c/(r3*r2) + 1.0/(r2*r2) ;

   /* Fill returning triangle values */
   ret = (Triangle  *) malloc (sizeof(Triangle)) ;
   if (ret == NullTriangle) {
      cp_error("Cannot allocate memory for temp triangle: aborting") ;
      goto exit ;
   }
   ret->ind[0] = (t[0].s1==t[2].s1 || t[0].s1==t[2].s2) ? t[0].s1->flag : t[0].s2->flag ;
   ret->ind[1] = (t[0].s1==t[2].s1 || t[0].s1==t[2].s2) ? t[0].s2->flag : t[0].s1->flag ;
   ret->ind[2] = (t[0].s1==t[2].s1 || t[0].s2==t[2].s1) ? t[2].s2->flag : t[2].s1->flag ;

   ret->p = log10 (t[0].side + t[1].side + t[2].side) ;
   /* ret->dp = 13.86 * eps2 / ( t[0].side + t[1].side + t[2].side ) ; */
   ret->r = r ;
   ret->c = c ;

   ret->dr = sqrt ( 2.0 * r*r * eps2 * ff ) ;
   ret->dc = sqrt ( 2.0 * (1.0-c*c) * eps2 * ff + 
                    3.0 * c*c * eps2*eps2 * ff*ff ) ;

   /* Determines the triangle orientation */
   a3 = atan2 (dy3 , dx3) ;
   a2 = atan2 (dy2 , dx2) ;
   if (a3 < 0.0 && a2 > 0.0) a3 += 2.0*(double)PI ;
   if (a3 > a2 && a3-a2 < (double)PI) ret->ot = COUNTERCLOCKWISE ;
   else ret->ot = CLOCKWISE ;

   ret->f = 0 ;
   *code = 0 ;

   exit :

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : SearchPairs
 * In        : Triangles lists
 * Out       : matching pairs indexes, number of matches
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static Match *
SearchPairs (Star **s1, Triangle **t1, int ntr1, 
             Star **s2, Triangle **t2, int ntr2, int *pnret)
{
   Match   *ret = (Match*)NULL ;
   int      i, j, i1, i2, j1, j2 ;
   int      npair = 0, nret = 0, nrej = 0 ;
   double   dr, dc, drc ;
   Pair    *pair = (Pair*)NULL ;
   double   rtolmax = 0, drmax ;
   int      nexp, nexp0 ;
   double   d1, d2, lp, bin ;
   double   scale = 0.0 ;

   if (l_param1->scale != 1.0 || l_param2->scale != 1.0) 
      scale = l_param1->scale / l_param2->scale ;

   /* Sort triangles for growing ratio r */
   cp_message(0,3,"Sorting triangles for sides ratio factor\n") ;
   qsort ((void*)t1, (size_t)ntr1, sizeof(Triangle*), 
          (int (*)(const void *, const void *))sort_tr_r) ;
   qsort ((void*)t2, (size_t)ntr2, sizeof(Triangle*), 
          (int (*)(const void *, const void *))sort_tr_r) ;

   /* Initialize triangles flags: store temporarily index */
   for (i=0; i<ntr1; i++) t1[i]->f = i ;
   for (i=0; i<ntr2; i++) t2[i]->f = i ;

   /* Scan triangles lists for matching pairs */
   /* Highly optimized for speed search */
   cp_message(0,3,"Scanning triangles lists for matching pairs\n") ;
   rtolmax = GetRmax(t2, ntr2) ;
#ifdef debug
   printf("rtolmax %g\n", rtolmax) ;
#endif
   nexp = nexp0 = MIN(ntr1, ntr2) ;
   pair = (Pair*) malloc ((size_t)nexp*sizeof(Pair)) ;
   i = j1 = j2 = 0 ;
   while (i < ntr1) {

      IF_ABORT_EXIT ;

      drmax = rtolmax + t1[i]->dr ;
      if (t2[j1]->r-t1[i]->r < -drmax) 
         while (j1 < ntr2-1 && t2[j1]->r-t1[i]->r < -drmax)
	    j1++ ;
      else
         while (j1 > 0 && t2[j1]->r-t1[i]->r > -drmax)
	    j1-- ;
      if (t2[j2]->r-t1[i]->r < drmax)
         while (j2 < ntr2-1 && t2[j2]->r-t1[i]->r < drmax) 
            j2++ ;
      else
         while (j2 > 0 && t2[j2]->r-t1[i]->r > drmax) 
            j2-- ;
      for (j = j1; j <= j2; j++) {
         dr = (t2[j]->r-t1[i]->r) / (t1[i]->dr+t2[j]->dr) ;
         dc = (t2[j]->c-t1[i]->c) / (t1[i]->dc+t2[j]->dc) ;
         drc = SQR (dc) + SQR (dr) ;
         if (drc < 1.0) {
            if (npair >= nexp-10) {
	       /* Reallocate mem for current pair */
               nexp += nexp0 ;
               pair = (Pair*) realloc (pair, (size_t)nexp*sizeof(Pair)) ;
               if (pair == (Pair*)NULL) {
                  cp_error ("Could not alloc memory for matching pairs") ;
                  goto exit ;
               }
            }
            pair[npair].i = i ;
	    pair[npair].j = j ;
	    pair[npair].d = t1[i]->p - t2[j]->p ;
            pair[npair].f = 0 ;
            npair++ ;
         }
      }
      i++ ;
   }
   if (npair == 0) {
      cp_message(0,2,"No match found\n") ;
      goto exit ;
   }

   /* Sort pairs for increasing d */
   cp_message (0,3,"Sort pairs for increasing scaling\n") ;
   qsort ((void*)pair, (size_t)npair, sizeof(Pair), 
          (int (*)(const void *, const void *))sort_pair) ;

   IF_ABORT_EXIT ;

   /* Purge false matches */
   cp_message(0,3,"Purge false matches\n") ;
   bin = 0.003 ;
   for (i = 0; i < npair-1; i++) {
      j = i+1 ;
      i1 = t1[pair[i].i]->ind[0] ;
      i2 = t2[pair[i].j]->ind[0] ;
      while (j < npair-1 && pair[j].d-pair[i].d < bin) {
         j1 = t1[pair[j].i]->ind[0] ;
         j2 = t2[pair[j].j]->ind[0] ;

         d1 = HYPOT (s1[i1]->xprj-s1[j1]->xprj, s1[i1]->yprj-s1[j1]->yprj) ;
         d2 = HYPOT (s2[i2]->xprj-s2[j2]->xprj, s2[i2]->yprj-s2[j2]->yprj) ;
         lp = log10 (d1)-log10 (d2) ;

         if (ABS(pair[i].d-lp) < bin) {
            pair[i].f++ ;
            pair[j].f++ ;
         }
         j++ ;
      }
   }

#ifdef debug
   if (verbose > 8) {
      char   line[MAX_RECORD_LENGTH] = "\n" ;
      FILE      *out ;
      char       tmpname[80] = "drdc\0" ;
      int    ii1 , ii2 ;
      out = fopen(tmpname, "w") ;
      for(i=0; i< npair; i++) {
         ii1 = pair[i].i;
         ii2 = pair[i].j;
         sprintf(line," %2d %2d %12g %12g %12g %12g %12g %12g %d\n",
               ii1, ii2, pair[i].d,
               t1[ii1]->r-t2[ii2]->r,
               t1[ii1]->c-t2[ii2]->c,
               (t1[ii1]->r-t2[ii2]->r)/sqrt(ABS(t1[ii1]->dr-t2[ii2]->dr)),
               (t1[ii1]->c-t2[ii2]->c)/sqrt(ABS(t1[ii1]->dc-t2[ii2]->dc)),
               t1[ii1]->p-t2[ii2]->p, pair[i].f ) ;
	    fputs(line, out) ;
      }
      /* Close temp file for DISPLAY mode */
      fclose(out) ;
   }
#endif
   /* Allocate memory for index table */
   if (!(ret = (Match*) malloc ((size_t)npair*sizeof(Match)))) {
      cp_error("Could not alloc memory for index table for matching stars") ;
      goto exit ;
   }

   /* Get best matches */
   nret = 0 ;
   for (i = 0; i < npair; i++) {
      if ((scale == 0.0 && pair[i].f > 5) || 
          (scale != 0.0 && pair[i].f > 4 && ABS (pair[i].d-log10(scale)) < 0.2)) {
         ret[nret].i = pair[i].i ;
	 ret[nret].j = pair[i].j ;
         ret[nret].f = 0 ;
         nret++ ;
      }
   }
   nrej = npair-nret ;

   /* Reallocate memory for true number of matches */
   cp_message (0, 3, "Reallocate memory for match matrix\n") ;
   ret = (Match*) realloc (ret, (size_t)nret*sizeof(Match)) ;
   CXC_MESSAGE (0, 2, "Matched %d triangles... ", nret) ;

   /* Output more info */
   if (nrej == 1) CXC_MESSAGE (0, 2, "Rejected %d pair\n",nrej) ;
   else CXC_MESSAGE (0, 2, "Rejected %d pairs\n", nrej) ;

   if (nret == 0) {
      if (ret != (Match*)NULL) free (ret) ;
      ret = (Match*)NULL ;
   }

   exit:

   /* Free memory */
   cp_message (0, 6, "Deallocating memory\n") ;
   if (pair != (Pair*)NULL) free (pair) ;
   *pnret = nret ;
   return (ret) ;
}

/*-----------------------------------------------------------------------------
 * Function  : purge
 * In        : triangles lists, matching pairs indexes
 * Out       : triangles lists for matching pairs only
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static int 
purge (Triangle  **t1,  Triangle  **t2,  Match *pair,   int nmatch, 
       Triangle ***tp1, Triangle ***tp2, int *npairs)
{
   Triangle **ts1 = *tp1 ;
   Triangle **ts2 = *tp2 ;
   int        i, j, i0 = *npairs ;

   ts1 = (Triangle **) realloc (ts1, (size_t)(nmatch+i0)*sizeof(Triangle*)) ;
   ts2 = (Triangle **) realloc (ts2, (size_t)(nmatch+i0)*sizeof(Triangle*)) ;

   for (i = 0; i < nmatch; i++) {
      j = i+i0 ;
      ts1[j] = (Triangle *) malloc (sizeof(Triangle)) ;
      ts2[j] = (Triangle *) malloc (sizeof(Triangle)) ;
      memcpy (ts1[j], t1[pair[i].i], sizeof(Triangle)) ;
      memcpy (ts2[j], t2[pair[i].j], sizeof(Triangle)) ;
   }
   *tp1 = ts1 ;
   *tp2 = ts2 ;
   *npairs = nmatch+i0 ;

   return 0 ;
}

static int 
purge_old (Triangle **t1,  Triangle **t2,  Match *pair,  int nmatch, 
           Triangle **ts1, Triangle **ts2, Match **opair, int *npairs)
{
   Triangle  *tts1 = *ts1 ;
   Triangle  *tts2 = *ts2 ;
   Match     *outp = *opair ;
   int        i, j, i0 = *npairs ;

   tts1 = (Triangle *)realloc( tts1, (size_t)(nmatch+i0)*sizeof(Triangle) ) ;
   tts2 = (Triangle *)realloc( tts2, (size_t)(nmatch+i0)*sizeof(Triangle) ) ;
   outp = (Match *)realloc( outp, (size_t)(nmatch+i0)*sizeof(Match) ) ;
   for ( i = 0; i<nmatch; i++ ) {
      j = i+i0 ;
      memcpy( tts1+j, t1[pair[i].i], sizeof(Triangle)) ;
      memcpy( tts2+j, t2[pair[i].j], sizeof(Triangle)) ;
      outp[j].i = outp[j].j = j ;
   }
   *npairs = nmatch+i0 ;
   *ts1 = tts1 ;
   *ts2 = tts2 ;
   *opair = outp ;
   return 0 ;
}


/*-----------------------------------------------------------------------------
 * Function  : check_transf
 * In        : matching triangles
 * Out       : subsample of input pairs
 * Purpose   : 
 *   
 *---------------------------------------------------------------------------*/

static Match *
check_transf (Star **s1, int ns1, Triangle **t1, 
              Star **s2, int ns2, Triangle **t2,
              int npairs, int *nout)
{
   Match  *rates = (Match*)NULL ;
   Match  *group = (Match*)NULL ;
   Match  *voted = (Match*)NULL ;
   int     i, nr, ng, nv ;

   /* Rates sharpest peaks in the distribution of log10_P */
   CXC_MESSAGE (1, 3, "Rating sharpest peaks... ") ;
   rates = rating (t1, t2, npairs, &nr) ;
   if (suspend)
      goto exit ;

   /* Select matching triangles by rotation angle */
   if (nr > 3) {
      CXC_MESSAGE (1, 3, "Selecting rotation angle... ") ;
      group = sel_angle (s1, t1, s2, t2, rates, nr, &ng) ;
      if (suspend)
         goto exit ;
   }
   else {
      CXC_MESSAGE (1,3,"few matches: skip\n") ;
      goto exit ;
   }

   if (group != (Match*)NULL) {
      /* Case 1: Strong peak: Casts votes to vertex stars */
      CXC_MESSAGE ( 1, 3, "Strong peak: Casting votes to vertex... ") ;
      voted = voting (t1, t2, group, ng, nout) ;
      if (suspend)
         goto exit ;
   }
   else {
      if (ng >= 0) {
         /* Case 2: Weak peak: Get selected group */
         CXC_MESSAGE (1, 3, "Weak peak: Get selected group\n") ;
         group = (Match*) malloc ((size_t)nr*sizeof(Match)) ;
         nv = 0 ;
         for (i = 0; i < nr; i++) 
            if (rates[i].f == ng) 
               memcpy (&group[nv++], &rates[i], sizeof(Match)) ;
         group = (Match*) realloc (group,(size_t)nv*sizeof(Match)) ;
         ng = nv ;
      }
      else {
         /* Case 3: No peak detected */
         CXC_MESSAGE (1, 2, "No peak detected\n") ;
         *nout = 0 ;
         goto exit ;
      }
   }

   if (voted == (Match*)NULL) {
      /* Try a pedantic solution on given group */
      CXC_MESSAGE (0, 3, "Try a pedantic solution on given group...\n") ;
      voted = pedantic (s1, ns1, t1, s2, ns2, t2, group, ng, nout) ;
      if (suspend)
         goto exit ;
      if (*nout < 0) {
         /* Weak pedantic solution: skip */
         CXC_MESSAGE (0, 3, "weak pedantic solution: skip\n") ;
         *nout = 0 ;
         free (voted) ;
         voted = (Match*)NULL ;
      }
   }

   exit :

   if (rates != (Match*)NULL) free (rates) ;
   if (group != (Match*)NULL) free (group) ;

   return voted ;
}

/*-----------------------------------------------------------------------------
 * Function  : rating
 * In        : matching triangles
 * Out       : subsample of input pairs
 * Purpose   : select groups of matching pairs within peaks in logM histogram
 *   
 *---------------------------------------------------------------------------*/

static Match *
rating (Triangle **t1, Triangle **t2, int npair, int *pnr)
{
   Match   *ret = (Match*)NULL ;
   Pair    *p = (Pair*)NULL ;
   Point   *histog = (Point*)NULL ;
   int      i, j, k, i1, i2, nbin, io ;
   double   bin, x, hlim, xo[30] ;

   IF_ABORT_RETURN (ret) ;

   p = (Pair*) malloc ((size_t)npair*sizeof(Pair)) ;
   for (i = 0; i < npair; i++) {
      p[i].i = i ;
      p[i].j = i ;
      p[i].d = t1[i]->p - t2[i]->p ;
      p[i].f = 0 ;
   }

   /* Sort pairs for increasing d */
   qsort ((void*)p, (size_t)npair, sizeof(Pair), 
          (int (*)(const void *, const void *))sort_pair) ;

   /* Get the complete histogram */
   bin = 0.0028 ;
   nbin = (int)(( p[npair-1].d - p[0].d ) / bin) + 1 ;
   histog = (Point*) malloc ((size_t)nbin*sizeof(Point)) ;

   i = i1 = i2 = 0 ;
   x = p[0].d ;
   while (x < p[npair-1].d) {
      while (i1 < npair-1 && p[i1].d < x) i1++ ;
      i2 = i1 ;
      while (i2 < npair-1 && p[i2].d <= x+bin) i2++ ;
      histog[i].x = x + bin/2.0 ;
      histog[i].y = -(i2-i1) ;
      histog[i].i = i1 ;
      i++ ;
      i1 = i2 ;
      x += bin ;
   }

   qsort ((void*)histog, (size_t)nbin, sizeof(Point), 
          (int (*)(const void *, const void *))sort_p_y) ;

   /* Get best peaks */
   ret = (Match*) malloc ((size_t)npair*sizeof(Match)) ;
   j = k = io = 0 ;
   hlim = MIN(histog[0].y/5.0, -2.0) ;
   do {
      i1 = histog[k].i ;
      i2 = i1 - (int)histog[k].y ;
      for ( i=i1; i<i2; i++) {
         ret[j].i = p[i].i ;
         ret[j].j = p[i].j ;
         ret[j].f = io ;
         j++ ;
      }
#ifdef debug
      printf("k %3d i1 %3d i2 %3d; n %d; outbin %d\n", k, i1, i2, i2-i1, io) ;
#endif
      xo[io] = histog[k].x ;
      if (k < nbin-1 && ABS(histog[k].x-histog[k+1].x) > 1.1*bin) 
         io++ ;
      else
          if (k < nbin-1) xo[io] = (histog[k].x+histog[k+1].x)/2.0 ;
      k++ ;
   } while (k < MIN(15, nbin) && histog[k].y < hlim) ;

   /* Resume some info */
   if (io == 0 || k == 1) {
      if (j > 1) 
         CXC_MESSAGE (0, 3, "Sharp distribution peak at logM = % 8.5f\n", xo[0]) ;
   }
   else {
      CXC_MESSAGE (0, 3, "Multiple distribution peaks: %d\n",k) ;
      for (i=0; i<io; i++)
         CXC_MESSAGE (6, 3, "Peak no %2d at logM = % 8.5f\n", i, xo[i]) ;
   }

   ret = (Match*) realloc (ret, (size_t)j*sizeof(Match)) ;
   *pnr = j ;

   free (p) ;
   free (histog) ;
   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : sel_angle
 * In        : matching triangles
 * Out       : subsample of input pairs
 * Purpose   : select groups of matching pairs within peak in angles histogram
 *             
 *---------------------------------------------------------------------------*/

static Match *
sel_angle(Star **s1, Triangle **t1, Star **s2, Triangle **t2, 
          Match *pair, int nmatch, int *pn)
{
   Match  *ret = (Match*)NULL ;
   Pair   *p = (Pair*)NULL ;
   Point  *histog = (Point*)NULL ;
   double  a, rp ;
   double  bin = 0.3 ;
   int     i, j, l, np, i1, i2, nmax, imax, nbin, nm = 0 ;
   int     r = 0, k = 0 ;

   IF_ABORT_RETURN (ret) ;

   do {

      /* Get rotation angle value for each matching group */
      p = (Pair*) malloc ((size_t)nmatch*sizeof(Pair)) ;
      np = 0 ;
      while (pair[k].f == r && k < nmatch) {
         a = comp_angle(s1, t1[pair[k].i], s2, t2[pair[k].j]) ;
         p[np].i = pair[k].i ;
         p[np].j = pair[k].j ;
         p[np].d = a*180.0/(double)PI ;
         p[np].f = 0 ;
         np++ ;
         k++ ;
      }
      qsort ((void*)p, (size_t)np, sizeof(Pair), 
             (int (*)(const void *, const void *))sort_pair) ;
#ifdef debug
      if (verbose > 6) {
         printf("nmatch %d k %d np %d\n", nmatch, k, np) ;
         printf("Rotation angle:  step %d\n", r) ;
         for (j=0; j<np; j++) printf(" % 9.4f", p[j].d) ;
         printf("\n") ;
      }
#endif

      /* Append angles in the range -180:-179 to the end of the vec */
      j = 0 ;
      while (j < np-1 && p[j].d < -179.0) j++ ;
      p = (Pair*) realloc ( p, (size_t)(np+j)*sizeof(Pair) ) ;
      for (l = 0; l < j; l++) {
         memcpy(&p[np], &p[l], sizeof(Pair)) ;
         p[np].d += 360.0 ;
         np++ ;
      }

      /* Get the interval which maximize the main peak */
      i1 = i2 = imax = nmax = 0 ;
      while (i1 < np-1) {
         while (i2 < np-1 && p[i2].d-p[i1].d < bin) i2++ ;
         if (i2-i1 > nmax) {
            nmax = i2-i1 ;
            imax = i1 ;
         }
         i1++ ;
         while (i1 < i2 && p[i2].d-p[i1].d > bin) i1++ ;
      }

      /* Get the complete histogram */
      bin = 0.45 ;
      nbin = 361/bin + 1 ;
      histog = (Point*) malloc ((size_t)nbin*sizeof(Point)) ;
      i = i1 = i2 = 0 ;
      a = p[imax+nmax/2].d-bin/2.0 ;
      while (a >= -180.0) a -= bin ;
      while (a < 180.0) {
         while (i1 < np-1 && p[i1].d < a) i1++ ;
         i2 = i1 ;
         while (i2 < np-1 && p[i2].d <= a+bin) i2++ ;
         histog[i].x = a + bin/2.0 ;
         histog[i].y = -(i2-i1) ;
         histog[i].i = i1 ;
         i++ ;
         i1 = i2 ;
         a += bin ;
      }
      histog = (Point*) realloc ( histog, (size_t)i*sizeof(Point)) ;
      qsort ((void*)histog, (size_t)i, sizeof(Point), 
             (int (*)(const void *, const void *))sort_p_y) ;
      
      rp =  histog[0].y / MIN( histog[1].y, -0.5 ) ;
#ifdef debug
      printf("histogram:\nangle % 7f: nbin %.0f; next bin % 7f %.0f;\n",
            histog[0].x,-histog[0].y,histog[1].x,-histog[1].y ) ;
#endif
      i1 = histog[0].i ;
      i2 = i1 - (int)histog[0].y ;
      if (rp > 3) {

         /* Strong peak */
         ret = (Match*) malloc ((size_t)(i2-i1)*sizeof(Match)) ;
         j = 0 ;
         for (i = i1; i < i2; i++) {  /** FIXME: i2 must be included? **/
            ret[j].i = p[i].i ;
            ret[j].j = p[i].j ;
            ret[j].f = 0 ;
            j++ ;
         }
         *pn = j ;
         CXC_MESSAGE (0, 3, "Sharp distribution peak at rotation angle = % 8.3f\n",histog[0].x) ;
         CXC_MESSAGE (2, 3, "Returning %d selected matching pairs among %d step %d\n", j, nmatch, r) ;
         break ;
      }
      else {
         if (rp > 2) 
            CXC_MESSAGE (0, 3, "Weak distribution peak at rotation angle = % 8.3f, step %d\n",
                  histog[0].x, r) ;
         if (i2-i1 > nm) {
            nm = i2-i1 ;
            *pn = r ;
         }
      }

      /* Free memory, get next group */  
      free (p) ;
      free (histog) ;
      r++ ;

   } while (k < nmatch-1) ;

   if (ret == (Match*)NULL) {
      if (nm > 2) 
         CXC_MESSAGE (0,3,"Weak peak %d in logM distribution\n", *pn) ;
      else {
         CXC_MESSAGE (0,3,"No peak detected in angle distribution\n") ;
         *pn = -1 ;
      }
   }
   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : voting
 * In        : triangles list 1 & 2 and list of matching pairs
 * Out       : selected matching pairs
 * Purpose   : select groups of stars matching pairs 
 *             
 *---------------------------------------------------------------------------*/

static Match *
voting (Triangle **t1, Triangle **t2, Match *group, int ng, int *pnr)
{
   Match   *ret = (Match*)NULL ;
   Match   *tmp = (Match*)NULL ;
   int      i, j, k, l, nv, nr, rlim ;

   IF_ABORT_RETURN (ret) ;

   /* Alloc memory for working arrays */
   tmp = (Match*) malloc ((size_t)(3*ng) * sizeof(Match)) ;
   ret = (Match*) malloc ((size_t)(3*ng) * sizeof(Match)) ;

   /* Accumulate vertex index */
   l = 0 ;
   for (k = 0; k < ng; k++) {
      i = group[k].i ;
      j = group[k].j ;
      for (nv = 0; nv < 3; nv++) {
         tmp[l].i = t1[i]->ind[nv] ;
         tmp[l].j = t2[j]->ind[nv] ;
         tmp[l].f = 1 ;
         l++ ;
      }
   }

   /* Cast votes to vertex */
   qsort ((void*)tmp, (size_t)l, sizeof(Match), 
          (int (*)(const void *, const void *))sort_match_i) ;
   j = k = 0 ;
   while (k < l) {
      i = 0 ;
      while (k+i < l && memcmp(&tmp[k], &tmp[k+i], sizeof(Match)) == 0) i++ ;
      memcpy ( &ret[j], &tmp[k], sizeof(Match) ) ;
      ret[j++].f = i ;
      k += i ;
   }
   free (tmp) ;
   nr = j ;
   ret = (Match*) realloc (ret, (size_t)nr*sizeof(Match)) ;

   /* Remove twice-assigned points */
   ret = uniq (ret, &nr) ;
   qsort ((void*)ret, (size_t)nr, sizeof(Match), 
          (int (*)(const void *, const void *))sort_match_j) ;
   ret = uniq (ret, &nr) ;

   /* Sort by vote and return points with higher votes */
   qsort ((void*)ret, (size_t)nr, sizeof(Match), 
          (int (*)(const void *, const void *))sort_match_r) ;
   rlim = MAX (MIN (ret[2].f/2, 15), 3) ;
   cp_message (0,4,"Max vote %d, limiting vote %d\n", ret[0].f, rlim) ;
   i = 0 ;
   while (i < nr && ret[i].f >= rlim) i++ ;

   if (i >= 3) {
      CXC_MESSAGE (0, 3, "Returning %d selected vertex\n", i) ;
      ret = (Match*) realloc (ret, (size_t)i*sizeof(Match)) ;
      *pnr = i ;
   }
   else {
      CXC_MESSAGE (0, 3, "too few vertex: voting failed\n") ;
      free (ret) ;
      ret = (Match*)NULL ;
      *pnr = 0 ;
   }

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : pedantic
 * In        : stars and matching triangles lists
 * Out       : indexes of stars of best matching triangle
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static Match *
pedantic (Star **si1, int ns1, Triangle **t1, 
          Star **si2, int ns2, Triangle **t2,
          Match *group, int ng, int *pnr)
{
   Match    *ret = (Match*)NULL ;
   Trasf    *tmp = NULL ;
   Star     **s1 = NULL, **s2 = NULL, **sc1 = NULL, **sc2 = NULL ;
   int       i, j, k, ii, j1, j2, nbin = 0, nmatch, ibest = -1, minmatch ;
   double    dx, dm, a, s, sc ;
   double    tol_pos = 20.0*l_param1->eps ;

   minmatch = (int)(0.5 * MIN (ns1, ns2)) ;
   CXC_MESSAGE (3, 3, "Minimum number to get a good match: %d;  Tol Radius: %g\n", 
               minmatch, tol_pos) ;

   IF_ABORT_EXIT ;

   /* Get working copy of input stars */
   s1 = star_ptr_dup (si1, (size_t)ns1) ;
   s2 = star_ptr_dup (si2, (size_t)ns2) ;
   sc1 = star_ptr_dup (si1, (size_t)ns1) ;
   sc2 = star_ptr_dup (si2, (size_t)ns2) ;

   /* Sort list s1 for increasing yprj; sc1 preserve original order */
   qsort ((void*)s1, (size_t)ns1, sizeof(Star*), 
          (int (*)(const void *, const void *))sort_star_yprj) ;

   /* Scan all input groups */
   nmatch = 0 ;
   for (i = 0; i < ng; i++) {

      IF_ABORT_EXIT ;

      /* Compute transformation between current pair */
      tmp = MatchTriangle (sc1, t1[group[i].i], sc2, t2[group[i].j]) ;

      /*Apply Transformation and sort for yprj */
      for (k = 0; k < ns2; k++) {
         s2[k]->xrot = tmp->xfit_coeff[0] + tmp->xfit_coeff[1]*s2[k]->xprj 
		                          + tmp->xfit_coeff[2]*s2[k]->yprj ;
         s2[k]->yrot = tmp->yfit_coeff[0] + tmp->yfit_coeff[1]*s2[k]->xprj 
		                          + tmp->yfit_coeff[2]*s2[k]->yprj ;
         s2[k]->mcal = s2[k]->mag + tmp->dmag ;
         s2[k]->flag = 0 ;
      }

      IF_ABORT_EXIT ;

      /* Sort list2 for increasing yrot */
      qsort ((void*)s2, (size_t)ns2, sizeof(Star*), 
             (int (*)(const void *, const void *))sort_star_yrot) ;

      /* Cross correlate two set of projected positions */
      nbin = ii = j1 = j2 = 0 ;
      while (ii < ns1) {
         while (j1 < ns2-1 && s2[j1]->yrot-s1[ii]->yprj < -tol_pos)
	    j1++ ;
         while (j2 < ns2-1 && s2[j2]->yrot-s1[ii]->yprj < tol_pos)
            j2++ ;
         for (j = j1; j <= j2; j++) {
	    dx = ABS (s2[j]->xrot - s1[ii]->xprj) ;
	    dm = ABS (s2[j]->mcal - s1[ii]->mag) ;
  	    if (dx <= tol_pos && dm <= 2.0*gparam.t0_mag && s2[j]->flag == 0) {
               s2[j]->flag = 1 ;
               nbin++ ;
               break ;
            }
         }
         ii++ ;
      }

      IF_ABORT_EXIT ;

      /* Get bestmatch */
      if (nbin > nmatch) {
         nmatch = nbin ;
         ibest = i ;
         if (tmp->inv == 0) s = 1.0 ;
         else s = -1.0 ;
         /* FIXME: use correct formula */
         a = get_angle (tmp->xfit_coeff, tmp->yfit_coeff) ;
         sc = t1[group[i].i]->p - t2[group[i].j]->p ;
         if (tmp->inv) 
            CXC_MESSAGE (6, 3,"   %3d match at logM = % 8.5f angle = % 8.3f - axis inversion\n",
            nmatch, sc, RADTODEG (a)) ;
         else 
            CXC_MESSAGE (6, 3,"   %3d match at logM = % 8.5f angle = % 8.3f\n",
            nmatch, sc, RADTODEG (a)) ;
      }
      free (tmp) ;
      tmp = (Trasf*)NULL ;

      if (nbin > minmatch) break ;
   }

   exit :

   if (s1) free (s1) ;
   if (s2) free (s2) ;
   if (sc1) free (sc1) ;
   if (sc2) free (sc2) ;
   if (tmp != (Trasf*)NULL) free (tmp) ;
   if (ibest < 0) return ret ;

   /* Fill returning values */
   ret = (Match*) malloc ((size_t)3*sizeof(Match)) ;
   i = group[ibest].i ;
   j = group[ibest].j ;
   for (k = 0; k < 3; k++) {
      ret[k].i = t1[i]->ind[k] ;
      ret[k].j = t2[j]->ind[k] ;
      ret[k].f = 1 ;
   }
   *pnr = 3 ;
   if (nbin < minmatch) 
      *pnr = -3 ;

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : comp_angle
 * In        : 3 corner stars
 * Out       : rotation angle
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static double 
comp_angle (Star **s1, Triangle *t1, Star **s2, Triangle *t2)
{
   double   xv1[3], yv1[3], xv2[3], yv2[3] ;
   double   x_dev[3] = { 1.0, 1.0, 1.0} ;
   double   y_dev[3] = { 1.0, 1.0, 1.0} ;
   double   xc[3], yc[3] ;
   double   xchisq, ychisq ;
   int      ifit[3] = { 1, 1, 1} ;
   int      i, ii ;
   void     (*PlateFunc)() = &FitFunc ;

   for (i = 0; i < 3; i++) {
      ii = t1->ind[i] ;
      xv1[i] = s1[ii]->xprj ;
      yv1[i] = s1[ii]->yprj ;
      ii = t2->ind[i] ;
      xv2[i] = s2[ii]->xprj ;
      yv2[i] = s2[ii]->yprj ;
   }

   /* Fit coordinates */
   (void) fit (xv2, yv2, xv1, x_dev, 3, xc, ifit, 3, &xchisq, PlateFunc) ;
   (void) fit (xv2, yv2, yv1, y_dev, 3, yc, ifit, 3, &ychisq, PlateFunc) ;

   return get_angle (xc, yc) ;
}

static double
get_angle (double   xc[3], double yc[3])
{
   double ret ;
   double inv = ( xc[1]*yc[2] > 0.0 ) ? 1.0 : -1.0 ;
   double c11 = inv * xc[1] ;
   double c12 = inv * xc[2] ;
   double c21 =       yc[1] ;
   double c22 =       yc[2] ;
   double r1, r2 ;

   r1 = ( c21 > 0.0 ) ? atan2 (  c21, c11 ) : 
        ( c21 < 0.0 ) ? atan2 ( -c21, c11 ) : 0.0 ;

   r2 = ( c12 > 0.0 ) ? atan2 (  c12, c22 ) : 
        ( c12 < 0.0 ) ? atan2 ( -c12, c22 ) : 0.0 ;

   ret = ( r1 + r2 ) / 2.0 ;
   /* if ( c11 < 0.0 && c22 < 0.0 ) ret -= PI ; */
#ifdef debug
   if (verbose>6)
      printf("ang %f; c11 %g c22 %g\n", ret, c11,c22) ;
#endif

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : MatchTriangle
 * In        : 3 corner stars
 * Out       : transformation coefficients
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static Trasf *
MatchTriangle (Star **s1, Triangle *t1, Star **s2, Triangle *t2)
{
   Trasf   *ret = (Trasf*)NULL ;
   double   xv1[3], yv1[3], xv2[3], yv2[3] ;
   double   x_dev[3] = { 1.0, 1.0, 1.0} ;
   double   y_dev[3] = { 1.0, 1.0, 1.0} ;
   double   xchisq, ychisq ;
   int      ifit[3] = { 1, 1, 1} ;
   int      i, i1, i2 ;
   void     (*PlateFunc)() = &FitFunc ;

   ret = (Trasf*) malloc (sizeof(Trasf)) ;
   if (ret == (Trasf*)NULL) return ret ;

   if (t1->ot == t2->ot) ret->inv = 0 ;
   else  ret->inv = 1 ;

   ret->dmag = 0.0 ;
   for (i = 0; i < 3; i++) {
      i1 = t1->ind[i] ;
      i2 = t2->ind[i] ;
      xv1[i] = s1[i1]->xprj ;
      yv1[i] = s1[i1]->yprj ;
      xv2[i] = s2[i2]->xprj ;
      yv2[i] = s2[i2]->yprj ;
      ret->dmag += s1[i1]->mag - s2[i2]->mag ;
   }
   ret->dmag /= 3.0 ;

   /* Fit coordinates */
   (void) fit (xv2, yv2, xv1, x_dev, 3, ret->xfit_coeff, ifit, 3, &xchisq, 
                 PlateFunc) ;
   (void) fit (xv2, yv2, yv1, y_dev, 3, ret->yfit_coeff, ifit, 3, &ychisq, 
                 PlateFunc) ;

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : GetRmax
 * In        : 
 * Out       : 
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static double 
GetRmax (Triangle **t, int ntr)
{
   int    i ;
   double rmed = 0.0 ;
   double rmax = 0.0 ;
   double ret ;
   for (i = 0; i < ntr; i++) {
      rmed += t[i]->dr ;
      rmax = MAX (rmax, t[i]->dr) ;
   }
   rmed /= MAX (1, ntr) ;
   ret = MIN (rmax, 15 * rmed) ;
   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : uniq
 * In        : 
 * Out       : 
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static Match *
uniq (Match *ret, int *pnr)
{
   int nr = *pnr ;
   int i, j = 0 ;
   while (j < nr-1) {
      if (ret[j].i == ret[j+1].i || ret[j].j == ret[j+1].j) {
         i = j+1 ;
         if (ret[j].f < ret[j+1].f) i = j ;
         while (i < nr-1) {
            memcpy (&ret[i], &ret[i+1], sizeof(Match)) ;
            i++ ;
         }
         nr-- ;
      }
      j++ ;
   }
   ret = (Match*) realloc (ret, (size_t)nr*sizeof(Match)) ;
   *pnr = nr ;
   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : get_dmag
 * In        : 
 * Out       : 
 * Purpose   : 
 *             
 *---------------------------------------------------------------------------*/

static double 
get_dmag (Star **s1, Star **s2, Match *pair, int nmatch)
{
   double ret = 0.0 ;
   int    i ;
   for (i = 0; i < nmatch; i++)
      ret += s1[pair[i].i]->mag - s2[pair[i].j]->mag ;
   ret /= (double) MAX (nmatch, 1) ;
   return ret ;
}
