/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: findshift.c
 * Author 	: Paolo Montegriffo
 * Created on	: October, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0.0 - November 9, 2000
 * Version      : 2.1.0 - June 28, 2001
 * Version      : 2.1.1 - July 11, 2001
 * Version      : 2.1.3 - November 14, 2001 : Modified MOSAIC mode
 * Version      : 2.1.7 - Jun 13, 2002: added control for GUI
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "cataxcorr.h"
#include "cataxcorr_p.h"
#include "cp_messages.h"

/*----------------------------------------------------------------------------
 * 							Private Types
 *--------------------------------------------------------------------------*/

struct p {
   int x ;
   int y ;
} ;

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern g_pars    gparam ;
extern boolean   mag_flag ;  /* TRUE: we are using magnitudes */
extern boolean   suspend ;

double max_shift ;

/*----------------------------------------------------------------------------
 * 							Defines
 *--------------------------------------------------------------------------*/

#ifndef MAX
#define MAX(a,b) (((a)>=(b)) ? (a) : (b))
#endif
#ifndef MIN
#define MIN(a,b) (((a)<=(b)) ? (a) : (b))
#endif
#ifndef ABS
#define ABS(a) ((a) < 0 ? (-(a)) : (a))
#endif

/*----------------------------------------------------------------------------
 * 							Private functions
 *--------------------------------------------------------------------------*/

static int sort_star_y(const Star **d1, const Star **d2)
{
   if ( (*d1)->yprj > (*d2)->yprj ) 
      return 1 ; 
   else 
      return -1 ;
}

static int is_opposite( struct p pos1, struct p pos2 )
{
   if( ABS(pos2.x-pos1.x) == 2 || ABS(pos2.y-pos1.y) == 2 ) 
      return 1 ;
   else 
      return 0 ;
}

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

extern StarMatch *FindShift   (StarList *sl1, StarList *sl2, int *pnmatch) ;
static void       GetRange    (Star **s, int n,
                               double *x0, double *x1, double *y0, double *y1) ;
static struct p  is_on_border (double x, double y, double lim[4]) ;

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

/*-----------------------------------------------------------------------------
 * Function  : FindShift
 * In        : Two star lists, number of stars in each one
 * Out       : Number of matching stars found, index table for matching stars
 * Purpouse  : Makes a Xcorrelation between the two input star lists;
 *	       the Xcorrelation is made up to a maximum given shift max_shift
 * Note	     : The algorithm is higly optimized to work speedly: input y 
 *             coordinates are sorted in ascending order
 *             The 2nd star list is calibrated in magnitude
 *             
 *---------------------------------------------------------------------------*/

StarMatch *
FindShift(StarList *sl1, StarList *sl2, int *pnmatch)
{
   Star   **s1 = star_ptr_dup(sl1->data, sl1->nstar) ;
   Star   **s2 = star_ptr_dup(sl2->data, sl2->nstar) ;
   int      nstar1 = sl1->nstar ;
   int      nstar2 = sl2->nstar ;
   StarMatch *match = NULL ;              /* returning index table */
   int      nmatch = 0 ;
   int      i, i2, i1, ii2, ii1, nbin=0, npeaks=0 ;
   int     *list2  = (int*)NULL, *list1  = (int*)NULL ;
   int     *blist2 = (int*)NULL, *blist1 = (int*)NULL ;
   double   dx, dy, dmag, dxi, dyi, dmagi ;
   double   bestdmag = 0.0 ;
   double   s1_lim[4], s2_lim[4] ;
   double   s1_xl, s1_yl, s2_xl, s2_yl, s1_d, s2_d, s_d, x_s = 0.0, y_s = 0.0 ;
   double   overlap ;
   double   tol_pos ;
   struct p cod, pn ;
   int      minmatch = (int)( 0.75*MIN(nstar2, nstar1) ) ;
   size_t   maxnbin = (size_t)MAX(nstar1, nstar2) ;

   max_shift = sl1->param->max_shift ;
   /* tol_pos = 0.5 * sl1->param.crow_rad ; */
   tol_pos = gparam.min_tol ;

   cp_message(1, 3, "Tol pos: %g; max_shift: %g\n", tol_pos, max_shift) ;
   /* if(max_shift==0.0) return match ;*/
   cod.x = cod.y = pn.x = pn.y = 0 ;

   /* Allocate memory for indexes arrays */
   if(!(list1 =  (int *) malloc (maxnbin*sizeof(int))) ||
      !(list2 =  (int *) malloc (maxnbin*sizeof(int))) ||
      !(blist1 = (int *) malloc (maxnbin*sizeof(int))) ||
      !(blist2 = (int *) malloc (maxnbin*sizeof(int))) ){
      cp_error("Could not alloc memory for cross-correlating the tables");
      goto exit;
   }

   /* Get coordinates range */
   GetRange(s1, nstar1, &s1_lim[0], &s1_lim[1], &s1_lim[2], &s1_lim[3]) ;
   printf("list1 range %d %g %g %g %g\n",nstar1,s1_lim[0],s1_lim[1], s1_lim[2],s1_lim[3]) ;
   GetRange(s2, nstar2, &s2_lim[0], &s2_lim[1], &s2_lim[2], &s2_lim[3]) ;
   printf("list2 range %d %g %g %g %g\n",nstar2,s2_lim[0],s2_lim[1], s2_lim[2],s2_lim[3]) ;

   /* Get min star density */
   s1_xl = s1_lim[1] - s1_lim[0] ;
   s1_yl = s1_lim[3] - s1_lim[2] ;
   s2_xl = s2_lim[1] - s2_lim[0] ;
   s2_yl = s2_lim[3] - s2_lim[2] ;
   s1_d = (double)nstar1 / (s1_xl*s1_yl) ;
   s2_d = (double)nstar2 / (s2_xl*s2_yl) ;
   s_d = MIN(s1_d, s2_d) ;

   overlap = MIN( MIN( s1_xl, s1_yl ), MIN( s2_xl, s2_yl ) ) / 2.0 ;
   if( gparam.mosaic == TRUE && max_shift > overlap ) {
      cp_warning("max_shift bigger than half coordinate range: reducing to %.1f",
         overlap) ;
      max_shift = overlap ;
   }

   /* Sort star lists for y coordinates */
   cp_message(0,3,"Sorting star lists for y coordinate\n") ;
   qsort((void*)s1, (size_t)nstar1, sizeof(Star*), 
         (int (*)(const void *, const void *))sort_star_y) ;
   qsort((void*)s2, (size_t)nstar2, sizeof(Star*), 
         (int (*)(const void *, const void *))sort_star_y) ;

   gui_update () ;
   if (suspend)
      goto exit ;

   /* Cross correlation */
   for (i1 = 0; i1 < nstar1; i1++) {
      if (gparam.mosaic == TRUE) 
         cod = is_on_border(s1[i1]->xprj, s1[i1]->yprj, s1_lim) ;
      if (gparam.mosaic != TRUE 
         || memcmp(&cod, &pn, sizeof(struct p)) != 0) {
         gui_update () ;
         if (suspend)
            goto exit ;
         for (i2 = 0; i2 < nstar2; i2++) {
            if (gparam.mosaic != TRUE || 
                 is_opposite(cod, is_on_border(s2[i2]->xprj, s2[i2]->yprj, s2_lim))) {
               if (s1[i1]->flag != 1 && s2[i2]->flag != 1) {
                  dy = s2[i2]->yprj - s1[i1]->yprj ;
                  if (gparam.mosaic == TRUE || ABS(dy) < max_shift) {
                     dx = s2[i2]->xprj - s1[i1]->xprj ;
                     dmag = s2[i2]->mag - s1[i1]->mag ;
	             if (gparam.mosaic == TRUE || ABS(dx) < max_shift) {
		        nbin = ii1 = ii2 = 0 ;
		        while (ii1 < nstar1-1) {
		           ii1++ ;
	                   if (ii1 != i1) {
			      while (ii2 > 0 && s2[ii2]->yprj-s1[ii1]->yprj-dy > -tol_pos)
  		                 ii2-- ;
			      while (ii2 < nstar2-1 && s2[ii2]->yprj-s1[ii1]->yprj-dy < -tol_pos)
		                 ii2++ ;
	        	      while (ii2 < nstar2-1 && (dyi=s2[ii2]->yprj-s1[ii1]->yprj-dy) < tol_pos) {
		                 if (ii2 != i2) {
			            if (s1[ii1]->flag != 1 && s2[ii2]->flag != 1) {
				       dxi = ABS(s2[ii2]->xprj - s1[ii1]->xprj - dx) ;
				       dmagi = ABS(s2[ii2]->mag - s1[ii1]->mag - dmag) ;
				       if (dxi < tol_pos && dmagi < gparam.t0_mag) {
                		          list1[nbin] = ii1 ;
				          list2[nbin] = ii2 ;
                		          nbin++ ;
                		          break ;
                		       }
			            }
			         }
			         ii2++ ;
			      }
        	           }
        	        }
        	        list1[nbin] = i1 ;
        	        list2[nbin] = i2 ;
                        nbin++ ;
        	        if (nbin > 1 && nbin >= nmatch) {
                           x_s = MIN( s1_lim[1], s2_lim[1]-dx ) - MAX( s1_lim[0], s2_lim[0]-dx) ;
                           y_s = MIN( s1_lim[3], s2_lim[3]-dy ) - MAX( s1_lim[2], s2_lim[2]-dy) ;
                           minmatch = (int)(0.6*s_d*x_s*y_s)+1 ;
        	           nmatch = nbin ;
		           bestdmag = -dmag ;
        	           for (i=0; i<nmatch; i++) {
        		      blist1[i] = list1[i] ;
        		      blist2[i] = list2[i] ;
        	           }
		           npeaks++ ;
        	           cp_message(0,4,"%2d: %2d matches at list %3d coinc %3d: dx=%10g dy=%10g dmag=%5g; minmatch %d\n",
                		      npeaks, nmatch, i2, i1, dx, dy, bestdmag, minmatch) ;
        	        }
                        if (nmatch > minmatch) break ;
	             }
	          }
	       }
            }
         }
      }
      if (nmatch > minmatch) break ;
   }
   if (nmatch == 0) {
      cp_error("Could not find any matching stars") ;
      goto exit ;
   }

   /* Allocate memory for index table */
   if (!(match = (StarMatch*) malloc ((size_t)(nmatch)*sizeof(StarMatch)))) {
      cp_error("Could not alloc memory for index table for matching stars") ;
      goto exit ;
   }
   /* Fill index table */
   for (i = 0; i < nmatch; i++) {
      match[i].s1 = s1[blist1[i]] ;
      match[i].s2 = s2[blist2[i]] ;
   }
   *pnmatch = nmatch ;

   /* Calibrate 2nd star list */
   for (i=0; i<nstar2; i++) 
      s2[i]->mcal = s2[i]->mag + bestdmag ;

   /* Checks the overlapping area fraction against max trans order */
   if ( (overlap = (x_s*y_s) / (s2_xl*s2_yl)) < 0.60 ) {
      cp_warning("Overlapping region covers %g %% of rotating field",overlap*100.0) ;
      cp_warning("May want to use a lower transformation order") ;
   }

   exit:

   /* Free memory */
   if (s1) free (s1) ;
   if (s2) free (s2) ;
   if (list2  != (int*)NULL) free (list2) ;
   if (list1  != (int*)NULL) free (list1) ;
   if (blist2 != (int*)NULL) free (blist2) ;
   if (blist1 != (int*)NULL) free (blist1) ;

   return (match) ;
}

static void GetRange(Star **s, int n,
                     double *x0, double *x1, double *y0, double *y1)
{
   int i ;
   *x0 = *x1 = s[0]->xprj ;
   *y0 = *y1 = s[0]->yprj ;
   for (i=0; i<n; i++) {
      *x0 = MIN(*x0, s[i]->xprj) ;
      *x1 = MAX(*x1, s[i]->xprj) ;
      *y0 = MIN(*y0, s[i]->yprj) ;
      *y1 = MAX(*y1, s[i]->yprj) ;
   }
   return ;
}

static struct p is_on_border(double x, double y, double lim[4])
{
   struct p ret ;
   ret.x = ret.y = 0 ;
   if ( x-lim[0] < max_shift ) ret.x = -1 ;
   if ( lim[1]-x < max_shift ) ret.x =  1 ;
   if ( y-lim[2] < max_shift ) ret.y = -1 ;
   if ( lim[3]-y < max_shift ) ret.y =  1 ;
   return ret ;
}

