/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: platemodel.c
 * Author 	: Paolo Montegriffo
 * Created on	: October, 2000	
 * Language	: ANSI C
 * Description	: 
 * Version      : 2.0 - January 9, 2002
 * Version      : 2.1.4 - February 15, 2002
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 							Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"
#include "platemodel.h"

#ifdef HAVE_GSL
#include <gsl/gsl_math.h>
#include <gsl/gsl_sf_legendre.h>
#include <gsl/gsl_linalg.h>
#define MAXORD 10
#else
#define MAXORD 5
#endif

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

const int  minstar[]   = { 3, 3, 30, 48, 60, 78, 98, 100, 120, 150, 200 } ;
const int  LINncoef[]  = { 1, 3 } ;
const int  SAOncoef[]  = { 1, 3, 6, 10, 13 } ;
const int  POLYncoef[] = { 1, 3, 6, 10, 15, 21, 28, 36, 45, 55, 66 } ;
const int  POLYRADncoef[] = { 1, 4, 7, 12, 17, 24, 31, 40, 49, 60, 71 } ;
const int  LEGEncoef[] = { 1, 3, 6, 10, 15, 21, 28, 36, 45, 55, 66 } ;


const model_type m_type[NUM_MODEL] = 
{ 
   { (void(*)())LINPlateFunc,  LINncoef,  1 },
   { (void(*)())SAOPlateFunc,  SAOncoef,  4 },
   { (void(*)())POLYPlateFunc, POLYncoef, MAXORD },
   { (void(*)())POLYRADPlateFunc, POLYRADncoef, MAXORD },
   { (void(*)())LEGEPlateFunc, LEGEncoef, 10 },
} ;

/*-----------------------------------------------------------------------------
 * Function  : LINPlateFunc
 * In        : pix coords for matching pairs, 
 *             no. of matching pairs
 * Out       : terms of fitting function
 * Purpouse  : Specify the form of fitting function: LINEAR function
 *---------------------------------------------------------------------------*/

void LINPlateFunc(double x, double y, double p[], int np)
{
   int    i ;

   for (i=0; i<np; i++) p[i] = 0.0 ;

   p[0]  = 1 ;      /* 0th order */
   p[1]  = x ;
   p[2]  = y ;      /* 1st order */
   return;
}

/*-----------------------------------------------------------------------------
 * Function  : SAOPlateFunc
 * In        : pix coords for matching pairs, 
 *             no. of matching pairs
 * Out       : terms of fitting function
 * Purpouse  : Specify the form of fitting function: SAO polynome
 *---------------------------------------------------------------------------*/

void SAOPlateFunc(double x, double y, double p[], int np)
{
   int    i ;
   double x2, y2, x3, y3, r2 ;

   x2 = x * x ;
   y2 = y * y ;
   x3 = x * x2 ;
   y3 = y * y2 ;
   r2 = x2 + y2 ;

   for (i=0; i<np; i++) p[i] = 0.0 ;

   p[0]  = 1 ;      /* 0th order */
   p[1]  = x ;
   p[2]  = y ;      /* 1st order */
   p[3]  = x2 ;
   p[4]  = y2 ;
   p[5]  = x*y ;    /* 2nd order */
   p[6]  = x3 ;
   p[7]  = y3 ;
   p[8]  = x2*y ;
   p[9]  = x*y2 ;   /* 3rd order */
   p[10] = r2 ;
   p[11] = x*r2 ;
   p[12] = y*r2 ;   /* 4th order */

   return;
}

/*-----------------------------------------------------------------------------
 * Function  : POLYPlateFunc
 * In        : pix coords for matching pairs, 
 *             no. of matching pairs
 * Out       : terms of fitting function
 * Purpouse  : Specify the form of fitting function   ** FIXME **
 *---------------------------------------------------------------------------*/

void POLYPlateFunc(double x, double y, double p[], int np)
{
#ifdef HAVE_GSL

   int    i, j, k ;

   for (i=0; i<np; i++) p[i] = 0.0 ;

   for (k=i=0; i<=MAXORD; i++) 
      for (j=0; j<=i; j++) 
         p[k++] = gsl_pow_int(x, i-j) * gsl_pow_int(y, j) ;

#else

   int    i ;
   double x2, y2, x3, y3, x4, y4, x5, y5 ;

   x2 = x * x ;
   y2 = y * y ;
   x3 = x * x2 ;
   y3 = y * y2 ;
   x4 = x2 * x2 ;
   y4 = y2 * y2 ;
   x5 = x3 * x2 ;
   y5 = y3 * y2 ;

   for (i=0; i<np; i++) p[i] = 0.0 ;

   p[0]  =  1      ;     /* 0th order */
   p[1]  =  x      ;
   p[2]  =       y ;     /* 1st order */
   p[3]  = x2      ;
   p[4]  =  x *  y ;
   p[5]  =      y2 ;     /* 2nd order */
   p[6]  = x3      ;
   p[7]  = x2 *  y ;
   p[8]  =  x * y2 ;
   p[9]  =      y3 ;     /* 3rd order */
   p[10] = x4      ;
   p[11] = x3 *  y ;
   p[12] = x2 * y2 ;
   p[13] =  x * y3 ;
   p[14] =      y4 ;     /* 4th order */
   p[15] = x5      ;
   p[16] = x4 *  y ;
   p[17] = x3 * y2 ;
   p[18] = x2 * y3 ;
   p[19] =  x * y4 ;
   p[20] =      y5 ;     /* 5th order */

#endif
   return;
}

extern double xx0, xx1, yy0, yy1 ;

void LEGEPlateFunc(double x, double y, double p[], int np)
{
#ifdef HAVE_GSL

   int    i, j, k ;
   double xn = 2.0*(x-xx0)/(xx1-xx0)-1.0 ;
   double yn = 2.0*(y-yy0)/(yy1-yy0)-1.0 ;

   for (i=0; i<np; i++) p[i] = 0.0 ;

   for (k=i=0; i<=MAXORD; i++) 
      for (j=0; j<=i; j++) 
         p[k++] = gsl_sf_legendre_Pl(i-j, xn) * gsl_sf_legendre_Pl(j, yn) ;
#endif
   return;
}



/*-----------------------------------------------------------------------------
 * Function  : POLYRADPlateFunc
 * In        : pix coords for matching pairs, 
 *             no. of matching pairs
 * Out       : terms of fitting function
 * Purpouse  : Specify the form of fitting function: POLYRAD
 *---------------------------------------------------------------------------*/

void POLYRADPlateFunc(double x, double y, double p[], int np)
{
#ifdef HAVE_GSL

   int    i, j, k ;

   for (i=0; i<np; i++) p[i] = 0.0 ;

   for (k=i=0; i<=MAXORD; i++) {
      for (j=0; j<=i; j++) 
         p[k++] = gsl_pow_int(x, i-j) * gsl_pow_int(y, j) ;
      if ( i>0 && GSL_IS_ODD(i) )
         p[k++] = gsl_pow_int( gsl_hypot(x, y), i) ;
   }

#else

   int    i ;
   double x2, y2, x3, y3, x4, y4, x5, y5, r ;

   x2 = x * x ;
   y2 = y * y ;
   x3 = x * x2 ;
   y3 = y * y2 ;
   x4 = x2 * x2 ;
   y4 = y2 * y2 ;
   x5 = x3 * x2 ;
   y5 = y3 * y2 ;
   r = hypot(x, y) ;

   for (i=0; i<np; i++) p[i] = 0.0 ;

   p[0]  =  1 ;          /* 0th order */
   p[1]  =  x ;
   p[2]  =  y ;
   p[3]  =  r ;          /* 1st order */
   p[4]  = x2 ;
   p[5]  = x*y ;
   p[6]  = y2 ;          /* 2nd order */
   p[7]  = x3 ;
   p[8]  = x2*y ;
   p[9]  = x*y2 ;
   p[10] = y3 ;
   p[11] = r*r*r ;       /* 3rd order */
   p[12] = x4      ;
   p[13] = x3 *  y ;
   p[14] = x2 * y2 ;
   p[15] =  x * y3 ;
   p[16] =      y4 ;     /* 4th order */
   p[17] = x5      ;
   p[18] = x4 *  y ;
   p[19] = x3 * y2 ;
   p[20] = x2 * y3 ;
   p[21] =  x * y4 ;
   p[22] =      y5 ; 
   p[23] = r*r*r*r*r ;    /* 5th order */

#endif
   return;
}

void 
model_set_cp(PlateMod *model)
{

#ifdef HAVE_GSL
   gsl_matrix      *a, *inv ;
   gsl_permutation *p ;
   int              s ;
   size_t           i, j ;

   /* Allocate memory for working arrays */
   a   = gsl_matrix_alloc ((size_t)2, (size_t)2) ;
   inv = gsl_matrix_alloc ((size_t)2, (size_t)2) ;
   p   = gsl_permutation_alloc ((size_t)2) ;

   for (i=0; i<2; i++)
      for (j=0; j<2; j++)
         gsl_matrix_set (a, i, j, model->pc[i][j]) ;

   gsl_linalg_LU_decomp (a, p, &s) ;
   gsl_linalg_LU_invert (a, p, inv) ;

   for (i=0; i<2; i++)
      for (j=0; j<2; j++)
         model->cp[i][j] = gsl_matrix_get (inv, i, j) ;

   gsl_matrix_free (a) ;
   gsl_matrix_free (inv) ;
   gsl_permutation_free (p) ;
#endif
}


