/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: platemodel.h
 * Author 	: Paolo Montegriffo
 * Created on	: September, 2001	
 * Language	: ANSI C
 * Description	: Type definitions for plate model structures
 * Version      : 2.1.4 - February 15, 2002
 *--------------------------------------------------------------------------*/

#ifndef PLATEMODEL_H
#define PLATEMODEL_H 1

#include "common.h"

BEGIN_C_DECLS

#include "cataxcorr.h"

#define FN_MAXPAR  72 /* Maximum number of coefficents for plate function*/

typedef enum 
{
   BUSY = -1,
   INIT = 0,
   LOW_ALIGN,
   HIGH_ALIGN,
   MODEL_FITTED
} model_stat ;

typedef struct model_type_tag
{
  void      (*PlateFunc)() ;
  const int  *fncoef ;
  int         maxord ;
} model_type ;


typedef struct PlateMod_tag {
   int        fitmodel ;
   model_type set ;
   int        order ;
   double     xfit_coeff[FN_MAXPAR] ;
   double     yfit_coeff[FN_MAXPAR] ;
   int        mask[FN_MAXPAR] ;
   int        nx_coeff ;
   int        ny_coeff ;
   double     dxdev ;
   double     dydev ;
   double     xchisq ;
   double     ychisq ;
   double     dmag ;
   double     crpix[2] ;      /* Pixel coord at reference point */
   double     crval[2] ;      /* Coord value at reference point */
   double     cdelt[2] ;
   double     pc[2][2] ;
   double     cp[2][2] ;
   double     crota2 ;
   int        frm1 ;
   StarMatch *match ;
   int        nmatch ;
   boolean    spheric ;
   model_stat status ;
   double     x, y ;
   boolean    saved ;
   double     dx, dy ;
} PlateMod ;

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

extern int  FitPlate             (Star **s1, Star **s2, Match *match, int nmatch, 
                                  PlateMod *model, double *chisq) ;
extern int  model_fit_distortion (StarMatch *match, int nmatch, 
                                  PlateMod *model, double *chisq) ;

extern void LINPlateFunc         (double x, double y, double p[], int np) ;

extern void SAOPlateFunc         (double x, double y, double p[], int np) ;

extern void POLYPlateFunc        (double x, double y, double p[], int np) ;

extern void POLYRADPlateFunc     (double x, double y, double p[], int np) ;

extern void LEGEPlateFunc        (double x, double y, double p[], int np) ;

extern void model_set_cp         (PlateMod *model) ;

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#define NullPlateMod (PlateMod*)NULL

enum {
  LIN = 0,
  SAO,
  POLY,
  POLYRAD,
  LEGENDRE,
  NUM_MODEL
} ;

extern const int  minstar[] ;
extern const int  LINncoef[] ;
extern const int  SAOncoef[] ;
extern const int  POLYncoef[] ;
extern const int  POLYRADncoef[] ;
extern const int  LEGEncoef[] ;
extern const model_type m_type[NUM_MODEL] ;

#ifdef debug
#define MODEL_CHECK(a) \
{\
   int  i_i; \
   char lab[40] ; \
   if (verbose > 4) { \
      printf("%-8s= %#20.14f / %-47s\n", "CRVAL1", (a)->crval[0], \
         "Coordinate at reference pixel") ; \
      printf("%-8s= %#20.14f / %-47s\n", "CRVAL2", (a)->crval[1], \
         "Coordinate at reference pixel") ; \
      printf("%-8s= %#20.14g / %-47s\n", "CRPIX1", (a)->crpix[0], \
         "Reference pixel") ; \
      printf("%-8s= %#20.14g / %-47s\n", "CRPIX2", (a)->crpix[1], \
         "Reference pixel") ; \
      printf("%-8s= %#20.14g / %-47s\n", "CDELT1", (a)->cdelt[0], \
         "Coordinate increment per pixel") ; \
      printf("%-8s= %#20.14g / %-47s\n", "CDELT2", (a)->cdelt[1], \
         "Coordinate increment per pixel") ; \
      printf("%-8s= %#20.14g / %-47s\n", "PC1_1", (a)->pc[0][0], \
         "Pixel transformation matrix") ; \
      printf("%-8s= %#20.14g / %-47s\n", "PC1_2", (a)->pc[0][1], " ") ; \
      printf("%-8s= %#20.14g / %-47s\n", "PC2_1", (a)->pc[1][0], " ") ; \
      printf("%-8s= %#20.14g / %-47s\n", "PC2_2", (a)->pc[1][1], " ") ; \
      for(i_i=0; i_i<(a)->nx_coeff; i_i++) {\
         sprintf(lab, "COi_%d", i_i+1) ; \
         printf("%-8s= %#20.14g ; %#20.14g / %-22s\n", lab, \
            (a)->xfit_coeff[i_i],(a)->yfit_coeff[i_i], \
            (i_i==0 ? "XY plate model coeff" : " " )) ; \
      } \
   }\
}
#else
#define MODEL_CHECK(a) 
#endif

END_C_DECLS

#endif     /* !PLATEMODEL_H */
/*--------------------------------------------------------------------------*/
