/*----------------------------------------------------------------------------
 * CataPack
 *
 * Authors: Paolo Montegriffo <paolo.montegriffo@oabo.inaf.it>
 *
 * Copyright (C) 2001-2006 INAF - Osservatorio Astronomico di Bologna.
 *
 * This program is free software; you can redistribute it and/or modify
 * it under the terms of the GNU General Public License as published by
 * the Free Software Foundation; either version 2 of the License, or
 * (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA 02111-1307, USA.
 *
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 					      Bologna Astronomical Observatory
 *----------------------------------------------------------------------------
 * File name 	: starlist.c
 * Author 	: Paolo Montegriffo
 * Created on	: February, 2002	
 * Language	: ANSI C
 * Description	: Functions for stalist and coinctab manipulation
 * Version      : 2.1.4 - February 14, 2002
 *--------------------------------------------------------------------------*/

/*----------------------------------------------------------------------------
 * 								Includes
 *--------------------------------------------------------------------------*/

#if HAVE_CONFIG_H
#  include "config.h"
#endif

#include "common.h"

#include "cataxcorr.h"
#include "cxc_decode.h"
#include "catalib.h"
#include "cp_messages.h"
#include "coordinates.h"
#include "starlist.h"

/*----------------------------------------------------------------------------
 * 							Function prototypes
 *--------------------------------------------------------------------------*/

static Star     **star_new            (size_t     n) ;

static int        read_coord          (char      *token,
                                       double    *x) ;
static int        write_coord         (char     **token,
                                       double     x,
                                       double     fc) ;
static int        star_list_check     (StarList  *star) ;

static int        star_list_limits    (StarList  *list,
                                       double    *xmin,
                                       double    *xmax,
                                       double    *ymin,
                                       double    *ymax) ;
static StarList  *star_list_cpy       (StarList  *in) ;

static int        stack_get_n         (Star     **s,
                                       int        ns) ;

/*----------------------------------------------------------------------------
 * 								Defines
 *--------------------------------------------------------------------------*/

#ifndef MIN
#define MIN(a,b) (((a) <= (b)) ? (a) : (b))
#endif
#ifndef MAX
#define MAX(a,b) (((a) >= (b)) ? (a) : (b))
#endif
#ifndef ABS
#define ABS(a) ((a) < 0 ? (-(a)) : (a))
#endif
#ifndef SQR
#define SQR(a) ((a)*(a))
#endif

/*----------------------------------------------------------------------------
 * 							Global variables	
 *--------------------------------------------------------------------------*/

extern g_pars    gparam ;

extern double    IntNull   ;
extern char      command[] ;
extern boolean   spheric ;

/*----------------------------------------------------------------------------
 * 							Private types
 *--------------------------------------------------------------------------*/

struct si {
   Star *s ;
   int   i ;
} ;

/*----------------------------------------------------------------------------
 * 							Private functions
 *--------------------------------------------------------------------------*/

static int sort_si (const struct si *s1, const struct si *s2)
{ 
   if ( s1->s->mag > s2->s->mag ) 
      return 1 ; 
   else 
      return -1 ; 
}

static int sort_star_idx (const Star **d1, const Star **d2)
{ 
   if ( (*d1)->index > (*d2)->index ) 
      return 1 ; 
   else 
      return -1 ;
}

static int sort_star_yprj (const Star **d1, const Star **d2)
{ 
   if ( (*d1)->yprj > (*d2)->yprj ) 
      return 1 ; 
   else 
      return -1 ;
}

static int sort_star_yrot (const Star **d1, const Star **d2)
{ 
   if ( (*d1)->yrot > (*d2)->yrot ) 
      return 1 ; 
   else 
      return -1 ;
}

static int sort_star_match_yprj (const StarMatch *d1, const StarMatch *d2)
{ 
   if ( (d1)->s1->yprj > (d2)->s1->yprj ) 
      return 1 ; 
   else 
      return -1 ;
}

/*----------------------------------------------------------------------------
 *							Function codes
 *--------------------------------------------------------------------------*/

static Star **
star_new (size_t n)
{
   size_t i ;
   Star **ret = (Star**)NULL ;
   ret = (Star**) malloc (n * sizeof(Star*)) ;
   if (ret == (Star**)NULL) return ret ;
   for (i = 0; i < n; i++) {
      ret[i] = (Star*) malloc (sizeof(Star)) ;
      ret[i]->id = (char*)NULL ;
   }
   return ret ;
}

Star **
star_destroy (Star **in, size_t n)
{
   size_t i ;
   for (i = 0; i < n; i++) {
      if (in[i]->id != (char*)NULL) 
         free (in[i]->id) ;
      if (in[i])
         free (in[i]) ;
   }
   if (in)
      free (in) ;
   return (Star**)NULL ;
}

/* Duplicate Star pointers vector, allocating memory */
Star **
star_ptr_dup (Star **in, size_t n)
{
   Star   **ret = (Star**) malloc (n * sizeof(Star*)) ;
   memcpy (ret, in, n*sizeof(Star*)) ;
   return ret ;
}

StarList *
star_list_new (size_t n)
{
   StarList *ret = (StarList*)NULL ;

   ret = (StarList*) calloc ((size_t)1, sizeof(StarList)) ;
   if (ret == NullStarList) return ret ;

   ret->data = star_new (n) ;
   if (ret->data == (Star**)NULL) {
      free (ret) ;
      return (StarList*)NULL ;
   }
   ret->nstar = n ;
   ret->spheric = ret->out_sph = FALSE ;
   ret->proj = ret->saved = FALSE ;
   ret->inv_density = 0.0 ;
   ret->param = NULL ;

   return ret ;
}

StarList * 
star_list_destroy (StarList *in)
{
   if (in->data != (Star**)NULL)
      star_destroy (in->data, in->nstar) ;
   /*if (in->param) free (in->param) ;*/
   if (in != (StarList*)NULL) 
      free (in) ;
   return (StarList*)NULL ;
}

StarList * 
star_list_free (StarList *in)
{
   if (in->data != (Star**)NULL)
      free (in->data) ;
   if (in != (StarList*)NULL) 
      free (in) ;
   return (StarList*)NULL ;
}

/* Duplicate a StarList: alloc mem for ret star */
StarList *
star_list_dup (StarList *in)
{
   StarList *ret = (StarList*)NULL ;
   size_t  i ;

   /* Create the returning star list */
   ret = star_list_new (in->nstar) ;
   if (ret == (StarList*)NULL) 
      return ret ;
   /* Copy the star list content */
   for (i = 0; i < in->nstar; i++) {
      memcpy (ret->data[i], in->data[i], sizeof(Star)) ;
      ret->data[i]->id = strdup (in->data[i]->id) ;
   }

   ret->spheric = in->spheric ;
   ret->out_sph = in->out_sph ;
   ret->origin[0] = in->origin[0] ;
   ret->origin[1] = in->origin[1] ;
   ret->proj  = in->proj ;
   ret->saved = in->saved ;
   ret->param = (def_pars*) malloc (sizeof(def_pars)) ;
   memcpy (ret->param, in->param, sizeof(def_pars)) ;

   return ret ;
}

/* Duplicate StarList allocating memory for Star pointers vector */
static StarList *
star_list_cpy (StarList *in)
{
   StarList *ret = (StarList*)NULL ;

   ret = (StarList*) malloc (sizeof(StarList)) ;
   if ( ret == (StarList*)NULL ) 
      return ret ;
   ret->data = star_ptr_dup (in->data, in->nstar) ;

   ret->nstar = in->nstar ;
   ret->spheric = in->spheric ;
   ret->out_sph = in->out_sph ;
   ret->origin[0] = in->origin[0] ;
   ret->origin[1] = in->origin[1] ;
   ret->proj = in->proj ;
   ret->saved = in->saved ;
   ret->param = in->param ;

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_read
 * In        : 
 * Out       : 
 * Purpouse  : 
 * Note      : *** Se un ID e' una stringa non funziona ******
 *             
 *---------------------------------------------------------------------------*/

StarList *
star_list_read (Finfo *f_info, def_pars *param)
{
   StarList *ret = NullStarList ;
   int       i, j, n, skip = 0, sx, sy ;
   char    **toks = (char**)NULL ;
   char      line[MAX_RECORD_LENGTH] ;
   FILE     *in ;
   int       mind = f_info->m_pos ;
   int       xind = f_info->x_pos ;
   int       yind = f_info->y_pos ;
   double    x, y ;

   CPKDEBUG ("star_list_read") ;

   if (xind < 0 || yind < 0) {
      cp_error ("Error in ooordinates specification: aborting") ;
      return ret ;
   }
   if (mind < 0)
      cp_warning ("Missing comparison magnitude for list %s",f_info->filename) ;

   /* Allocate memory */
   ret = star_list_new ((size_t)(f_info->n_rows)) ;
   if (ret == NullStarList) {
      cp_error ("Cannot alloc memory for star list %s: aborting",f_info->filename) ;
      return ret ;
   }

   CXC_MESSAGE (1, 2, "Reading frame file %s\n", f_info->filename) ;
   /* Open the file */
   in = fopen (f_info->filename, "r") ;

   /* Skip the header */
   for (j = 0; j < f_info->n_head; j++) fgets (line, MAX_RECORD_LENGTH, in) ;

   /* Read file columns */
   j = 0 ;
   while (j < f_info->n_rows && fgets (line, MAX_RECORD_LENGTH, in) != NULL) {
      toks = tokenize (line," ",&n) ;
      if (n == f_info->n_cols && strncmp (toks[0], "#", (size_t)1)) {

         /* Get ID string */
         ret->data[j]->id = (char*) malloc ((strlen (toks[0])+1)*sizeof(char)) ;
         strcpy (ret->data[j]->id, toks[0]) ;

         /* Get Magnitude */
         if (mind>=0) ret->data[j]->mag = atof (toks[mind]) ;
         else ret->data[j]->mag = IntNull ;

         /* Checks for null values */
         for (i = 0; i < gparam.N_NullVal; i++) 
            if (ret->data[j]->mag == gparam.NullValues[i]) 
               ret->data[j]->mag = IntNull ;
         ret->data[j]->mcal = ret->data[j]->mag ;

         /* Get positions */
         sx = sy = 0 ;
         if (f_info->fmt[xind].type != 's') 
            x = atof (toks[xind]) ;
         else {
            sx = read_coord (toks[xind], &x) ;
            x = x * 360.0 / 24.0 ;
         }
         if (f_info->fmt[yind].type != 's') 
            y = atof (toks[yind]) ;
         else
            sy = read_coord (toks[yind], &y) ;

         if (sx||sy) {
            (void)free_carray (toks, n) ;
            star_list_destroy (ret) ;
            cp_error ("Unrecognized coordinates format in list %s: aborting",
            f_info->filename) ;
            return NullStarList ;
         }
         
         ret->data[j]->xpix = ret->data[j]->xprj = ret->data[j]->xrot = x ;
         ret->data[j]->ypix = ret->data[j]->yprj = ret->data[j]->yrot = y ;
         ret->data[j]->flag = 0 ;
	 ret->data[j]->p = NULL ;
         j++ ;
      }
      else 
         if (n > 0) {
            skip++ ;
            cp_warning ("Skipping: '%s'\n", line) ;
         }
      (void)free_carray (toks, n) ;
   }
   fclose (in) ;
   if (skip != 0) cp_warning ("Skipped %d not-empty lines", skip) ;

   /* Put index values */
   for (j = 0; j < f_info->n_rows; j++) ret->data[j]->index = j ;

   if (f_info->spheric) ret->spheric = TRUE ;
   ret->out_sph = ret->spheric ;
   ret->proj = FALSE ;
   ret->saved = FALSE ;
   if (param == NULL)
      ret->param = file_info_set_default (ret->spheric) ;
   else
      ret->param = param;

   cp_message (0,3,"Read %d records\n", (int)ret->nstar) ;
   STARCHECK (ret->data, ret->nstar) ;

   star_list_center (ret, &ret->origin[0], &ret->origin[1]) ;

   /* Check for uniq id numbers */
   /* if( star_list_check(ret) ) {
      star_list_destroy(ret) ;
      ret = NullStarList ;
      cp_error("Star list %s has not uniq id numbers: aborting",f_info->filename) ;
   } */

   return ret ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_save
 * In        : Finfo , Star List, plate model struct
 * Out       : 
 * Purpouse  : 
 * 
 *---------------------------------------------------------------------------*/

void 
star_list_save (char *filename, Finfo *f_info, StarList *list)
{
   char   **toks, *p1 ;
   char     line[MAX_RECORD_LENGTH], lineout[MAX_RECORD_LENGTH], desc[MAX_RECORD_LENGTH] ;
   int      i, j, n ;
   FILE    *in, *out ;
   size_t   nn ;
   Star   **s ;

   CPKDEBUG ("star_list_save") ;

   if (list->saved) 
      return ;

   s = star_ptr_dup (list->data, list->nstar) ;
   qsort ((void*)s, list->nstar, sizeof(Star*), 
          (int (*)(const void *, const void *))sort_star_idx) ;

   /* Open the files */
   cp_message (1,2,"Saving current rotated star list '%s'\n", filename) ;
   in  = fopen (f_info->filename, "r") ;
   out = fopen (filename, "w") ;

   /* Add Comment line */
   sprintf (line,"#COMMAND= %s\n",command) ;
   fputs (line, out) ;

   /* Copy the header except for descriptors header line */
   for (j = 0; j < f_info->n_head; j++) {
      fgets (line, MAX_RECORD_LENGTH, in) ;
      if (strncmp (line, "#N ", (size_t)3))
         fputs (line, out) ;
   }

   /* Copy file columns */
   strcpy (desc, "#N ") ;
   j = 0;
   /* for (j = 0; j < f_info->n_rows; j++) { */
   while (j < f_info->n_rows) {

      /* Read line, tokenize it, get column spacing, format output column */
      strcpy (lineout, "") ;
      fgets (line, MAX_RECORD_LENGTH, in) ;
      p1 = line ;
      toks = tokenize (line," ",&n) ;

      if (n == f_info->n_cols && strncmp (toks[0], "#", (size_t)1)) {

	 for (i = 0; i < n; i++) {
            nn = (size_t)(strstr (p1, toks[i]) - p1) + strlen (toks[i]) ;
            if (i == f_info->x_pos && gparam.append == FALSE) 
               write_coord (&toks[i], s[j]->xpix, 24.0/360.0) ;
            else if (i == f_info->y_pos && gparam.append == FALSE) 
               write_coord (&toks[i], s[j]->ypix, 1.0) ;
            sprintf (lineout, "%s %*s", lineout, (int)nn-1, toks[i]) ;
            /* append RA&Dec after x y columns */
            if (i == f_info->y_pos && gparam.append == TRUE) {
               write_coord (&toks[i], s[j]->xpix, 24.0/360.0) ;
               sprintf (lineout, "%s %*s", lineout, (int)nn-1, toks[i]) ;
               write_coord (&toks[i], s[j]->ypix, 1.0) ;
               sprintf (lineout, "%s %*s", lineout, (int)nn-1, toks[i]) ;
            }

            /* Format descriptors record */
            if (j == 0) {
               if (i == f_info->x_pos && spheric && gparam.append == FALSE)
        	  sprintf (desc, "%s %*s", desc, (int)nn-1, "RA") ;
               else if (i == f_info->y_pos && spheric && gparam.append == FALSE) 
        	  sprintf (desc, "%s %*s", desc, (int)nn-1, "Dec") ;
               else 
        	  sprintf (desc, "%s %*s", desc, (int)nn-1, f_info->labels[i]) ;
               if (i == f_info->y_pos && spheric && gparam.append == TRUE)
        	  sprintf (desc, "%s%15s%15s", desc, "RA", "Dec") ;
            }
            p1 += nn ;
	 }
	 /* Put descriptors header line */
	 if (j == 0) {
            strip_newline (desc) ;
            strcat (desc, "\n#\n\0") ;
            fputs (desc, out) ;
	 }
	 j++;
      }
      else 
         strncpy (lineout, line, strlen (line) + 1);

      (void) free_carray (toks, n) ;
      strip_newline (lineout) ;
      strcat (lineout,"\n\0") ;

      fputs (lineout,out) ;
   }
   
   fclose (in) ;   
   fclose (out) ;
   list->saved = TRUE ;
   free (s) ;
}

static int
read_coord (char *token, double *x)
{
   char   **toks = (char**)NULL ;
   double   d, s ;
   int      n ;

   *x = 0.0 ;
   toks = tokenize (token,":",&n) ;   
   if (n == 3) {
      d = atof (toks[0]) ;
      /* s = (d >= 0.0) ? 1.0 : -1.0 ; */
      s = (strncmp (toks[0], "-", 1L) == 0) ? -1.0 : 1.0;
      *x = s * (ABS (d) + 
                atof (toks[1])/60.0 + 
                atof (toks[2])/3600.0 ) ;
      gparam.deg_fmt = FALSE ;
   }
   else {
      *x = atof (token) ;
      gparam.deg_fmt = TRUE ;
   }

   (void) free_carray (toks, n) ;
   if (n == 3) return 0 ;
   else return 1 ;
}

static int
write_coord (char **token, double x, double fc)
{
   char    outstring[25], dec[6] ;
   double  h, m, s, sg ;
   size_t  nd ;
   char   *pp = strstr (*token, ".") ;

   if (spheric)
      nd = 8 ;
   else
      nd = (pp != (char*)NULL) ? MAX (strlen (*token) - (size_t)(pp - *token), 2) : 2 ;

   if (gparam.deg_fmt)
      sprintf (outstring, "%.*f", (int)nd-1, x) ;
   else {
      sg = (x >= 0.0) ? 1.0 : -1.0 ;
      x = ABS (x) ;
      x = x*fc ;
      h = floor (x) ;
      x = (x - h) * 60.0 ;
      m = floor (x) ;
      x = (x - m) * 60.0 ;
      s = floor (x) ;
      x = (x - s) ;
      sprintf (dec, "%.*f", ((fc == 1.0) ? 2 : 3), x) ;
      if (sg > 0) 
         sprintf (outstring, "% 02d:%02d:%02d%-4s", (int)h, (int)m, (int)s, 
               strstr (dec, ".")) ;
      else
         sprintf (outstring, "-%02d:%02d:%02d%-4s", (int)h, (int)m, (int)s, 
               strstr (dec, ".")) ;
   }

   nd = strlen (outstring) + 1 ;
   if (nd > strlen (*token))
      *token = (char*) realloc (*token, nd*sizeof(char)) ;
   strcpy (*token, outstring) ;

   return 0 ;
}

static int int_sort (const void * d1, const void * d2)
{ 
   if (*(const int*)d1 > *(const int*)d2) return 1 ; else return -1 ; 
}

/***FIXME: USELESS ***/
static int 
star_list_check (StarList *star)
{
   int    *tmp = (int*)NULL ;
   size_t  i ;

   /* Allocate mem */
   tmp = (int*) calloc (star->nstar, sizeof(int)) ;
   /* Cp ID numbers in tmp vec */
   for (i = 0; i < star->nstar; i++)
      /* tmp[i] = star->data[i].id */ ;
   /* Sort ID numbers */
   qsort (tmp, star->nstar, sizeof(int), int_sort) ;
   /* Scan for multiple ID */
   i=0 ;
   while (i < star->nstar-(size_t)1 && tmp[i] != tmp[i+1])
      i++ ;
   /* Free mem */
   free (tmp) ;

   if (i != star->nstar-(size_t)1) return 1 ;
   return 0 ;
}

double 
star_list_density (StarList *list)
{
   size_t  i ;
   double  fx1, fx2, fy1, fy2 ;

   fx1 = fx2 = list->data[0]->xrot ;
   fy1 = fy2 = list->data[0]->yrot ;

   for (i = 0; i < list->nstar; i++) {
      fx1 = MIN (fx1, list->data[i]->xrot) ;
      fx2 = MAX (fx2, list->data[i]->xrot) ;
      fy1 = MIN (fy1, list->data[i]->yrot) ;
      fy2 = MAX (fy2, list->data[i]->yrot) ;
   }
   return ((fx2-fx1)*(fy2-fy1) / MAX (list->nstar, 1)) ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_get_record
 * In        : Finfo , record index
 * Out       : 
 * Purpouse  : 
 * 
 *---------------------------------------------------------------------------*/

void
star_list_get_record (Finfo *f_info, int n, char *buffer)
{
   FILE      *in ;
   char       line[MAX_RECORD_LENGTH] ;
   int        i ;

   in = fopen (f_info->filename, "r") ;
   /* Skip the header */
   for (i = 0; i < f_info->n_head; i++) 
      fgets (line, MAX_RECORD_LENGTH, in) ;
   /* Skip other records */
   for (i = 0; i < n ; i++) 
      fgets (line, MAX_RECORD_LENGTH, in) ;

   fgets (line, MAX_RECORD_LENGTH, in) ;
   strip_newline (line) ;
   sprintf (buffer, " %s ", line) ;
   fclose (in) ;
}

/*-----------------------------------------------------------------------------
 * Function  : stack_init
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
 *---------------------------------------------------------------------------*/

StackTab *
stack_init (StarList *list, int n, int nfiles)
{
   StackTab  *stack = (StackTab*)NULL ;
   size_t     i ;
   int        status = 1 ;

   CPKDEBUG ("stack_init") ;

   stack = (StackTab*) malloc (sizeof(StackTab)) ;
   if (!stack) goto exit ;

   stack->tab = (Stack**) malloc (list->nstar*sizeof(Stack*)) ;
   if (!stack->tab) goto exit ;

   for (i = 0; i < list->nstar; i++) {
      stack->tab[i] = (Stack*) malloc (sizeof(Stack)) ;
      stack->tab[i]->data = (Star**) calloc ((size_t)nfiles, sizeof(Star*)) ;
      if (!stack->tab[i] || !stack->tab[i]->data) goto exit ;
      stack->tab[i]->data[n] = list->data[i] ;
      stack->tab[i]->id = i+1 ;
      stack->tab[i]->ns  = bin_switch_on (0, n) ;
      stack->tab[i]->ns2 = bin_switch_on (0, n-32) ;
   }

   stack->nfiles = nfiles ;
   stack->nf  = bin_switch_on (0, n) ;
   stack->nf2 = bin_switch_on (0, n-32) ;
   stack->n = 1 ;
   stack->ref = star_list_cpy (list) ;
   if (!stack->ref) goto exit ;

   /* Link ref stars with coinc tab entries */
   for (i = 0; i < list->nstar; i++)
      stack->ref->data[i]->p = stack->tab[i] ;

   status = 0 ;

   exit:

   if (status == 1) {
      cp_error ("Cannot alloc memory for initialize coinc structure: aborting") ;
      stack_destroy (stack) ;
      return NULL ;
   }

   return stack ;
}

/*-----------------------------------------------------------------------------
 * Function  : stack_destroy
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
 *---------------------------------------------------------------------------*/

void 
stack_destroy (StackTab *stack)
{
   size_t i ;

   if (stack == (StackTab *)NULL) 
      return ;
   
   for (i = 0; i < stack->ref->nstar; i++) {
      if (stack->tab[i]->data) free (stack->tab[i]->data) ;
      if (stack->tab[i]) free (stack->tab[i]) ;
   }
   if (stack->tab) free (stack->tab) ;
   if (stack->ref) free (stack->ref) ;
   free (stack) ;
   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : coinc_tab_update
 * In        : 
 * Out       :  
 * Purpouse  :  
 * Note      : 
 *---------------------------------------------------------------------------*/

int 
stack_add (StackTab **stack, StarList *list, int nfile, 
           StarMatch *match, int nmatch)
{
   int       i, j, status = 1 ;
   size_t    newdim ;
   Stack    *tab ;

   CPKDEBUG ("stack_add") ;

   /* Some checks first */
   if (nfile >= (*stack)->nfiles) {
      cp_error ("Illegal file number request") ;
      return status ;
   }
   if (bin_check_n ((*stack)->nf, nfile) ||
       bin_check_n ((*stack)->nf2, nfile-32)) {
      cp_error ("Illegal  request: File %d already added", nfile) ;
      return status ;
   }

   /* Reallocate memory for coinc struct to recover new stars */
   newdim = (*stack)->ref->nstar + list->nstar - nmatch ;

   (*stack)->tab = (Stack**) realloc ((*stack)->tab, newdim*sizeof(Stack*)) ;
   (*stack)->ref->data = (Star**) realloc ((*stack)->ref->data, newdim*sizeof(Star*)) ;
   if (!(*stack)->tab || !(*stack)->ref->data) goto exit ;

   for (i = 0; i < (int)list->nstar - nmatch; i++) {
      j = i + (*stack)->ref->nstar ;
      (*stack)->tab[j] = (Stack*) malloc (sizeof(Stack)) ;
      (*stack)->tab[j]->data = (Star**) calloc ((size_t)(*stack)->nfiles, sizeof(Star*)) ;
      if (!(*stack)->tab[j] || !(*stack)->tab[j]->data) goto exit ;
   }

   /* Link matching stars */
   for (i = 0; i < list->nstar; i++)
      list->data[i]->p = NULL ;
   for (i = 0; i < nmatch; i++)
      match[i].s2->p = match[i].s1->p ;

   /* Fill the coinc tab */
   j = (int)(*stack)->ref->nstar ;
   for (i = 0; i < list->nstar; i++) {
      if (list->data[i]->p) {
         /* Matched entry */
         tab = (Stack*)list->data[i]->p ;
         tab->data[nfile] = list->data[i] ;
         tab->ns  = bin_switch_on (tab->ns,  nfile) ;
         tab->ns2 = bin_switch_on (tab->ns2, nfile-32) ;
         (*stack)->n = MAX ((*stack)->n, 
               stack_get_n (tab->data, (*stack)->nfiles)) ;
      }
      else {
         /* New entry */
         (*stack)->ref->data[j] = (*stack)->tab[j]->data[nfile] = list->data[i] ;
         (*stack)->tab[j]->id = j+1 ;
         (*stack)->tab[j]->ns  = bin_switch_on (0, nfile) ;
         (*stack)->tab[j]->ns2 = bin_switch_on (0, nfile-32) ;
         (*stack)->ref->data[j]->p = list->data[i]->p = (*stack)->tab[j] ;
         j++ ;
      }
   }
   (*stack)->ref->nstar = newdim ;
   (*stack)->nf  = bin_switch_on ((*stack)->nf,  nfile) ;
   (*stack)->nf2 = bin_switch_on ((*stack)->nf2, nfile-32) ;

   status = 0 ;

   exit:

   if (status == 1) {
      cp_error ("Cannot alloc memory for updating coinc structure: aborting") ;
      stack_destroy ((*stack)) ;
   }

   return status ;
}

static int 
stack_get_n (Star **s, int ns)
{
   int i, n ;
   for (n = i = 0; i < ns; i++)
      if (s[i])
         n++ ;
   return n ;
}

int 
stack_remove (StackTab **stack, int nfile)
{
   int       i, k, status = 1 ;
   size_t    j ;
   Star    **ref_tmp ;
   Stack   **cnc_tmp ;

   CPKDEBUG ("stack_remove") ;

   if (nfile >= (*stack)->nfiles) {
      cp_error ("Illegal file number request") ;
      return status ;
   }

   /* Fill common stars fields */
   ref_tmp = (Star**)  malloc ((*stack)->ref->nstar*sizeof(Star*)) ;
   cnc_tmp = (Stack**) malloc ((*stack)->ref->nstar*sizeof(Stack*)) ;
   j = 0 ;
   for (i = 0; i < (*stack)->ref->nstar; i++) {

      if ((*stack)->tab[i]->data[nfile]) {
         (*stack)->tab[i]->data[nfile]->p = NULL ;
         (*stack)->tab[i]->data[nfile] = NULL ;
         (*stack)->tab[i]->ns  = bin_switch_off ((*stack)->tab[i]->ns, nfile) ;
         (*stack)->tab[i]->ns2 = bin_switch_off ((*stack)->tab[i]->ns2, nfile-32) ;
      }

      if ((*stack)->tab[i]->ns != 0 /* && (*stack)->tab[i]->ns2 != 0 */) {
         /* Save reference to non-empty records */
         ref_tmp[j] = (*stack)->ref->data[i] ;
         cnc_tmp[j] = (*stack)->tab[i] ;
         j++ ;
      }
      else 
         free ((*stack)->tab[i]) ;
   }

   ref_tmp = (Star**)  realloc (ref_tmp, j*sizeof(Star*)) ;
   cnc_tmp = (Stack**) realloc (cnc_tmp, j*sizeof(Stack*)) ;

   free ((*stack)->ref->data) ;
   free ((*stack)->tab) ;

   (*stack)->ref->data = ref_tmp ;
   (*stack)->tab = cnc_tmp ;
   (*stack)->ref->nstar = j ;
   (*stack)->n = 0 ;

   /* Reindex stars */
   for (i = 0; i < (int)j; i++) {
      (*stack)->tab[i]->id = i+1 ;
      k = 0 ;
      while (k < (*stack)->nfiles && !(*stack)->tab[i]->data[k]) k++ ;
      (*stack)->ref->data[i] = (*stack)->tab[i]->data[k] ;
      (*stack)->n = MAX ((*stack)->n, 
            stack_get_n ((*stack)->tab[i]->data, (*stack)->nfiles)) ;
   }
   (*stack)->nf  = bin_switch_off ((*stack)->nf,  nfile) ;
   (*stack)->nf2 = bin_switch_off ((*stack)->nf2, nfile-32) ;

   status = 0 ;

   return status ;
}

/*-----------------------------------------------------------------------------
 * Function  : coinc_tab_save
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
 *---------------------------------------------------------------------------*/

void 
coinc_tab_save (StackTab *stack, Finfo **f_info, char *outfile )
{
   int    i, j ;
   char   line[MAX_RECORD_LENGTH] ;
   FILE * out;

   CPKDEBUG ("coinc_tab_save") ;

   out = fopen (outfile, "w") ;
   strcpy (line, "# \0") ;
   sprintf (line, "%s %5s", line, "ID") ;
   sprintf (line, "%s %10s", line, f_info[0]->filename) ;
   for (i = 1; i < stack->nfiles; i++) {
      sprintf (line, "%s %10s", line, f_info[i]->filename) ;
      if (gparam.fitmodel != NONE) sprintf (line, "%s.rot", line) ; /** FIXME **/
   }
   sprintf (line, "%s %5s\n#", line, "NS") ;
   fprintf (out, "%s\n",line) ;

   for (i = 0; i < (int)stack->ref->nstar; i++) {
      strcpy (line, " \0") ;
      sprintf (line, "%s %6d", line, stack->tab[i]->id) ;
      for (j = 0; j < stack->nfiles; j++) 
	 if (stack->tab[i]->data[j])
            sprintf (line, "%s %10s", line, stack->tab[i]->data[j]->id) ;
         else
            sprintf (line, "%s %10s", line, "0") ;
      sprintf (line, "%s %10u", line, stack->tab[i]->ns) ;
      if (stack->nfiles > 32) {
	 sprintf (line, "%s:%u", line, stack->tab[i]->ns2) ;
      }
      fprintf (out, "%s\n", line) ;
   }
   fclose (out) ;

   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_get_coinc
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
 *---------------------------------------------------------------------------*/

StarList *
star_list_get_coinc (StackTab *stack)
{
   return stack->ref ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_center
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
 *---------------------------------------------------------------------------*/

int 
star_list_center (StarList *list, double *xcenter, double *ycenter)
{
   double  x0, x1, y0, y1 ;
   int      i ;

   x0 = x1 = list->data[0]->xpix ;
   y0 = y1 = list->data[0]->ypix ;
   for (i = 1; i < (int)list->nstar; i++) {
      x0 = MIN (x0, list->data[i]->xpix) ;
      x1 = MAX (x1, list->data[i]->xpix) ;
      y0 = MIN (y0, list->data[i]->ypix) ;
      y1 = MAX (y1, list->data[i]->ypix) ;
   }
   *xcenter = ( x1 + x0 ) / 2.0 ;
   *ycenter = ( y1 + y0 ) / 2.0 ;
   return 0 ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_flag_crowded
 * In        : star list, position tolerance
 * Out       : modified star list
 * Purpouse  : flag crowded stars to 1 except for the brightest of each group
 * Note      : highly optimized for speed search
 *---------------------------------------------------------------------------*/

void 
star_flag_crowded (StarList *list, double tol)
{
   int    i, j, k, nn = 0 ;
   int    ns = (int)list->nstar ;
   Star **star ;
   struct si *s ;

   CPKDEBUG ("star_flag_crowded") ;

   for (i = 0; i < ns; i++) list->data[i]->flag = 0 ;
   if (tol == 0.0) return ;

   /* Get two copies of input star pointers */
   star = star_ptr_dup (list->data, list->nstar) ;
   s = (struct si *) malloc (list->nstar * sizeof(struct si)) ;
   if (!star || !s) {
      cp_error ("Cannot flag crowded stars: aborting") ;
      if (star) free (star) ;
      if (s) free (s) ;
      return ;
   }

   /* Sort for y coordinates */
   qsort ((void*)star, list->nstar, sizeof(Star*), 
          (int (*)(const void *, const void *))sort_star_yprj) ;

   for (i = 0; i < list->nstar; i++) {
      s[i].s = star[i] ;
      s[i].i = i ;
   }

   /* Get a copy of input star list sorted for brightness */
   qsort ((void*)s, list->nstar, sizeof(struct si), 
          (int (*)(const void *, const void *))sort_si) ;

   /* Scan star list */
   i = 0 ;
   while (i < ns-1) {
      j = k = s[i].i ;
      if (star[j]->flag == 0 && star[j]->mag != IntNull) {
         while (k > 0 && star[j]->yprj-star[k]->yprj < tol) k-- ;
         while (k < ns-1 && star[k]->yprj-star[j]->yprj < tol) {
            k++ ;
            if (fabs (star[k]->xprj-star[j]->xprj) < tol && k != j &&
                star[k]->mag > star[j]->mag + 0.5) {
               /* Flag crowded stars */
               if (star[k]->flag != 1) nn++ ;
	       star[k]->flag = 1 ;
            }
         }
      }
      else
         star[j]->flag = 1 ;
      i++ ;
   }
   cp_message (0,3,"Flagged %d stars\n", nn) ;

   free (star) ;
   free (s) ;

   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_flag_crow_r
 * In        : star list, position tolerance
 * Out       : modified star list
 * Purpouse  : flag crowded stars to 1 except for the brightest of each group
 * Note      : same as 'star_flag_crowded' but works on xyrot fields
 *---------------------------------------------------------------------------*/

void 
star_flag_crow_r (StarList *list, double tol)
{
   int    i, j, k, nn = 0 ;
   int    ns = (int)list->nstar ;
   Star **star = list->data ;
   struct si *s ;

   CPKDEBUG ("star_flag_crow_r") ;

   for (i = 0; i < ns; i++) star[i]->flag = 0 ;
   if (tol == 0.0) return ;

   /* Get two copies of input star pointers */
   star = star_ptr_dup (list->data, list->nstar) ;
   s = (struct si *) malloc (list->nstar * sizeof(struct si)) ;
   if (!star || !s) {
      cp_error ("Cannot flag crowded stars: aborting") ;
      if (star) free (star) ;
      if (s) free (s) ;
      return ;
   }

   /* Sort for y coordinates */
   qsort ((void*)star, list->nstar, sizeof(Star*), 
          (int (*)(const void *, const void *))sort_star_yrot) ;

   for (i = 0; i < list->nstar; i++) {
      s[i].s = star[i] ;
      s[i].i = i ;
   }

   /* Get a copy of input star list sorted for brightness */
   qsort ((void*)s, list->nstar, sizeof(struct si), 
          (int (*)(const void *, const void *))sort_si) ;

   /* Scan star list */
   i = 0 ;
   while (i < ns-1) {
      j = k = s[i].i ;
      if (star[j]->flag == 0 && star[j]->mag != IntNull) {
         while (k > 0 && star[j]->yrot-star[k]->yrot < tol) k-- ;
         while (k < ns-1 && star[k]->yrot-star[j]->yrot < tol) {
            k++ ;
            if (fabs (star[k]->xrot-star[j]->xrot) < tol && k != j &&
                star[k]->mag > star[j]->mag + 1.0) {
               /* Flag crowded stars */
               if (star[k]->flag != 1) nn++ ;
	       star[k]->flag = 1 ;
            }
         }
      }
      /* else
         star[j]->flag = 1 ; */
      i++ ;
   }
   cp_message (0,3,"Flagged %d stars\n", nn) ;

   free (star) ;
   free (s) ;

   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_flag_mask
 * In        : 
 * Out       : 
 * Purpouse  : 
 * Note      : 
 *             
 *---------------------------------------------------------------------------*/

void 
star_flag_mask (StarList *list, shape mask_shape, double mask_p[])
{
   double  p1, p2, p3, p4, d, cd ;
   int     i, nn = 0 ;

   CPKDEBUG ("star_flag_mask") ;

   /* Set default values */
   switch (mask_shape) {
      case CIRCLE :
         if (mask_p[0]<-9.99e9 || mask_p[1]<-9.99e9) 
            star_list_center (list, &p1, &p2) ;
         if (mask_p[2]<-9.99e9) p3 = 0.0 ;
         if (mask_p[3]<-9.99e9) p4 = 1.0e9 ;
         break ;
      case RECT :
         if (mask_p[0]<-9.99e9 || mask_p[1]<-9.99e9 ||
             mask_p[2]<-9.99e9 || mask_p[3]<-9.99e9) 
            star_list_limits (list, &p1, &p2, &p3, &p4) ;
         break ;
      case MAG : 
         break ;
      case SHAPE :
         return ;
      default :
         cp_error ("Unrecognized mask shape: aborting") ;
         return ;
   }
   if (mask_p[0] < -9.99e9) mask_p[0] = p1 ;
   if (mask_p[1] < -9.99e9) mask_p[1] = p2 ;
   if (mask_p[2] < -9.99e9) mask_p[2] = p3 ;
   if (mask_p[3] < -9.99e9) mask_p[3] = p4 ;

   /* Mask values */
   switch (mask_shape) {
      case CIRCLE :
         for (i = 0; i < (int)list->nstar; i++) {
         cd = (list->spheric) ? cos (list->data[i]->ypix*M_PI/180.0) : 1.0 ;
            d = SQR ((list->data[i]->xpix-mask_p[0])*cd) + 
                SQR ((list->data[i]->ypix-mask_p[1])) ;
            if (d < SQR (mask_p[2]) || d > SQR (mask_p[3])) {
               list->data[i]->flag = 1 ;
               nn++ ;
            }
         }
         break ;
      case RECT :
         for (i = 0; i < (int)list->nstar; i++) {
            if (list->data[i]->xpix < mask_p[0] ||
                list->data[i]->xpix > mask_p[1] ||
                list->data[i]->ypix < mask_p[2] ||
                list->data[i]->ypix > mask_p[3]) {
               list->data[i]->flag = 1 ;
               nn++ ;
            }
         }
         break ;
      case MAG : 
         break ;
      case SHAPE :
         break ;
   }
   cp_message (0, 3, "Masked %d stars\n", nn) ;

   return ;
}

/*-----------------------------------------------------------------------------
 * Function  : star_list_limits
 * In        : 
 * Out       : 
 * Purpouse  : 
 *             
  *---------------------------------------------------------------------------*/

static int 
star_list_limits (StarList *list, double *xmin,  double *xmax, 
                  double *ymin, double *ymax)
{
   int      i ;

   *xmin = *xmax = list->data[0]->xpix ;
   *ymin = *ymax = list->data[0]->ypix ;
   for (i = 0; i < (int)list->nstar; i++) {
      *xmin = MIN (*xmin, list->data[i]->xpix) ;
      *xmax = MAX (*xmax, list->data[i]->xpix) ;
      *ymin = MIN (*ymin, list->data[i]->ypix) ;
      *ymax = MAX (*ymax, list->data[i]->ypix) ;
   }

   return 0 ;
}

void
star_check (Star **s, size_t n)
{
   size_t i ;
   cp_message (0,4,"index    id    mag       xpix      ypix\n") ;
   for (i = 0; i < MIN (6, n); i++)
      cp_message (0,4,"%5d %5s %f %f %f\n", s[i]->index, s[i]->id, 
            s[i]->mag, s[i]->xpix, s[i]->ypix) ;
}

void 
star_list_pix2prj (StarList *list, double origin[])
{
   if (list->proj && !memcmp (list->origin, origin, (size_t)2*sizeof(double))) 
      return ;
   coord_pix2prj (list->data, (int)list->nstar, list->spheric, origin) ;
   memcpy (list->origin, origin, (size_t)2*sizeof(double)) ;
   list->proj = TRUE ;
   return ;
}

/* Get working subsample from input star list */
StarList *
star_list_sample (StarList *sl, boolean mag)
{
   StarList  *out_sl ;
   size_t     i, n ;

   CPKDEBUG ("star_list_sample") ;

   out_sl = (StarList *) malloc (sizeof(StarList)) ;
   out_sl->data = (Star**) malloc (sl->nstar * sizeof(Star*)) ;

   for (n = i = 0; i < sl->nstar; i++) {
      if ((mag == FALSE || sl->data[i]->mag != IntNull) && 
            sl->data[i]->flag == 0) {
         out_sl->data[n] = sl->data[i] ;
         n++ ;
      }
   }
   out_sl->data = (Star**) realloc (out_sl->data, n*sizeof(Star*)) ;
   out_sl->nstar = n ;
   out_sl->spheric = sl->spheric ;
   out_sl->out_sph = sl->out_sph ;
   out_sl->param = sl->param ;
   out_sl->origin[0] = sl->origin[0] ;
   out_sl->origin[1] = sl->origin[1] ;

   return out_sl ;
}

